(***************************************************************************

Java context types

Krakatoa Project - Dmons research team - LRI - Universit Paris XI

$Id: java_context_types.ml,v 1.25 2007/03/11 14:28:40 nrousset Exp $

***************************************************************************)


(*s types and environments *)

type accessibility = [ `PUBLIC | `PROTECTED | `PRIVATE | `NONE ];;

type java_type =
  | Byte_type
  | Short_type
  | Integer_type
  | Float_type
  | Boolean_type
  | Null_type
  | Type_type
  | Class_type of class_entry
  | Interface_type of interface_entry
  | Array_type of java_type
  | Prop_type
  | Type_var of string
  | Logic_type of string * java_type list
  | Real_type
      
and local_variable_entry =
    {
      local_variable_entry_name : string;
      mutable local_variable_entry_trans_name : string;
      mutable local_variable_entry_is_ref : bool;
      local_variable_entry_type : java_type;
    }
    
and instance_variable_entry =
    {
      instance_variable_entry_tag : int;
      instance_variable_entry_name : string;
      mutable instance_variable_entry_trans_name : string;
      instance_variable_entry_accessibility : accessibility;
      instance_variable_entry_is_static : bool;
      instance_variable_entry_is_final : bool;
      instance_variable_entry_class_or_interface : java_env_entry;
      instance_variable_entry_type : java_type;
      instance_variable_entry_is_ghost : bool;
      instance_variable_entry_is_model : bool;
    }
    
and constant_entry =
    {
      constant_entry_name : string;
      constant_entry_interface : interface_entry;
      constant_entry_type : java_type;
    }

and method_entry = 
    {
     method_entry_name : string;
     mutable method_entry_trans_name : string;
     method_entry_accessibility : accessibility;
     method_entry_is_static : bool;
     method_entry_is_pure : bool;
     method_entry_return_type : java_type ;
     method_entry_class_or_interface : java_env_entry;
     mutable method_entry_is_predicate : bool;
     method_entry_routine :  routine_entry;
     method_entry_sep_specs : string list;
     mutable method_entry_graph : java_env_entry list
   }
    
and logic_type_entry = {
    mutable logic_type_entry_name : string;
    mutable logic_type_entry_vars : string list
  }
      
and logic_entry = 
    {
     logic_entry_name : string;
     logic_entry_return_type : java_type;
     mutable logic_entry_effects : effects ;
     logic_entry_parameters : t;
    }
    
and axiom_entry = 
    {
      axiom_entry_name : string;
      mutable axiom_entry_effects : effects ;
    }
    
and constructor_entry = 
    {
     mutable constructor_entry_trans_name : string;      
     constructor_entry_class : class_entry;
     constructor_entry_routine : routine_entry;
     mutable constructor_entry_graph : java_env_entry list
    }

and routine_entry =
    {
     routine_entry_parameters : t;
     mutable routine_entry_parameters_as_local_var : t;
     mutable routine_entry_produce_body : bool;
     mutable routine_entry_local_vars : t;
     mutable routine_entry_effects : effects;
     mutable routine_entry_int_array_writes_nothing : bool;
     mutable routine_entry_float_array_writes_nothing : bool;
     mutable routine_entry_bool_array_writes_nothing : bool;
     mutable routine_entry_obj_array_writes_nothing : bool;
     mutable routine_entry_fields_writes_nothing : 
       instance_variable_entry Inttagset.t;
     mutable routine_entry_use_transactions : bool
   }

and effects = 
    {
      reads_alloc : bool;
      writes_alloc : bool;
      writes_everything : bool;
      reads_alloc_table : bool;
      writes_alloc_table : bool;
      reads : instance_variable_entry Inttagset.t;
      writes : instance_variable_entry Inttagset.t;
      reads_int_array : bool;
      reads_int_assigned_array : bool;
      writes_int_array : bool;
      reads_float_array : bool;
      writes_float_array : bool;
      reads_bool_array : bool;
      writes_bool_array : bool;
      reads_obj_array : bool;
      writes_obj_array : bool;
      throws : bool (*i class_entry Inttagset.t i*);
      breaks : bool (*i label_entry Inttagset.t i*);
      continue : bool
    }

and package_entry =
    {
      package_entry_name : string;
      mutable package_entry_directories : string list;
      mutable package_entry_contents_read : bool;
      mutable package_entry_contents : t;
    }
    
and class_entry =
    {
      class_entry_package : string;
      class_entry_name : string;
      mutable class_entry_extends : class_entry option;
      mutable class_entry_implements : interface_entry list;
      mutable class_entry_contents : t;
      mutable class_entry_constructors : constructor_entry list;
      mutable class_entry_invariant_effects : effects;
      mutable class_entry_static_invariant_effects : effects;
      mutable class_entry_invariant_args : (string * Why.base_type) list;
      mutable class_entry_static_invariant_args : (string * Why.base_type) list;
      mutable class_entry_restore_invariant_args : (string * Why.base_type) list;
      mutable class_entry_representation_invariant_effects : effects;
    }

and interface_entry =
    {
      interface_entry_name : string;
      mutable interface_entry_extends : interface_entry list;
      mutable interface_entry_contents : t;
    }

and java_env_entry = 
  | Package_entry of package_entry
  | Class_entry of class_entry
  | Interface_entry of interface_entry
  | Instance_variable_entry of instance_variable_entry
  | Constant_entry of constant_entry
  | Method_entry of method_entry list
  | Constructor_entry of constructor_entry
  | Local_variable_entry of local_variable_entry 
  | Logic_type_entry of logic_type_entry 
  | Logic_entry of logic_entry

and t = (string * java_env_entry) list
;;

