(***************************************************************************

Lexer for JavaCard source files

VerifiCard Project - Dmons research team - LRI - Universit Paris XI

$Id: java_lexer.mll,v 1.18 2007/03/11 14:28:40 nrousset Exp $

***************************************************************************)


{

  open Java_parser

  exception Lexical_error of string

  let string_buffer = Buffer.create 17

  let kw_table = 
    let table = Hashtbl.create 17 in
    let _ = 
      List.iter
	(fun (s,t) -> Hashtbl.add table s t)
	[ "abstract", ABSTRACT;
	  "boolean", BOOLEAN;
	  "break", BREAK(Location.dummy);
	  "byte", BYTE;
	  "byvalue", BYVALUE;
	  "case", CASE;
	  "cast", CAST;
	  "catch", CATCH;
	  "char", CHAR;
	  "class", CLASS;
	  "const", CONST;
	  "continue", CONTINUE(Location.dummy);
	  "default", DEFAULT;
	  "do", DO(Location.dummy);
	  "double", DOUBLE;
	  "else", ELSE;
	  "extends", EXTENDS;
	  "false", FALSE(Location.dummy);
	  "final", FINAL;
	  "finally", FINALLY;
	  "float", FLOAT;
	  "for", FOR(Location.dummy);
	  "future", FUTURE;
	  "generic", GENERIC;
	  "goto", GOTO;
	  "if", IF(Location.dummy);
	  "implements", IMPLEMENTS;
	  "import", IMPORT;
	  "inner", INNER;
	  "instanceof", INSTANCEOF;
	  "int", INT;
	  "interface", INTERFACE;
	  "long", LONG;
	  "native", NATIVE;
	  "new", NEW(Location.dummy);
	  "null", NULL(Location.dummy);
	  "operator", OPERATOR;
	  "outer", OUTER;
	  "package", PACKAGE;
	  "private", PRIVATE;
	  "protected", PROTECTED;
	  "public", PUBLIC;
	  "rest", REST;
	  "return", RETURN(Location.dummy);
	  "short", SHORT;
	  "static", STATIC;
	  "super", SUPER(Location.dummy);
	  "switch", SWITCH(Location.dummy);
	  "synchronized", SYNCHRONIZED(Location.dummy);
	  "this", THIS(Location.dummy);
	  (* "threadsafe" ? *)
	  "throw", THROW(Location.dummy);
	  "throws", THROWS;
	  "transient", TRANSIENT;
	  "true", TRUE(Location.dummy);
	  "try", TRY(Location.dummy);
	  "var", VAR;
	  "void", VOID;
	  "volatile", VOLATILE;
	  "while", WHILE(Location.dummy);	
	]
    in table

  let find_loc lexbuf =
    Location.find 
      !Java_parser_base.cur_base 
      (Lexing.lexeme_start lexbuf) 
      (Lexing.lexeme_end lexbuf) 

  let id_or_kw lexbuf s =
    let l = find_loc lexbuf in
    try
      let k = Hashtbl.find kw_table s in
      (*i
	prerr_string "Keyword ";
	prerr_endline s;
	i*)
      begin
	match k with
	  | FALSE _ -> FALSE(l)
	  | NEW _ -> NEW(l)
	  | NULL _ -> NULL(l)
	  | SUPER _ -> SUPER(l)
	  | THIS _ -> THIS(l)
	  | TRUE _ -> TRUE(l)
	  | _ -> k
      end
    with
	Not_found ->
	  (*i
	    prerr_string "Ident ";
	    prerr_endline s;
	    i*)
	  ID(l,s)


  let remove_last_char s = String.sub s 0 (pred (String.length s))

  let jml_spec_token base jml_string =
(*i
    Format.fprintf Config.log "In file %s, parsing JML Spec: %s@."
      (Location.base_filename base) jml_string;
i*)
    match Jml_syntax.parse_jml_specification base jml_string with 
    | Ast_types.Jml_pure -> JML_PURE
    | Ast_types.Jml_spec_public -> JML_SPEC_PUBLIC
    | Ast_types.Jml_declaration d -> JML_DECLARATIONS(d)
    | Ast_types.Jml_method_specification s -> JML_METHOD_SPECIFICATION(s)
    | Ast_types.Jml_loop_annotation la -> JML_LOOP_ANNOTATION(la)
    | Ast_types.Jml_assertion a -> JML_ASSERTION(a)
    | Ast_types.Jml_annotation_statement Ast_types.Set_statement s -> JML_ANNOTATION_STATEMENT(Ast_types.Set_statement(s))
    | Ast_types.Jml_axiom(id,e) -> JML_AXIOM(id,e)
    | Ast_types.Jml_type(id, idl) -> JML_TYPE(id, idl)
    | Ast_types.Jml_logic_reads(id,t,p,r) -> JML_LOGIC_READS(id,t,p,r)
    | Ast_types.Jml_logic_def(id,t,p,e) -> JML_LOGIC_DEF(id,t,p,e)


  let char_for_backslash = function
      'n' -> '\n'
    | 't' -> '\t'
    | 'b' -> '\b'
    | 'r' -> '\r'
   (* | 'f' -> '\f' pas accepte par ocaml *)
    | c   -> c

}

  let backslash_escapes =
    ['\\' '"' '\'' 'n' 't' 'b' 'r' 'f' (* octal manque ! *)]

rule token = parse
  | [' ' '\t']+
      { token lexbuf }
  | '\n' 
      { Java_parser_base.next_line lexbuf; token lexbuf }
  (* evil DOS (NR) *)
  | "\r\n" 
      { Java_parser_base.next_line lexbuf; token lexbuf }
  | "/**/"
      { token lexbuf }
  | "/**"
      { Buffer.clear string_buffer;
	DOC_COMMENT(doc_comment lexbuf) }
  | "/*@"
      { Buffer.clear string_buffer;
	let base = 
	  Location.offset 
	    !Java_parser_base.cur_base (Lexing.lexeme_end lexbuf) 
	in
	jml_spec_token base (jml_spec lexbuf) }
  | "//@"
      { Buffer.clear string_buffer;
	let base = 
	  Location.offset 
	    !Java_parser_base.cur_base (Lexing.lexeme_end lexbuf) 
	in
	jml_spec_token base (jml_line_spec lexbuf) }
  | "/*"
      { comment lexbuf; token lexbuf }
  | "//" ([^'\n''@'][^'\n']*)? '\n'
      { Java_parser_base.next_line lexbuf; token lexbuf }
  | ['A'-'Z''a'-'z''_']['A'-'Z''a'-'z''0'-'9''_']* 
      { id_or_kw lexbuf (Lexing.lexeme lexbuf) }
  | ';'
      { SEMICOLON }
  | ','
      { COMMA }
  | '.'
      { DOT }
  | '+'                 
      { PLUS(find_loc lexbuf) }
  | '-'       
      { MINUS(find_loc lexbuf) }
  | "++"                 
      { PLUSPLUS(find_loc lexbuf) }
  | "--"       
      { MINUSMINUS(find_loc lexbuf) }
  | '*'
      { STAR }
  | '/'
      { SLASH }
  | '%'
      { PERCENT }
  | "&"
      { AMPERSAND }
  | "|"
      { VERTICALBAR }
  | "&&"
      { AMPERSANDAMPERSAND }
  | "||"
      { VERTICALBARVERTICALBAR }
  | "!"
      { BANG(find_loc lexbuf) }
  | "~"
      { TILDA(find_loc lexbuf) }
  | "^"
      { CARET }
  | "?"
      { QUESTIONMARK }
  | ":"
      { COLON }
  | "<<" | ">>" | ">>>"
      { SHIFT(Lexing.lexeme lexbuf) }
  | "=" 
      { EQ }
  | "*=" | "/=" | "%=" | "+=" | "-=" 
  | "<<=" | ">>=" | ">>>=" | "&=" | "^=" | "|="
      { ASSIGNOP(Lexing.lexeme lexbuf) }
  | ">" | "<" | "<=" | ">="
      { COMP(Lexing.lexeme lexbuf) }
  | "==" | "!="
      { EQOP(Lexing.lexeme lexbuf) }

      (* decimal constants *)

  | '0' | ['1'-'9']['0'-'9']*
      { INTEGER(find_loc lexbuf,int_of_string (Lexing.lexeme lexbuf)) }
  | ('0' | ['1'-'9']['0'-'9']*) ['l''L']
      { LONGINTEGER 
	  (find_loc lexbuf,
	   Int64.of_string (remove_last_char (Lexing.lexeme lexbuf))) }

      (* octal constants *)

  | '0'['0'-'7']+          
      { INTEGER
	  (find_loc lexbuf,
	   int_of_string ("0o" ^ (Lexing.lexeme lexbuf))) }
  | '0'['0'-'7']+ ['l''L']         
      { LONGINTEGER 
	  (find_loc lexbuf,
	   Int64.of_string 
	     ("0o" ^ (remove_last_char (Lexing.lexeme lexbuf)))) }

      (* hexadecimal constants *)

  | '0'['x''X']['0'-'9' 'A'-'F' 'a'-'f']+          
    { INTEGER(find_loc lexbuf,int_of_string (Lexing.lexeme lexbuf)) }
  | '0'['x''X']['0'-'9' 'A'-'F' 'a'-'f']+['l''L'] 
    { LONGINTEGER
	(find_loc lexbuf,
	 Int64.of_string (remove_last_char (Lexing.lexeme lexbuf))) }

      (* floating-point constants *)

  | (['0'-'9']+ as _int) '.' (['0'-'9']* as _frac) 
      (['e''E'](['-''+']?['0'-'9']+ as _exp))? ['f''F''d''D'] ?
      { REAL (find_loc lexbuf,Lexing.lexeme lexbuf) }

  | '.' (['0'-'9']+ as _frac) (['e''E'](['-''+']?['0'-'9']+ as _exp))? ['f''F''d''D'] ?
      { REAL (find_loc lexbuf,Lexing.lexeme lexbuf) }

  | (['0'-'9']+ as _int) ['e''E'] (['-''+']?['0'-'9']+ as _exp) ['f''F''d''D'] ?
      { REAL (find_loc lexbuf,Lexing.lexeme lexbuf) }

      (* character constants *)

  | "'" _ "'"
      { CHARACTER
	  (find_loc lexbuf,
	   Lexing.lexeme_char lexbuf 1) }

  | "'" '\\' backslash_escapes "'"
      { CHARACTER
	  (find_loc lexbuf,
	   char_for_backslash(Lexing.lexeme_char lexbuf 1)) }

  | "'" '\\' 'u' ['0'-'9''A'-'F'] ['0'-'9''A'-'F'] ['0'-'9''A'-'F'] ['0'-'9''A'-'F'] "'"
      { CHARACTER
	  (find_loc lexbuf, '#') } (* gd : unrepresented unicode character *)

  | '('
      { LEFTPAR(find_loc lexbuf) }
  | ')'
      { RIGHTPAR }
  | '{'
      { LEFTBRACE }
  | '}'
      { RIGHTBRACE }
  | '['
      { LEFTBRACKET }
  | ']'
      { RIGHTBRACKET }
  | '"'
      { let start = Lexing.lexeme_start lexbuf in
	Buffer.clear string_buffer;
	let eend = string lexbuf in
        let loc = Location.find !Java_parser_base.cur_base start eend in
	STRING(loc, Buffer.contents string_buffer) }
  | _ 
      { raise (Lexical_error (Lexing.lexeme lexbuf)) }
  | eof
      { EOF }

and string = parse
  | '"'
      { Lexing.lexeme_end lexbuf }
  | '\\' backslash_escapes
      { Buffer.add_string string_buffer (Lexing.lexeme lexbuf);
	string lexbuf }
  | '\\' _ 
      { raise (Lexical_error "unknown escape sequence in string") }
  | ['\n' '\r']
      { (* no \n anymore in strings since java 1.4 *)
	raise (Lexical_error "string not terminated");
	(* Java_parser_base.next_line lexbuf; 
	Buffer.add_string string_buffer (Lexing.lexeme lexbuf);
	string lexbuf*) }
  | [^ '\n' '\\' '"']+ 
      { Buffer.add_string string_buffer (Lexing.lexeme lexbuf);
	string lexbuf }
  | eof
      { raise (Lexical_error "string not terminated") }

and comment = parse
  | "*/"                
      { () }
  | '\n'
      { Java_parser_base.next_line lexbuf; comment lexbuf }
  | [^'*''\n']+             
      { comment lexbuf }
  | _                   
      { comment lexbuf }
  | eof                 
      { raise (Lexical_error "comment not terminated") }

and doc_comment = parse
  | "*/"                
      { Buffer.contents string_buffer }
  | '\n'
      { Java_parser_base.next_line lexbuf;  
	Buffer.add_string string_buffer (Lexing.lexeme lexbuf);
	doc_comment lexbuf }
  | [^'*''\n']+
      { Buffer.add_string string_buffer (Lexing.lexeme lexbuf);
	doc_comment lexbuf }
  | _                   
      { Buffer.add_string string_buffer (Lexing.lexeme lexbuf);
	doc_comment lexbuf }
  | eof
      { raise (Lexical_error "comment not terminated") }

and comment_in_jml = parse
  | "*/"                
      { Buffer.add_string string_buffer (Lexing.lexeme lexbuf) }
  | '\n'
      { Java_parser_base.next_line lexbuf; 
	Buffer.add_string string_buffer (Lexing.lexeme lexbuf);
	comment_in_jml lexbuf }
  | [^'*''\n']+             
      { Buffer.add_string string_buffer (Lexing.lexeme lexbuf);
	comment_in_jml lexbuf }
  | _                   
      { Buffer.add_string string_buffer (Lexing.lexeme lexbuf);
	comment_in_jml lexbuf }
  | eof                 
      { raise (Lexical_error "jml spec not terminated") }

and jml_spec = parse
  | "/*"
      { Buffer.add_string string_buffer (Lexing.lexeme lexbuf);
	comment_in_jml lexbuf; 
	jml_spec lexbuf }
  | "//" [^'\n']* '\n'
      { Java_parser_base.next_line lexbuf; 
	Buffer.add_string string_buffer (Lexing.lexeme lexbuf);
	jml_spec lexbuf }
  | "*/"                
      { Buffer.contents string_buffer }
  | '\n' | "\r\n"
      { Java_parser_base.next_line lexbuf;  
	Buffer.add_string string_buffer (Lexing.lexeme lexbuf);
	jml_spec lexbuf }
  | [^'@''*''\n''/']+
      { Buffer.add_string string_buffer (Lexing.lexeme lexbuf);
	jml_spec lexbuf }
  | _                   
      { Buffer.add_string string_buffer (Lexing.lexeme lexbuf);
	jml_spec lexbuf }
  | eof
      { raise (Lexical_error "jml spec not terminated") }

and jml_line_spec = parse
  | '\n' | '\r' | "\r\n"
      { Java_parser_base.next_line lexbuf;  
	Buffer.contents string_buffer}
  | [^'\n''\r']+
      { Buffer.add_string string_buffer (Lexing.lexeme lexbuf);
	jml_line_spec lexbuf }
(*
  | _                   
      { Buffer.add_string string_buffer (Lexing.lexeme lexbuf);
	jml_line_spec lexbuf }
*)
  | eof
      { raise (Lexical_error "jml spec not terminated") }


