(***************************************************************************

Main file for Krakatau

VerifiCard Project - Dmons research team - LRI - Universit Paris XI

$Id: main.ml,v 1.58 2007/07/23 21:51:21 paulin Exp $

***************************************************************************)

open Format

(* Some stuff on files *)

let open_file_preserving_header_gen magic_string_begin f =
(*i
  if Sys.file_exists f then
    begin
      Sys.rename f (f ^ ".bak");
      let c_new = open_out f
      and c_old = open_in (f ^ ".bak") 
      in
      try
	while true do
	  let s = input_line c_old in
	  output_string c_new (s^"\n");
	  if s = magic_string_begin then raise Exit;
	done;
	assert false
      with 
	| Exit -> c_new,Some c_old
	| End_of_file -> 
	    (* the magic line has not been found !!! *) 
	    output_string c_new "\n\n";
	    output_string c_new magic_string_begin;
	    output_string c_new "\n";
	    close_in c_old;
	    c_new, None
    end
  else
i*)
    let c_new = open_out (f ^ ".tmp")  in
    output_string c_new "\n\n";
    output_string c_new magic_string_begin;
    output_string c_new "\n";
    c_new, f
;;

let close_file_preserving_footer_gen magic_string_end c_new f =
  output_string c_new "\n\n";
  output_string c_new magic_string_end;
  output_string c_new "\n";
  close_out c_new;
  if Sys.file_exists f then 
    begin
      let old_md5 = Digest.file f
      and new_md5 = Digest.file (f ^ ".tmp")
      in
      if old_md5 <> new_md5
      then
	begin
	  Sys.rename f (f ^ ".bak");
	  Sys.rename (f ^ ".tmp") f
	end
      else
	Sys.remove (f ^ ".tmp")
    end
  else
    Sys.rename (f ^ ".tmp") f

let open_file_preserving_header = 
  open_file_preserving_header_gen
    "(*** THIS FILE IS GENERATED BY KRAKATOA - DO NOT EDIT ***)"

let close_file_preserving_footer = 
  close_file_preserving_footer_gen
    ""

let open_file_preserving_header_why = 
  open_file_preserving_header_gen
    "(* THIS FILE IS GENERATED BY KRAKATOA - DO NOT EDIT *)"

let close_file_preserving_footer_why = 
  close_file_preserving_footer_gen
    ""
	  
let open_file_preserving_header_sx = 
  open_file_preserving_header_gen
    ";; THIS FILE IS GENERATED BY KRAKATOA - DO NOT EDIT"

let close_file_preserving_footer_sx = 
  close_file_preserving_footer_gen
    ""
	  
let open_file_preserving_header_rv = 
  open_file_preserving_header_gen
    ";; THIS FILE IS GENERATED BY KRAKATOA - DO NOT EDIT"

let close_file_preserving_footer_rv = 
  close_file_preserving_footer_gen
    ""
	  
let open_file_preserving_header_pvs = 
  open_file_preserving_header_gen
    "%% THIS FILE IS GENERATED BY KRAKATOA - DO NOT EDIT"

let close_file_preserving_footer_pvs = 
  close_file_preserving_footer_gen
    ""
	  
(* generation of files Class_spec.why Class_spec.v and Class_m.why
   for a given compilation unit Class and all methods m occuring
   on the command_line
*)

let translate_comp_unit spec_files f cu =
(*
  let (d,b) = (Filename.dirname f,Filename.basename f) in
 let filename,_ = Config.split_extension b in
  let filename = Filename.concat !Config.source_package filename in
*)
(*
  if !Config.use_value_model
  then Translate_value_model.tr_compilation_unit cu 
  else 
    if !Config.use_global_memory_model
    then Translate_global_memory_model.tr_compilation_unit cu 
      else 
	*)
    fprintf Config.log "Translation compilation unit %s.@." f;
    Translate_local_memory_model.tr_compilation_unit cu 
;;


let all_spec_files comp_units = 
  List.map
    (fun (f,_) -> 
       let d,b = Filename.dirname f, Filename.basename f in
       if d = !Config.source_package
       then
	 fst (Config.split_extension b)
       else
	 fst (Config.split_extension f))
    comp_units
;;

(* the main program *)

let main () =
  (* banner *)
  printf "Krakatoa version %s - %s@." Version.version Version.date;
  try

    (* parsing command line arguments *)
    Config.parse_args ();

    if !Config.parse_only 
    then
      begin
	List.iter
	  (fun f -> let _ = Java_syntax.parse f in ())
	  (Config.files_to_parse ());
	  printf "Parsing OK.@.";
	  exit 0;
      end;

    (* creating the initial environment from the value of KrakatoaPath *)
    let init_env = Typing.initial_env() in
    if !Config.dump_env then
      begin
	fprintf Config.log "dumping initial environment in krakatoa.phase0env.log.@.";
	let c = open_out "krakatoa.phase0env.log" in
	let form = formatter_of_out_channel c in
	Java_context.fprintf_java_context form init_env;
	fprintf form "@.";
	close_out c
      end;

    (* Typing phase 1 : collecting all class and interface names in all
       imported packages *)
    fprintf Config.log "Typing phase 1 begins.@.";
    let genv = Typing.phase1_imported_files init_env in
    fprintf Config.log "Typing phase 1 ends.@.";
    if !Config.dump_env then
      begin
	fprintf Config.log "dumping environment after phase1 in krakatoa.phase1env.log.@.";
	let c = open_out "krakatoa.phase1env.log" in
	let form = formatter_of_out_channel c in
	Java_context.fprintf_java_context form genv;
	fprintf form "@.";
	close_out c
      end;

    (* Typing phase 2 : collecting all declarations, resolving type names 
       in them *)
    fprintf Config.log "Typing phase 2 begins.@.";
    let partially_typed_imported_comp_units =
      List.map
	(fun (f,cu) -> 
	   fprintf Config.log "Typing phase 2 of %s.@." f;
	   (f,Typing.phase2 genv f cu))
	!Typing.additional_files_to_type
    in
    fprintf Config.log "Handling constructor overloading.@.";    
    Typing.constructor_overloading genv;
    fprintf Config.log "Typing phase 2 ends.@.";    
    if !Config.dump_env then
      begin
	fprintf Config.log "dumping environment after phase2 in krakatoa.phase2env.log.@.";
	let c = open_out "krakatoa.phase2env.log" in
	let form = formatter_of_out_channel c in
	Java_context.fprintf_java_context form genv;
	fprintf form "@.";
	close_out c
      end;
    (*i
      List.iter
      (fun (f,cu) -> 
      printf "********* PARTIALLY TYPED AST OF %s ******************@." f;
      Ast.printf_compilation_unit cu;
      printf "@.";
      printf "********* END OF PARTIALLY TYPED AST *****************@.")
      partially_typed_imported_comp_units;
      i*)

    (* Typing phase 3: name resolution and typing of body of methods 
       and initializers of static fields *)

    fprintf Config.log "Typing phase 3 begins.@.";
    let typed_imported_comp_units =
      List.map
	(fun (f,cu) -> 
	   if !Config.verbose >= 2 then
	     printf "Typing phase 3 of %s.@." f;
	   (f,Typing.phase3 genv f cu))
	partially_typed_imported_comp_units
    in
    fprintf Config.log "Typing phase 3 ends.@.";
    if !Config.dump_env then
      begin
	fprintf Config.log 
	  "dumping environment after phase3 in krakatoa.phase3env.log.@.";
	let c = open_out "krakatoa.phase3env.log" in
	let form = formatter_of_out_channel c in
	Java_context.fprintf_java_context form genv;
	fprintf form "@.";
	close_out c
      end;
    (*i
      List.iter
      (fun (f,cu) -> 
      printf "********* FULL TYPED AST OF %s ******************@." f;
      Ast.printf_compilation_unit cu;
      printf "@.";
      printf "********* END OF FULL TYPED AST *****************@.")
      typed_imported_comp_units;
      i*)

    fprintf Config.log "Start of the pre-translation phase.@.";

    (* The pre-translation phase: computing of method effects, 
       disambiguation of names, and such *)
    Pre_translate.pre_translation genv 
      (List.map snd typed_imported_comp_units);

    fprintf Config.log "End of the pre-translation phase.@.";


    (* Call graph *)
    fprintf Config.log "Start of the call graph phase.@.";
    List.iter (fun (_, cu) -> Call_graph.compilation_unit cu) typed_imported_comp_units;
    if !Config.print_graph then Call_graph.print_graph genv;
    fprintf Config.log "End of the call graph phase.@.";


    (* the specifications from classes *)


    let spec_files = all_spec_files typed_imported_comp_units in

    let (why_decls,coq_decls) =
      List.fold_left
	(fun (wa,ca) (f,cu) ->
	   let (w,c) = translate_comp_unit spec_files f cu (List.map snd typed_imported_comp_units)
	   in (List.rev_append w wa,List.rev_append c ca))
	([],[])
	typed_imported_comp_units
    in
    let why_decls = why_decls @ coq_decls in
    let fields = Model.get_fields_transaction_2 [] genv in
    let field_names = List.map (fun ive -> ive.Java_context_types.instance_variable_entry_trans_name) fields in
    let field_names_restore = List.map (fun ive -> ive.Java_context_types.instance_variable_entry_trans_name ^ "_restore") fields in
    let beginTransaction_pre = Why.LPred("eq_int", [Why.LVar("transactionDepth"); Why.LConst(Why.Prim_int("0"))]) in
    let beginTransaction_reads = List.append ["intA"; "boolA"; "objA"] field_names in
    let beginTransaction_writes = List.append ["transactionDepth"; "intA_restore"; "boolA_restore"; "objA_restore"; "intA_assigned"] field_names_restore in
    let beginTransaction_arrays_body = Why.LAnd(Why.LPred("eq_int", [Why.LApp("array_acc", [Why.LVar("intA_restore"); Why.LVar("v"); Why.LVar("i")]); 
                                                                     Why.LApp("array_acc", [Why.LVar("intA"); Why.LVar("v"); Why.LVar("i")])]), 
						Why.LAnd(Why.LPred("eq_bool", [Why.LApp("array_acc", [Why.LVar("intA_assigned"); Why.LVar("v"); Why.LVar("i")]); 
                                                                               Why.LConst(Why.Prim_bool(false))]), 
                                                Why.LAnd(Why.LPred("eq_bool", [Why.LApp("array_acc", [Why.LVar("boolA_restore"); Why.LVar("v"); Why.LVar("i")]);
                                                                               Why.LApp("array_acc", [Why.LVar("boolA"); Why.LVar("v"); Why.LVar("i")])]), 
                                                         Why.LPred("eq", [Why.LApp("array_acc", [Why.LVar("objA_restore"); Why.LVar("v"); Why.LVar("i")]); 
                                                                                Why.LApp("array_acc", [Why.LVar("objA"); Why.LVar("v"); Why.LVar("i")])])))) in
    let beginTransaction_arrays = 
      Why.LForall("v", ([],"value"), 
		  Why.LForall("i", ([],"int"), beginTransaction_arrays_body)) 
    in
    let beginTransaction_fields = Model.store_fields fields in
    let beginTransaction_post = 
      Why.LAnd(Why.LPred("eq_int", 
			 [Why.LVar("transactionDepth"); 
			  Why.LConst(Why.Prim_int("1"))]), 
               Why.LAnd(beginTransaction_arrays, beginTransaction_fields)) 
    in 
    let beginTransaction_annot =  
      Why.Annot_type(beginTransaction_pre, 
		     Why.unit_type, 
		     beginTransaction_reads, 
		     beginTransaction_writes, 
		     beginTransaction_post, []) 
    in
    let beginTransaction_type = 
      Why.Prod_type("tt", Why.unit_type, beginTransaction_annot) 
    in
    let beginTransaction_decl = 
      Why.Param(false,"JCSystem_beginTransaction_parameter", 
		beginTransaction_type) 
    in 
    let abortTransaction_pre = 
      Why.LPred("eq_int", 
		[Why.LVar("transactionDepth"); 
		 Why.LConst(Why.Prim_int("1"))]) 
    in 
    let abortTransaction_reads = 
      List.append ["volatile"; "intA_restore"; "boolA_restore"; "objA_restore"] field_names_restore 
    in
    let abortTransaction_writes = 
      List.append ["transactionDepth"; "intA"; "boolA"; "objA"] field_names 
    in
    let abortTransaction_arrays_if = 
      Why.LApp("negb", 
	       [Why.LApp("acc", [Why.LVar("volatile"); Why.LVar("v")])]) 
    in
    let abortTransaction_arrays_then = 
      Why.LAnd(Why.LPred("eq_int", 
			 [Why.LApp("array_acc", 
				   [Why.LVar("intA"); 
				    Why.LVar("v"); 
				    Why.LVar("i")]); 
                          Why.LApp("array_acc", 
				   [Why.LVar("intA_restore"); 
				    Why.LVar("v"); 
				    Why.LVar("i")])]), 
               Why.LAnd(Why.LPred("eq_bool", 
				  [Why.LApp("array_acc", 
					    [Why.LVar("boolA"); 
					     Why.LVar("v"); 
					     Why.LVar("i")]);
                                   Why.LApp("array_acc", 
					    [Why.LVar("boolA_restore"); 
					     Why.LVar("v"); 
					     Why.LVar("i")])]), 
                        Why.LPred("eq", 
				  [Why.LApp("array_acc", 
					    [Why.LVar("objA"); 
					     Why.LVar("v"); 
					     Why.LVar("i")]); 
                                   Why.LApp("array_acc", 
					    [Why.LVar("objA_restore"); 
					     Why.LVar("v"); 
					     Why.LVar("i")])]))) 
    in
    let abortTransaction_arrays_else = Why.LAnd(Why.LPred("eq_int", [Why.LApp("array_acc", [Why.LVar("intA"); Why.LVar("v"); Why.LVar("i")]); 
                                                                     Why.LApp("array_acc", [Why.LVar("intA@"); Why.LVar("v"); Why.LVar("i")])]), 
                                                Why.LAnd(Why.LPred("eq_bool", [Why.LApp("array_acc", [Why.LVar("boolA"); Why.LVar("v"); Why.LVar("i")]); 
                                                                               Why.LApp("array_acc", [Why.LVar("boolA@"); Why.LVar("v"); Why.LVar("i")])]), 
                                                         Why.LPred("eq", [Why.LApp("array_acc", [Why.LVar("objA"); Why.LVar("v"); Why.LVar("i")]); 
                                                                                Why.LApp("array_acc", [Why.LVar("objA@"); Why.LVar("v"); Why.LVar("i")])]))) in
    let abortTransaction_arrays = 
      Why.LForall("i", ([],"int"), 
		  Why.LIf(abortTransaction_arrays_if, 
			  abortTransaction_arrays_then, abortTransaction_arrays_else)) 
    in
    let abortTransaction_fields = Model.restore_fields fields in
    let abortTransaction_post = 
      Why.LAnd(Why.LPred("eq_int", 
			 [Why.LVar("transactionDepth"); 
			  Why.LConst(Why.Prim_int("0"))]), 
	       Why.LAnd(Why.LForall("v", ([],"value"), 
				    abortTransaction_arrays), abortTransaction_fields)) 
    in 
    let abortTransaction_annot =  
      Why.Annot_type(abortTransaction_pre, Why.unit_type, abortTransaction_reads, abortTransaction_writes, abortTransaction_post, []) 
    in
    let abortTransaction_type = 
      Why.Prod_type("tt", Why.unit_type, abortTransaction_annot) 
    in
    let abortTransaction_decl = Why.Param(false,"JCSystem_abortTransaction_parameter", abortTransaction_type) in 
    let possibleCardTear_post = Why.LAnd(Why.LPred("eq_int", [Why.LVar("transactionDepth"); Why.LVar("transactionDepth@")]), 
                                         Why.LAnd(Why.LPred("eq", [Why.LVar("intA"); Why.LVar("intA@")]), 
                                                  Why.LAnd(Why.LPred("eq", [Why.LVar("boolA"); Why.LVar("boolA@")]), 
                                                           Why.LAnd(Why.LPred("eq", [Why.LVar("objA"); Why.LVar("objA@")]), 
                                                                              Model.unchange_fields fields)))) in
    let clear_transient_data = Why.LForall("i",
                                           ([],"int"),
                                           Why.LAnd(Why.LPred("eq_int", [Why.LApp("array_acc", [Why.LVar("intA"); Why.LVar("v"); Why.LVar("i")]); 
									 Why.LConst(Why.Prim_int("0"))]), 
				                    Why.LAnd(Why.LPred("eq_bool", [Why.LApp("array_acc", [Why.LVar("boolA"); Why.LVar("v"); Why.LVar("i")]); 
										   Why.LConst(Why.Prim_bool(false))]), 
							     Why.LPred("eq", [Why.LApp("array_acc", [Why.LVar("objA"); Why.LVar("v"); Why.LVar("i")]); 
										    Why.LVar("Null")])))) in
    let possibleCardTear_cardtear = 
      Why.LAnd(Why.LForall("v",
                           ([],"value"), 
                           Why.LImpl(Why.LPred("eq_bool", [Why.LApp("acc", [Why.LVar("volatile"); Why.LVar("v")]);
							   Why.LConst(Why.Prim_bool(true))]),
                                     clear_transient_data)),
               Why.LAnd(Why.LImpl(Why.LPred("neq_int", [Why.LVar("transactionDepth@"); Why.LConst(Why.Prim_int("0"))]),
                                  Why.LAnd(Model.restore_fields fields, 
                                           Why.LForall("v",
                                                       ([],"value"), 
                                                       Why.LImpl(Why.LPred("eq_bool", [Why.LApp("acc", [Why.LVar("volatile"); Why.LVar("v")]); 
										       Why.LConst(Why.Prim_bool(false))]),
                                                                 Why.LForall("i", ([],"int"), abortTransaction_arrays_then))))), 
                        Why.LAnd(Why.LImpl(Why.LPred("eq_int", [Why.LVar("transactionDepth@"); Why.LConst(Why.Prim_int("0"))]),
                                           Why.LAnd(Model.unchange_fields fields,
                                                    Why.LForall("v", 
                                                                ([],"value"), 
                                                                Why.LImpl(Why.LPred("eq_bool", [Why.LApp("acc", [Why.LVar("volatile"); Why.LVar("v")]);
												Why.LConst(Why.Prim_bool(false))]),
                                                                          Why.LForall("i", ([],"int"), abortTransaction_arrays_else))))),
                                 Why.LPred("eq_int", [Why.LVar("transactionDepth"); Why.LConst(Why.Prim_int("0"))])))) in
    let possibleCardTear_annot = Why.Annot_type(Why.LTrue, 
						Why.unit_type, 
						abortTransaction_reads, 
						abortTransaction_writes, 
						possibleCardTear_post, 
						[("CardTear",possibleCardTear_cardtear)]) in
    let possibleCardTear_type = Why.Prod_type("tt", Why.unit_type, possibleCardTear_annot) in
    let possibleCardTear_decl = Why.Param(false,"possibleCardTear", possibleCardTear_type) in
    let arrayCopyNonAtomic_pre = Why.LAnd(Why.LPred("neq", [Why.LVar("src"); Why.LVar("Null")]),
                                          Why.LAnd(Why.LPred("ge_int", [Why.LVar("srcOff"); Why.LConst(Why.Prim_int("0"))]),
						   Why.LAnd(Why.LPred("le_int", [Why.LApp("add_int", [Why.LVar("srcOff"); 
												      Why.LVar("length")]); Why.LApp("arraylength", [Why.LVar("alloc"); Why.LVar("src")])]),
							    Why.LAnd(Why.LPred("neq", [Why.LVar("dest"); Why.LVar("Null")]),
								     Why.LAnd(Why.LPred("ge_int", [Why.LVar("destOff"); Why.LConst(Why.Prim_int("0"))]),
									      Why.LAnd(Why.LPred("le_int", [Why.LApp("add_int", [Why.LVar("destOff"); Why.LVar("length")]); 
													Why.LApp("arraylength", [Why.LVar("alloc"); Why.LVar("dest")])]),
										       Why.LPred("ge_int", [Why.LVar("length"); Why.LConst(Why.Prim_int("0"))]))))))) in
    let arrayCopyNonAtomic_post_inTransaction_forall =
      Why.LForall("i", 
                  ([],"int"), 
                  Why.LImpl(Why.LAnd(Why.LPred("le_int", [Why.LConst(Why.Prim_int("0")); Why.LVar("i")]),
                                     Why.LPred("lt_int", [Why.LVar("i"); Why.LVar("length")])),
                            Why.LAnd(Why.LImpl(Why.LPred("eq_int", [Why.LApp("array_acc", [Why.LVar("intA@"); 
                                                                                      Why.LVar("dest"); 
                                                                                      Why.LApp("add_int", [Why.LVar("destOff"); Why.LVar("i")])]);
							       Why.LApp("array_acc", [Why.LVar("intA_restore@"); 
									              Why.LVar("dest_restore"); 
                                                                                      Why.LApp("add_int", [Why.LVar("destOff"); Why.LVar("i")])])]),
					       Why.LPred("eq_int", [Why.LApp("array_acc", [Why.LVar("intA_restore"); 
									              Why.LVar("dest_restore"); 
										      Why.LApp("add_int", [Why.LVar("destOff"); Why.LVar("i")])]);
							       Why.LApp("array_acc", [Why.LVar("intA"); 
									              Why.LVar("dest"); 
										      Why.LApp("add_int", [Why.LVar("destOff"); Why.LVar("i")])])])),
				     Why.LImpl(Why.LPred("neq_int", [Why.LApp("array_acc", [Why.LVar("intA@"); 
									               Why.LVar("dest"); 
										       Why.LApp("add_int", [Why.LVar("destOff"); Why.LVar("i")])]);
							        Why.LApp("array_acc", [Why.LVar("intA_restore@"); 
									               Why.LVar("dest_restore"); 
										       Why.LApp("add_int", [Why.LVar("destOff"); Why.LVar("i")])])]),
					       Why.LPred("eq_int", [Why.LApp("array_acc", [Why.LVar("intA_restore"); 
									              Why.LVar("dest_restore"); 
										      Why.LApp("add_int", [Why.LVar("destOff"); Why.LVar("i")])]);
							       Why.LApp("array_acc", [Why.LVar("intA_restore@"); 
									              Why.LVar("dest_restore"); 
										      Why.LApp("add_int", [Why.LVar("destOff"); Why.LVar("i")])])]))))) 
    in
    let arrayCopyNonAtomic_post_inTransaction_modif = 
      Why.LPred("array_modifiable", [Why.LVar("alloc");
		                     Why.LVar("intA_restore@");
		                     Why.LVar("intA_restore");
		                     Why.LApp("array_sub_loc", [Why.LVar("dest_restore");
					                        Why.LVar("destOff");
					                        Why.LApp("sub_int", [Why.LApp("add_int", [Why.LVar("destOff"); Why.LVar("length")]); Why.LConst(Why.Prim_int("1"))])])])
    in
    let arrayCopyNonAtomic_post_inTransaction = Why.LAnd(arrayCopyNonAtomic_post_inTransaction_forall, arrayCopyNonAtomic_post_inTransaction_modif) in
    let arrayCopyNonAtomic_post = 
      Why.LAnd(Why.LPred("eq_int", [Why.LVar("result"); Why.LApp("add_int", [Why.LVar("destOff"); Why.LVar("length")])]),
               Why.LAnd(Why.LForall("i", 
				     ([],"int"),
				     Why.LImpl(Why.LAnd(Why.LPred("le_int", [Why.LConst(Why.Prim_int("0")); Why.LVar("i")]),
							Why.LPred("lt_int", [Why.LVar("i"); Why.LVar("length")])),
					       Why.LPred("eq_int", [Why.LApp("array_acc", [Why.LVar("intA"); 
										           Why.LVar("dest"); 
											   Why.LApp("add_int", [Why.LVar("destOff"); Why.LVar("i")])]);
								    Why.LApp("array_acc", [Why.LVar("intA@"); 
										           Why.LVar("src"); 
											   Why.LApp("add_int", [Why.LVar("srcOff"); Why.LVar("i")])])]))),
			Why.LAnd(Why.LPred("array_modifiable", [Why.LVar("alloc");
								Why.LVar("intA@");
								Why.LVar("intA");
								Why.LApp("array_sub_loc", [Why.LVar("dest");
											   Why.LVar("destOff");
											   Why.LApp("sub_int", [Why.LApp("add_int", [Why.LVar("destOff"); 
															             Why.LVar("length")]); 
														Why.LConst(Why.Prim_int("1"))])])]),
				 Why.LAnd(Why.LImpl(Why.LPred("neq_int", [Why.LVar("transactionDepth"); Why.LConst(Why.Prim_int("0"))]),
                                                    arrayCopyNonAtomic_post_inTransaction),
				          Why.LAnd(Why.LImpl(Why.LPred("eq_int", [Why.LVar("transactionDepth"); Why.LConst(Why.Prim_int("0"))]),
						             Why.LPred("eq", [Why.LVar("intA_restore"); Why.LVar("intA_restore@")])),
						   Why.LAnd(Model.unchange_fields fields,
							    Why.LPred("eq_int", [Why.LVar("transactionDepth"); Why.LVar("transactionDepth@")]))))))) in
    let arrayCopyNonAtomic_tear = 
      Why.LAnd(Why.LImpl(Why.LPred("eq_bool", [Why.LApp("acc", [Why.LVar("volatile"); Why.LVar("dest")]); Why.LConst(Why.Prim_bool(false))]),
			 Why.LAnd(Why.LForall("i", 
				              ([],"int"),
				              Why.LImpl(Why.LAnd(Why.LPred("le_int", [Why.LConst(Why.Prim_int("0")); Why.LVar("i")]),
							         Why.LPred("lt_int", [Why.LVar("i"); Why.LVar("length")])),
					                Why.LOr(Why.LPred("eq_int", [Why.LApp("array_acc", [Why.LVar("intA"); 
										                            Why.LVar("dest"); 
											                    Why.LApp("add_int", [Why.LVar("destOff"); Why.LVar("i")])]);
								                     Why.LApp("array_acc", [Why.LVar("intA@"); 
										                            Why.LVar("src"); 
											                    Why.LApp("add_int", [Why.LVar("srcOff"); Why.LVar("i")])])]),
                                                                Why.LPred("eq_int", [Why.LApp("array_acc", [Why.LVar("intA"); 
										                            Why.LVar("dest"); 
											                    Why.LApp("add_int", [Why.LVar("destOff"); Why.LVar("i")])]);
								                     Why.LApp("array_acc", [Why.LVar("intA@"); 
										                            Why.LVar("dest"); 
											                    Why.LApp("add_int", [Why.LVar("destOff"); Why.LVar("i")])])])))),
				  Why.LAnd(Why.LImpl(Why.LPred("neq_int", [Why.LVar("transactionDepth"); Why.LConst(Why.Prim_int("0"))]),
                                                     Why.LAnd(Why.LPred("eq", [Why.LVar("dest_restore"); Why.LVar("dest")]),
						              Why.LForall("i", 
									  ([],"int"),
									  Why.LImpl(Why.LAnd(Why.LPred("le_int", [Why.LConst(Why.Prim_int("0")); Why.LVar("i")]),
											     Why.LPred("lt_int", [Why.LVar("i"); Why.LVar("length")])),
										    Why.LOr(Why.LPred("eq_int", [Why.LApp("array_acc", [Why.LVar("intA_restore"); 
														                        Why.LVar("dest"); 
															                Why.LApp("add_int", [Why.LVar("destOff"); Why.LVar("i")])]);
													         Why.LApp("array_acc", [Why.LVar("intA@"); 
														                        Why.LVar("src"); 
															                Why.LApp("add_int", [Why.LVar("srcOff"); Why.LVar("i")])])]),
											    Why.LPred("eq_int", [Why.LApp("array_acc", [Why.LVar("intA_restore"); 
														                        Why.LVar("dest"); 
															                Why.LApp("add_int", [Why.LVar("destOff"); Why.LVar("i")])]);
													         Why.LApp("array_acc", [Why.LVar("intA_restore@");
														                        Why.LVar("dest");
															                Why.LApp("add_int", [Why.LVar("destOff"); Why.LVar("i")])])])))))),
				           Why.LImpl(Why.LPred("eq_int", [Why.LVar("transactionDepth"); Why.LConst(Why.Prim_int("0"))]),
						     Why.LPred("eq", [Why.LVar("intA_restore"); Why.LVar("intA_restore@")]))))),
               possibleCardTear_cardtear) in
    let arrayCopyNonAtomic_annot = Why.Annot_type(arrayCopyNonAtomic_pre, 
						  Why.int_type, 
						  abortTransaction_reads, 
						  "intA_restore"::abortTransaction_writes,
						  arrayCopyNonAtomic_post,
						  [("CardTear",arrayCopyNonAtomic_tear)])
    in
    let arrayCopyNonAtomic_type = Why.Prod_type("dest_restore", Why.Base_type([], "value"), Why.Prod_type("src", Why.Base_type([], "value"), 
                                                Why.Prod_type("srcOff", Why.int_type,
                                                              Why.Prod_type("dest", Why.Base_type([], "value"),
                                                                            Why.Prod_type("destOff", Why.int_type,
                                                                                          Why.Prod_type("length", Why.int_type, arrayCopyNonAtomic_annot)))))) in
    let _arrayCopyNonAtomic_decl = Why.Param(false,"arrayCopyNonAtomic_parameter", arrayCopyNonAtomic_type) in
    let why_decls = 
      if !Config.javacard
      then beginTransaction_decl::abortTransaction_decl::why_decls
      else why_decls
    in
    let why_decls = 
      if !Config.cardtear
      then (* arrayCopyNonAtomic_decl:: *) possibleCardTear_decl::why_decls
      else why_decls
    in

 
    let why_decls = Model.output_why_model genv @ why_decls in

    let why_spec_filename = 
      Filename.concat !Config.source_package "Krak_spec.why" 
    in
    fprintf Config.log "Start building file %s.@." why_spec_filename;
    let c,f = open_file_preserving_header why_spec_filename in
    let form = formatter_of_out_channel c in
    Why.fprintf_why_decls form 
      (List.filter (function Why.Def _ -> false | _ -> true) why_decls);
    close_file_preserving_footer c f;
    fprintf Config.log "File %s written@." why_spec_filename; 



    (* The translation phase: generation of the model *)

(*
    let coq_model_name = 
      Filename.concat !Config.source_package 
	(if !Config.use_global_memory_model
	 then "Krak_global_model.v" 
	 else "Krak_model.v")
    in
    let c,f = open_file_preserving_header coq_model_name in
    let form = formatter_of_out_channel c in
    Model.output_model form genv;
    fprintf form "@.";
    fprintf form "(* Require Export ZArith. *)@.";
    fprintf form "Require Export Why.@.";
    fprintf form "(* Add LoadPath \"%s\". *)@." !Config.krakatoa_lib;
    fprintf form "(* Require Export replace. *)@.";
    fprintf form "(* Require Export Krak_model. *)@.";
    fprintf form "Export KrakClasses.@.";
    fprintf form "Export KrakGen.@.";
    fprintf form "Export KrakObj.@.";
    fprintf form "Export CGen.@.@.";
(*
    fprintf form "Load %s_tactics."
      (if !Config.use_global_memory_model then "krak" else "local");
    fprintf form "@.";    
*)
    fprintf form "(* Load %s_model_imports. *)@.@." 
      (if !Config.use_global_memory_model
       then "global" else "local"); 
    Why.output_coq_decls form coq_decls;
    close_file_preserving_footer c f;
    fprintf Config.log "File %s written@." coq_model_name; 
*)

(*
    if !Config.output_language = `SIMPLIFY
    then 
      begin
	let sx_model_name = 
	  Filename.concat !Config.source_package "Krak_model.sx" 
	in
	let c,f = open_file_preserving_header_sx sx_model_name in
	let form = formatter_of_out_channel c in
	Model.output_sx_model form genv;
	Why.output_sx_decls form coq_decls;
	close_file_preserving_footer_sx c f;
	fprintf Config.log "File %s written@." sx_model_name; 
      end;
    
    if !Config.output_language = `HARVEY
    then 
      begin
	let rv_model_name = 
	  Filename.concat !Config.source_package "Krak_model.rv" 
	in
	let c,f = open_file_preserving_header_rv rv_model_name in
	let form = formatter_of_out_channel c in
	Model.output_rv_model form genv;
	Why.output_rv_decls form coq_decls;
	close_file_preserving_footer_rv c f;
	fprintf Config.log "File %s written@." rv_model_name;
      end;
    
    if !Config.output_language = `PVS
    then 
      begin
	let pvs_model_name = 
	  Filename.concat !Config.source_package "Krak_model.pvs" 
	in
	let c,f = open_file_preserving_header_pvs pvs_model_name in
	let form = formatter_of_out_channel c in
	Model.file_copy 
	  (Filename.concat !Config.krakatoa_lib "local_memory.pvs") form;
	fprintf form "Krak_model : THEORY@\nBEGIN@\nimporting modifiable@\n";
	Model.output_pvs_model form genv;
	Why.output_pvs_decls form coq_decls;
	fprintf form "END Krak_model@.";
	close_file_preserving_footer_pvs c f;
	fprintf Config.log "File %s written@." pvs_model_name; 
      end;
*)
    
    (* The spec_imports.v file *)
    let spec_imports_filename = 	    
      Filename.concat !Config.source_package "spec_imports.v" 
    in
    let c,c_old = open_file_preserving_header spec_imports_filename in
    let form = formatter_of_out_channel c in
    fprintf form "Add LoadPath \"%s\".@." !Config.krakatoa_lib;
    fprintf form "Require Export %s_memory_why.@."
      (if !Config.use_global_memory_model then "global" else "local");
    fprintf form "Require Export krak_tactics.@.";
    fprintf form "Require Export Krak_spec_why.@.";
(*
    fprintf form "Require Export Krak_%smodel.@."
      (if !Config.use_global_memory_model then "global_" else "");
*)
(*
    fprintf form "Load %s_tactics.@."
      (if !Config.use_global_memory_model then "krak" else "local");
*)
    close_file_preserving_footer c c_old;
    fprintf Config.log "File %s written@." spec_imports_filename; 

    (* the why programs files *)

    List.iter 
      (fun decl ->
	 match decl with
	   | Why.Def(name,f) -> 
	       printf "Generating Why program %s@." name;
	       let why_filename = 	    
		 Filename.concat !Config.source_package (name ^ ".why") 
	       in
	       let c,c_old = open_file_preserving_header why_filename in
	       let form = formatter_of_out_channel c in
	       Why.fprintf_why_decl form (Why.Def(name ^ "_body",f));
	       fprintf form "@.";
	       close_file_preserving_footer c c_old;
	       fprintf Config.log "File %s written@." why_filename; 
	   | _ -> ()) 
      why_decls;

    (* generation of Krak_validations.v *)

    if !Config.why_validation then
      begin    
	let coq_filename = 	    
	  Filename.concat !Config.source_package "Krak_validations.v" 
	in
	let c,c_old = open_file_preserving_header coq_filename in
	let form = formatter_of_out_channel c in
	Model.output_krak_validations form genv spec_files;
	close_file_preserving_footer c c_old;
	fprintf Config.log "File %s written@." coq_filename; 
      end;

    (* generation of the makefile *)
    begin
      let name = Filename.concat !Config.source_package "makefile" in
      let c = open_out name in
      Model.output_makefile 
	(formatter_of_out_channel c) genv spec_files;
      close_out c;
      fprintf Config.log "File %s written@." name; 
    end;
    Config.close_log ();
    (* that's all folks *)
    exit 0
  with
    | Typing.Typing_error msg ->
	Config.close_log ();
	eprintf "Typing error: %s@." msg; exit 1
    | Typing.Ill_typed msg ->
	Config.close_log ();
	eprintf "General typing error: %s@." msg; exit 1
    | Pre_translate.Unsupported msg ->
	Config.close_log ();
	eprintf "Unsupported feature in translation: %s@." msg; exit 2
    | Pre_translate.Ill_typed msg ->
	Config.close_log ();
	eprintf "Translation failed due to ill-typed statement: %s@." msg; exit 2
    | Sys_error msg ->
	Config.close_log ();	
	eprintf "sys error: %s@." msg; exit 1
(*
    | e ->
	fprintf Config.log "Unexpected exception! (%s)@." (Printexc.to_string e); 
	Config.close_log ();
	eprintf "unexpected error. Giving up.@.";
	raise e
*)
;;
  
  


Printexc.catch main ();;
