

open Format;;
open Java_context_types;;

let output_header form =
  fprintf form "(* Coq Model generated by Krakatoa. Do not edit *)@.";
  fprintf form "@.";
  fprintf form "Add LoadPath \"%s\".@." !Config.krakatoa_lib;
  fprintf form "Require Export preambule.@.";
(*  fprintf form "Require Export classes.@."; *)
  fprintf form "Require Export %s_memory.@."
    (if !Config.use_global_memory_model
     then "global" else "local");
  fprintf form "Require Export ZArith.@.";
  fprintf form "Require Export java_base.@.";
  fprintf form "Require Export Rbase.@.";
  fprintf form "Require Export Rfunctions.@.";
(* NR  fprintf form "Require Export List.@."; *)
  fprintf form "@.";
  fprintf form "@.";
;;

let rec get_all_classes acc env =
  List.fold_left
    (fun acc (_,entry) -> match entry with
       | Package_entry pe ->
	   get_all_classes acc pe.package_entry_contents
       | Interface_entry ie ->
           begin
	     match ie.interface_entry_extends with
	       | [] -> (ie.interface_entry_name, Some "Object")::acc
               | x::l ->
                  let lacc = List.map (fun x -> (ie.interface_entry_name, Some x.interface_entry_name)) l in 
                  let acc = List.append lacc acc in
                  (ie.interface_entry_name, Some x.interface_entry_name)::acc
	   end
       | Class_entry te ->
	   begin
	     match te.class_entry_extends with
	       | None -> 
		   if te.class_entry_name = "Object"
		   then (te.class_entry_name, None)::acc
		   else (te.class_entry_name, Some "Object")::acc
	       | Some x -> (te.class_entry_name, Some x.class_entry_name)::acc
	   end
       | _ -> acc)
    acc
    env
;;


let rec topo_sort a b acc =
  match a with
    | [] -> 
	if b=[] then List.rev acc else topo_sort b [] acc
    | x::c -> 
	match x with
	  | (_, None) -> topo_sort c b (x::acc)
	  | (_, Some super) ->
              if List.mem_assoc super acc
              then topo_sort c b (x::acc)
	      else topo_sort c (x::b) acc
;;

let get_classes env =
  let l = get_all_classes [] env in
    topo_sort l [] []
;;


let rec get_fields acc env =
  List.fold_left
    (fun acc (_,entry) -> match entry with
       | Package_entry pe ->
	   get_fields acc pe.package_entry_contents
       | Class_entry te ->
	   List.fold_left
	   (fun acc (_,entry) -> match entry with
	      | Instance_variable_entry ive ->
		  if ive.instance_variable_entry_is_static
		  then acc
		  else
		    (ive,te)::acc
	      | _ -> acc)
	   acc
	   te.class_entry_contents
       | _ -> acc)
    acc
    env
;;


let rec get_fields_transaction acc env =
  List.fold_left
    (fun acc (_,entry) -> match entry with
       | Package_entry pe ->
	   get_fields_transaction acc pe.package_entry_contents
       | Class_entry te ->
	   List.fold_left
	   (fun acc (_,entry) -> match entry with
	      | Instance_variable_entry ive ->
		  if ive.instance_variable_entry_is_static
		  then acc
		  else
                    if ive.instance_variable_entry_is_ghost 
                    then acc
                    else      
		      (ive,te)::acc
	      | _ -> acc)
	   acc
	   te.class_entry_contents
       | _ -> acc)
    acc
    env
;;

let rec get_fields_transaction_2 acc env =
  List.fold_left
    (fun acc (_,entry) -> match entry with
       | Package_entry pe ->
	   get_fields_transaction_2 acc pe.package_entry_contents
       | Class_entry te ->
	   List.fold_left
	   (fun acc (_,entry) -> match entry with
              | Method_entry mel -> 
                  List.fold_left
                    (fun acc me -> 
                       if me.method_entry_routine.routine_entry_use_transactions
                       then
                         Inttagset.fold
                           (fun ive acc -> 
			      if List.mem ive acc || ive.instance_variable_entry_is_static || ive.instance_variable_entry_is_ghost 
			      then acc 
			      else ive::acc) 
                           me.method_entry_routine.routine_entry_effects.reads
                           acc
                       else acc)
                    acc
                    mel
	      | _ -> acc)
	   acc
	   te.class_entry_contents
       | _ -> acc)
    acc
    env
;;

let rec store_fields fields =
  match fields with 
  | [ive] -> 
    let f = ive.instance_variable_entry_trans_name in
      Why.LPred("eq", [Why.LVar(f ^ "_restore"); Why.LVar(f)])
  | hd::tl -> 
     Why.LAnd(store_fields [hd], store_fields tl)
  | [] -> Why.LTrue
;;


let rec restore_fields fields =
  match fields with 
  | [ive] -> 
    let f = ive.instance_variable_entry_trans_name in
      Why.LPred("eq", [Why.LVar(f); Why.LVar(f ^ "_restore")])
  | hd::tl -> 
     Why.LAnd(restore_fields [hd], restore_fields tl)
  | [] -> Why.LTrue
;;


let rec unchange_fields fields =
  match fields with 
  | [ive] -> 
    let f = ive.instance_variable_entry_trans_name in
      Why.LPred("eq", [Why.LVar(f); Why.LVar(f ^ "@")])
  | hd::tl -> 
     Why.LAnd(unchange_fields [hd], unchange_fields tl)
  | [] -> Why.LTrue
;;


let rec output_all_classes form classes =
  List.iter
    (fun (this,super) -> 
       fprintf form "| %s : classIdI@." this)
       classes
;;

(* NR 
let rec output_list_superclasses l =
  match l with
    | [] -> "nil"
    | e::l -> "(cons " ^ e ^ " " ^ (output_list_superclasses l) ^ ")" 
;;
*)

let rec output_all_superclasses form classes =
  List.iter
    (fun  (this,super) -> 
       match super with
	 | None -> fprintf form "  | %s => None@." this
	 | Some x -> fprintf form "  | %s => (Some %s)@." this x)
    classes
;;

let rec output_all_wf_classes form classes =
  List.iter
    (fun  (this,super) -> 
       match super with
	 | None ->  
	     fprintf form 
	     "Lemma wf_%s : (Acc (fun i1 i2 => (def (super i2) i1)) %s).@."
	     this this;
	     fprintf form 
	       "constructor; simpl; intros.@.";
	     fprintf form 
	       "case (undef_not_def (A:=classId) (x:=None)) with y; auto.@.";
	     fprintf form 
	       "Defined.@.";
	     fprintf form 
	       "@."
	 | Some x -> 
	     fprintf form 
	     "Lemma wf_%s : (Acc (fun i1 i2 => (def (super i2) i1)) %s).@."
	     this this;
	     fprintf form 
	       "constructor; simpl; intros.@.";
	     fprintf form 
	       "replace y with %s; auto.@." x;
	     fprintf form 
	       "apply wf_%s.@." x;
	     fprintf form 
	       "Defined.@.";
	     fprintf form 
	       "@.")
    classes
;;

let output_classes form env =
  let classes = get_classes env in
    fprintf form "(* Classes *)@.";
    fprintf form "@.";
    fprintf form "Inductive classIdI : Set :=@.";
    output_all_classes form classes;
    fprintf form ".@.";
    fprintf form "Definition classId := classIdI.@.";
    fprintf form "@.";
    fprintf form "Definition ObjectClass := Object.@.";
    fprintf form "@.";
    fprintf form "Definition eq_classId_dec : forall (i1 i2:classId), {i1=i2}+{~i1=i2}.@.";
    fprintf form "decide equality. Defined.@.";
    fprintf form "@.";
    fprintf form "Definition super (i:classId) : (option classId) :=@.";
    fprintf form "  match i with@.";
    output_all_superclasses form classes;
    fprintf form "  end.@.";
    fprintf form "@.";
    fprintf form "Lemma superObject : (super ObjectClass) = None.@.";
    fprintf form "auto. Save.@.";
    fprintf form "@.";
    output_all_wf_classes form classes;
    fprintf form "Lemma super_wf : (well_founded (fun i1 i2 => (def (super i2) i1))).@.";
    fprintf form "red; intros.@.";
    fprintf form "case a.@.";
    List.iter
      (fun  (this,_) -> fprintf form "apply wf_%s.@." this)
      classes;
    fprintf form "Defined.@.";
    fprintf form "@.";
;;

let rec fprintf_type form t =
  match t with
  | Byte_type -> fprintf form "(ByteT classId)"
  | Short_type -> fprintf form "(ShortT classId)"
  | Integer_type -> fprintf form "(IntT classId)"
  | Float_type -> fprintf form "(FloatT classId)"
  | Boolean_type -> fprintf form "(BoolT classId)"
  | Null_type -> fprintf form "(NullT classId)"
  | Class_type(te) -> fprintf form "(ClassT classId %s)" te.class_entry_name
  | Interface_type(te) -> 
      (*i fprintf form "(InterfaceT classId %s)" te.interface_entry_name i*)
      fprintf form "(ClassT classId %s)" te.interface_entry_name
  | Array_type(t) -> fprintf form "(ArrayT classId %a)" fprintf_type t
  | Type_type -> assert false
  | Prop_type | Type_var _ | Logic_type _ | Real_type -> assert false
;;

let name_field (ive,te) =
  (ive.instance_variable_entry_trans_name,
   ive.instance_variable_entry_type,
   te.class_entry_name)
;;

let output_fields form env fields = 
  let l = List.map name_field fields in
  fprintf form "(* fields *)@.";
  fprintf form "@.";
  fprintf form "Inductive fieldIdI : Set :=@.";
  List.iter
      (fun (f,t,c) ->
	 fprintf form "| %sI : fieldIdI@." f)
      l;
    fprintf form ".@.";
    fprintf form "@.";
    fprintf form "Definition fieldId := fieldIdI.@.";
    List.iter
      (fun (f,t,c) ->
	 fprintf form "Definition %s : fieldId := %sI.@." f f)
      l;
    fprintf form "@.";
    fprintf form "Definition eq_fieldId_dec : @.";
    fprintf form " forall (f1 f2:fieldId), {f1=f2}+{~f1=f2}.@.";
    fprintf form "Proof. decide equality. Defined.@.";
    fprintf form "@.";
    fprintf form "Definition defined_in : fieldId -> classId :=@.";
    fprintf form " fun (f:fieldId) =>@.";
    fprintf form " match f with@.";
    List.iter
      (fun (f,t,c) ->
	 fprintf form "| %s => %s@." f c)
      l;
    fprintf form " end.@.";
    fprintf form "@.";
    fprintf form "Lemma no_field_in_Object :@.";
    fprintf form "  forall (f:fieldId),~(defined_in f)=ObjectClass.@.";
    fprintf form "Proof. induction f; auto. Save.@.";
    fprintf form "@.";
    fprintf form "(* types of fields *)@.";
    fprintf form "@.";
    fprintf form "Definition fieldT : fieldId -> (javaT classId) :=@.";
    fprintf form " fun (f:fieldId) =>@.";
    fprintf form " match f with@.";
    List.iter
      (fun (f,t,c) ->
	 fprintf form "| %s => %a@." f fprintf_type t)
      l;
    fprintf form "end.@.";
    fprintf form "@.";
;;

 


let output_model form env =
  output_header form;
  fprintf form "Module KrakClasses.@.";
  fprintf form "@.";
  output_classes form env;
  if !Config.use_global_memory_model
  then
    begin
      let fields = get_fields [] env in
      output_fields form env fields;
    end;
  fprintf form "End KrakClasses.@.";
  fprintf form "@.";
  fprintf form "Import KrakClasses.@.";
  fprintf form "Module KrakObj := (ObjectAdr KrakClasses).@.";
  fprintf form "Module KrakGen := KrakObj.CGen.@.";
  fprintf form "@.";
(*  output_tactic form *)
;;

let instanceof_pred x =
  Why.LPred("instanceof",
	    [Why.LVar("h");
	     Why.LVar("v");
	     Why.LApp("ClassType",
			[Why.LVar(x)])])

let output_why_model env =
  let classes = get_classes env in
  List.fold_left
    (fun acc (this, super) -> 
       match super with
	 | None -> acc
	 | Some x -> 
	     Why.Axiom("instanceof_" ^ this ^"_" ^ x,
		       Why.LForall("h",([],"alloc_table"),
				   Why.LForall("v",([],"value"),
					       Why.LImpl(instanceof_pred this,
							 instanceof_pred x))))::
             (* NR - to be able to prove t1 <: t2 -> instanceof alloc v t1[] -> instanceof alloc v t2[] *)
             (Why.Axiom("subtype_" ^ this ^ "_" ^ x,
		       Why.LPred("subType", [Why.LApp("ClassType", [Why.LVar this]); Why.LApp("ClassType", [Why.LVar x])]))
	     ::acc))
    []
    classes
;;


let output_sx_model form env =
  fprintf form "@.";
  let classes = get_classes env in
  List.iter
    (fun  (this,super) -> 
       match super with
	 | None -> ()
	 | Some x ->
                fprintf form 
	          "@[(BG_PUSH@ (FORALL (h v) (IMPLIES@ (instanceof h v (ClassType %s))@ "
	          this;
	        fprintf form 
		  "(instanceof h v (ClassType %s)))))@]@.@."
	          x) 
    classes;
  fprintf form "@.";
(*  output_tactic form *)
;;

let output_rv_model form env =
  fprintf form "@.";
  let classes = get_classes env in
  List.iter
    (fun  (this,super) -> 
       match super with
	 | None -> ()
	 | Some x ->
	        fprintf form 
	          "@[(forall H V (->@ (instanceof H V (classType %s))@ "
	          this;
	        fprintf form 
	          "(instanceof H V (classType %s))))@]@.@."
	          x)
    classes;
  fprintf form "@.";
(*  output_tactic form *)
;;


let file_copy src form =
  let cin = open_in src in
  try
    while true do
      let s = input_line cin in fprintf form "%s\n" s
    done
  with End_of_file ->
    close_in cin

let output_pvs_model form env =
  fprintf form "@.";
  let classes = get_classes env in
  List.iter
    (fun (this,_) -> fprintf form "%s : classId@\n" this)
    classes;
  fprintf form "@\n";
  List.iter
    (fun  (this,super) -> 
       match super with
	 | None -> ()
	 | Some x ->
	        fprintf form 
	          "@[<hov 2>instanceof_%s_%s: AXIOM@ forall (h:store) (v:value): instanceof(h,v,ClassType(%s))@ implies instanceof(h,v,ClassType(%s))@]@.@." this x this x)
    classes;
(*  output_tactic form *)
;;



(* makefile *)

let rec print_all_paths form env =
  List.iter
    (fun (_,entry) -> match entry with
       | Package_entry pe ->
	   if not (List.mem !Config.source_package pe.package_entry_directories) then
	     () (*i fprintf form " \\@.\t-I %s" pe.package_entry_directory i*);
	   print_all_paths form pe.package_entry_contents
       | _ -> ())
    env
;;


let string_tr s ch1 ch2 =
  let t = String.copy s in
  for i=0 to pred(String.length t) do
    if String.get t i = ch1 then String.set t i ch2
  done;
  t
;;

let rec find_methods_to_prove acc env =
  List.rev(
  List.fold_left
    (fun acc (_,entry) -> match entry with
       | Package_entry pe ->
	   find_methods_to_prove acc pe.package_entry_contents
       | Class_entry te ->
	   let acc =
	     List.fold_left
	       (fun acc (_,e) ->
		  match e with
		    | Method_entry l ->
			List.fold_left
			  (fun acc me ->
			     if me.method_entry_routine.routine_entry_produce_body
			     then
				 let rec generate_m l i = 
                                   match l with
                                     | [] -> [te.class_entry_name ^ "_" ^ me.method_entry_name ^ "_nre"]
                                     | hd::tl -> 
                                       let base = te.class_entry_name ^ "_" ^ me.method_entry_name in
				       let new_m =
					 if hd = "" 
					 then base ^ "_f" ^ string_of_int i 
                                         else base ^ "_" ^ hd
				       in
				       new_m::(generate_m tl (i + 1))
				 in
				 if me.method_entry_sep_specs = []
				 then
				   (te.class_entry_name ^ "_" ^ me.method_entry_trans_name)::acc
				 else
	                           List.append (generate_m me.method_entry_sep_specs 0) acc
			     else acc)
			  acc l
		    | _ -> acc)
	       acc te.class_entry_contents
	   in
	   List.fold_left
	     (fun acc ce ->
		if ce.constructor_entry_routine.routine_entry_produce_body
		then ce.constructor_entry_trans_name::acc
		else acc)
	     acc te.class_entry_constructors
       | _ -> acc)
    acc
    env)
;;

let find_methods_to_prove env =
  find_methods_to_prove [] env
;;

let output_krak_validations form env spec_files =
  fprintf form "@.";
  fprintf form "Require Import Krak_spec_why.@.";
  let l = find_methods_to_prove env in
  List.iter
    (fun f -> 
       fprintf form "Require Import %s_valid.@." f)
    l;
  fprintf form "@.";
  fprintf form "Definition typeof := fun (A:Set) (_:A) => A.@.";
  fprintf form "@.";
  List.iter
    (fun f -> 
       fprintf form 
	 "Definition %s_krakatoa_validation : (typeof _ %s_parameter) := %s_body.@."
	 f f f)
    l;
  fprintf form "@."
;;

let string_of_prover = function
  | `COQ -> "coq"
  | `YICES -> "yices"
  | `SIMPLIFY -> "simplify"
  | `HARVEY -> "harvey"
  | `PVS -> "pvs"
;;

let why fmt f = fprintf fmt "%s.why" f
let simplify fmt f = fprintf fmt "%s_why.sx" f
let harvey fmt f = fprintf fmt "%s_why.rv" f

let print_files = Format_misc.print_list (fun fmt () -> fprintf fmt "\\@\n  ")

let output_makefile form env spec_files =
  let cmdline = ref "" in
  for 
    i=0 to Array.length Sys.argv - 1 do
      cmdline := !cmdline ^ " " ^ Sys.argv.(i);
  done; 
  fprintf form "# makefile generated by Krakatoa. Do not edit@.";
  fprintf form "@.";
  fprintf form "#@.";
  fprintf form "# This makefile was generated by the command line :@.";
  fprintf form "#%s" !cmdline;
  fprintf form "@.";
  fprintf form "#@.";
  fprintf form "@.";
  fprintf form "KRAKATOALIB=%s@." !Config.krakatoa_lib;
  fprintf form "@.";
(*  fprintf form "COQDIRS=-I $(KRAKATOALIB)%s" !Config.coqdirs; *)
  if !Config.coqdirs <> "" 
  then fprintf form "COQDIRS=-I %s" !Config.coqdirs
  else fprintf form "COQDIRS=";
  print_all_paths form env;
  fprintf form "@.";  
  fprintf form "@.";  
  fprintf form "TIMEOUT ?= 10@\n@\n";	    
  fprintf form "WHYEXEC=why";
  fprintf form "@.";  
  fprintf form "@.";  
  fprintf form "WHY=$(WHYEXEC)%s --coq-v8 %s %s %s --coq-tactic \"intuition\"@." 
    (if !Config.why_split then " --split-user-conj" else "")
    (if !Config.why_exp_goal then " --exp goal" else "") 
    (if !Config.why_fast_wp then "--fast-wp" else "")
    (if !Config.why_validation then " --valid" else ""); 
  fprintf form "@.";


  (* the main makefile rule *)
  fprintf form "all: gui@.@."; (* (string_of_prover !Config.output_language); *)

  let l = find_methods_to_prove env in

(*
  Make rules for SIMPLIFY output
*)
  fprintf form "@.";  
  fprintf form "WHYSIMPLIFY=$(WHYEXEC) %s %s %s --simplify --no-simplify-prelude@."
    (if !Config.why_split then " --split-user-conj" else "") 
    (if !Config.why_exp_goal then " --exp goal" else "") 
    (if !Config.why_fast_wp then "--fast-wp" else "");
  fprintf form "@.";

  fprintf form "simplify: %a@\n" (print_files simplify) l;
  fprintf form "\t@@echo 'Running Simplify on proof obligations' && (dp -timeout $(TIMEOUT) $^)@\n@\n";
  fprintf form "%%_why.sx: Krak_spec.why %%.why@\n";
  fprintf form "\t@@echo 'why -simplify [...] $*.why' && $(WHYSIMPLIFY) $(KRAKATOALIB)/local_memory.why Krak_spec.why $*.why@\n@\n";

(*
  Make rules for GWHY output
*)
  fprintf form "@.";  
  fprintf form "WHYGUI=g$(WHYEXEC)%s %s %s @."
    (if !Config.why_split then " --split-user-conj" else "")
    (if !Config.why_exp_goal then " --exp goal" else "") 
    (if !Config.why_fast_wp then "--fast-wp" else "");
  fprintf form "@.";

  fprintf form "gui: @\n";
  fprintf form "\t$(WHYGUI) $(KRAKATOALIB)/local_memory.why Krak_spec.why %a@\n@\n" (print_files why) l;



(*
  if !Config.output_language = `YICES
  then
*)
    begin
      fprintf form "@.";  
      fprintf form "WHYYICES=$(WHYEXEC)%s %s %s --smtlib --encoding mono@."
        (if !Config.why_split then " --split-user-conj" else "")
        (if !Config.why_exp_goal then " --exp goal" else "") 
        (if !Config.why_fast_wp then "--fast-wp" else "");
      fprintf form "@.";
    end;
  (* prove with Yices *)
(*  if !Config.output_language = `YICES then *)
    begin
      fprintf form "yices: ";
      List.iter (fun s -> fprintf form " %s_why.smt" s) l;
      fprintf form "@.";
      fprintf form "@.";
    end;



(*
  Make rules for HARVEY output
*)
    fprintf form "@.";  
    fprintf form "WHYHARVEY=$(WHYEXEC)%s --harvey --no-harvey-prelude@." 
      (if !Config.why_split then " --split-user-conj" else ""); 
    fprintf form "@.";

    fprintf form "harvey: %a@." (print_files harvey) l;
    fprintf form "\t@@echo 'Running haRVey on proof obligations' && (dp -timeout $(TIMEOUT) $^)@\n@\n";
    fprintf form "%%_why.rv: Krak_spec.why %%.why@\n";
    fprintf form "\t@@echo 'why -harvey [...] $*.why' && $(WHYHARVEY) $(KRAKATOALIB)/local_memory.why Krak_spec.why $*.why@\n@\n";


(*
  if !Config.output_language = `PVS
  then
*)
    begin
      fprintf form "@.";  
      fprintf form "WHYPVS=$(WHYEXEC) --pvs --pvs-preamble \"importing Krak_spec_why\"@." ;
      fprintf form "@.";
    end;


  (* prove with Coq *)
(*  if !Config.output_language = `COQ then *)
    begin
      fprintf form "coq:";
      List.iter (fun s -> fprintf form " %s_why.vo" s) l;
      if !Config.why_validation then
	begin
	  List.iter (fun s -> fprintf form " %s_valid.vo" s) l;
	  fprintf form " Krak_validations.vo"
	end;
      fprintf form "@.";
      fprintf form "@.";
    end;

(*  if !Config.output_language = `PVS then *)
    begin
      fprintf form "pvs:";
      List.iter (fun s -> fprintf form " %s_why.pvs" s) l;
      fprintf form "@.";
      fprintf form "\techo '(typecheck \"Krak_spec_why\")' > pvsbatch.el@.";
      List.iter 
	(fun s -> 
	   fprintf form "\techo '(typecheck \"%s_why\")' >> pvsbatch.el@." s)
	l;
      fprintf form "\tpvs -q -v 3 -batch -l pvsbatch.el@.";
      fprintf form "@.";
    end;

  (* rule for building memory_model.why and memory_model_why.v *)
  let source_model =
    if !Config.use_global_memory_model
    then "global" else "local"
  in
  fprintf form "%s_memory.why: $(KRAKATOALIB)/%s_memory.why@." 
    source_model source_model;
  fprintf form "\tcp $(KRAKATOALIB)/%s_memory.why %s_memory.why@.@." 
    source_model source_model;
  fprintf form "%s_memory_why.v: $(KRAKATOALIB)/%s_memory_why.v@." 
    source_model source_model;
  fprintf form "\tcp $(KRAKATOALIB)/%s_memory_why.v %s_memory_why.v@.@." 
    source_model source_model;
  fprintf form "krak_tactics.v: $(KRAKATOALIB)/krak_tactics.v @.";
  fprintf form "\tcp $(KRAKATOALIB)/krak_tactics.v krak_tactics.v@.@.";
  if !Config.why_validation then 
    begin
      fprintf form "%s_memory_valid.v: $(KRAKATOALIB)/%s_memory_valid.v@." 
	source_model source_model;
      fprintf form "\tcp $(KRAKATOALIB)/%s_memory_valid.v %s_memory_valid.v@.@." 
	source_model source_model;
    end;


  (* rule for executing why on files generated by krakatoa *)
  fprintf form "Krak_spec_why.v";
  if !Config.why_validation then fprintf form " Krak_spec_valid.v";
  fprintf form " \\@.";
  List.iter
    (fun s -> fprintf form "%s_why.v \\@." s)
    l;
  fprintf form "\t: %s_memory.why" source_model;
  fprintf form " \\@.\t\t%s_memory_why.v" source_model;
  fprintf form " \\@.\t\tkrak_tactics.v";
  fprintf form " \\@.\t\tKrak_spec.why";
  List.iter
    (fun s -> fprintf form " \\@.\t\t%s.why" s)
    l;
  fprintf form "@.\t@@echo Running why on generated programs...@.";
(*  let add_load_path = "Add LoadPath \\\"" ^ !Config.krakatoa_lib ^ "\\\". "
  in *)
  fprintf form 
    "@.\t$(WHY) --coq-preamble \"Require Export krak_tactics.\" %s_memory.why" source_model;
  fprintf form " \\@.\t\tKrak_spec.why";
  fprintf form 
    "@.\t$(WHY) --coq-preamble \"Require Export Krak_spec_why.\" %s_memory.why" source_model;
  fprintf form " \\@.\t\tKrak_spec.why";
  List.iter
    (fun s -> fprintf form " \\@.\t\t%s.why" s)
    l;
  fprintf form "@.@.";

  (* running why with yices output *)
(*
  if !Config.output_language = `YICES
  then
*)
    begin
      fprintf form "Krak_spec_why.smt \\@.";
      List.iter
	(fun s -> fprintf form "%s_why.smt \\@." s)
	l;
      fprintf form "\t: %s_memory.why" source_model;
      fprintf form " \\@.\t\tKrak_spec.why";
      List.iter
	(fun s -> fprintf form " \\@.\t\t%s.why" s)
	l;
      fprintf form "@.\t@@echo Running why on generated programs...@.";
      fprintf form "@.\t$(WHYYICES) %s_memory.why" source_model;
      fprintf form " \\@.\t\tKrak_spec.why";
      List.iter
	(fun s -> fprintf form " \\@.\t\t%s.why" s)
	l;
      fprintf form "@.@.";
    end;

(*
  if !Config.output_language = `HARVEY
  then
*)
    begin
      fprintf form "Krak_spec_why.rv \\@.";
      List.iter
	(fun s -> fprintf form "%s_why.rv \\@." s)
	l;
      fprintf form "\t: %s_memory.why" source_model;
      fprintf form " \\@.\t\tKrak_spec.why";
      List.iter
	(fun s -> fprintf form " \\@.\t\t%s.why" s)
	l;
      fprintf form "@.\t@@echo Running why on generated programs...@.";
      fprintf form "@.\t$(WHYHARVEY) %s_memory.why" source_model;
      fprintf form " \\@.\t\tKrak_spec.why";
      List.iter
	(fun s -> fprintf form " \\@.\t\t%s.why" s)
	l;
      fprintf form "@.@.";
    end;

(*
  if !Config.output_language = `PVS
  then
*)
    begin
      fprintf form "Krak_spec_why.pvs \\@.";
      List.iter
	(fun s -> fprintf form "%s_why.pvs \\@." s)
	l;
      fprintf form "\t: %s_memory.why" source_model;
      fprintf form " \\@.\t\tKrak_spec.why";
      List.iter
	(fun s -> fprintf form " \\@.\t\t%s.why" s)
	l;
      fprintf form "@.\t@@echo Running why on generated programs...@.";
      fprintf form "@.\t$(WHYPVS) %s_memory.why" source_model;
      fprintf form " \\@.\t\tKrak_spec.why";
      List.iter
	(fun s -> fprintf form " \\@.\t\t%s.why" s)
	l;
      fprintf form "@.@.";
    end;

  (* dependency of the coq object on the importations files *)
(*
  List.iter
    (fun f -> 
       let f = Filename.basename f in 
       fprintf form "%s_param.vo: %s_model_imports.vo@." f source_model)
    spec_files;
*)

  List.iter
    (fun s -> fprintf form "%s_why.vo: local_memory_why.vo krak_tactics.vo Krak_spec_why.vo@." s)
    l;
  fprintf form "@.";


  (* dependency of the krak validations *)
  if !Config.why_validation then 
    begin
      fprintf form "Krak_validations.vo:";
      List.iter
	(fun s -> fprintf form " %s_valid.vo" s)
	l;
      fprintf form "@.@.";
    end;

  fprintf form "# generic rules@.";
  fprintf form "@.";
  fprintf form ".SUFFIXES: .v .vo@.";
  fprintf form "@.";
  fprintf form ".v.vo: @.";
  fprintf form "\t@@echo coqc [...] $< && coqc $(COQDIRS)%s $<@." !Config.coqopts;
  fprintf form "@.";


(*
  if !Config.output_language = `HARVEY
  then
*)
    begin
      fprintf form ".SUFFIXES: .rv .rvcheck@.";
      fprintf form "@.";
      fprintf form ".rv.rvcheck: @.";
      fprintf form "\techo '(' > $*.all.rv@." ;      
      fprintf form "\tcat $(KRAKATOALIB)/local_memory_why.rv Krak_spec_why.rv \\@.";
      fprintf form "\t\t >> $*.all.rv@." ;
      fprintf form "\techo ')' >> $*.all.rv@." ;      
      fprintf form "\trvc $*.all.rv@." ;
      fprintf form "@.";
    end;

(*
  if !Config.output_language = `PVS
  then
*)
    begin
      fprintf form ".SUFFIXES: .pvs@.";
      fprintf form "@.";
    end;

  fprintf form "# regenerate@.";
  fprintf form "@.";
  fprintf form ".PHONY: regenerate@.";
  fprintf form "@.";
  fprintf form "regenerate:@.";
  fprintf form "\tcd .. && %s@." !cmdline;
  fprintf form "@.";
  
  fprintf form "# clean and depend@.";
  fprintf form "@.";
  fprintf form "clean:@.";
  fprintf form "\trm -f *.vo *.bak Krak_spec_why.v *_valid.v %s_memory.why %s_memory_why.v@." 
    source_model source_model;
  fprintf form "@.";
  if !Config.output_language = `COQ then begin
  fprintf form "depend .depend: %s_memory_why.v" source_model;
  fprintf form " \\@.\tKrak_spec_why.v";
  List.iter
    (fun s -> fprintf form " \\@.\t\t%s_why.v" s)
    l;
  fprintf form "@.";
  if Version.has_coq then
    begin
      fprintf form "\trm -f .depend@.";
      fprintf form "\tcoqdep -I %s/user-contrib $(COQDIRS) *.v >> .depend@."
	Version.coqlib;
    end
  else 
    fprintf form "\ttouch .depend@.";
  fprintf form "@.";
  fprintf form "include .depend@.";
  fprintf form "@."
  end
;;


(* JAVACARD *)
(* TO DO *)
(*
let fields = get_fields_transaction_2 [] genv;;
let field_names = List.map (fun ive -> ive.Java_context_types.instance_variable_entry_trans_name) fields;;
let field_names_restore = List.map (fun ive -> ive.Java_context_types.instance_variable_entry_trans_name ^ "_restore") fields;;
*)

(* TRANSACTIONS *)

(* beginTransaction() *)
(*
let beginTransaction_pre = Why.LPred("eq_int", [Why.LVar("transactionDepth"); Why.LConst(Why.Prim_int(0))]);;
let beginTransaction_reads = List.append ["intA"; "boolA"; "objA"] field_names;;
let beginTransaction_writes = List.append ["transactionDepth"; "intARestore"; "boolARestore"; "objARestore"] field_names_restore;;
let beginTransaction_arrays_body = Why.LAnd(Why.LPred("eq_int", [Why.LApp("array_acc", [Why.LVar("intARestore"); Why.LVar("v"); Why.LVar("i")]); 
                                                                     Why.LApp("array_acc", [Why.LVar("intA"); Why.LVar("v"); Why.LVar("i")])]), 
                                                Why.LAnd(Why.LPred("eq_bool", [Why.LApp("array_acc", [Why.LVar("boolARestore"); Why.LVar("v"); Why.LVar("i")]);
                                                                               Why.LApp("array_acc", [Why.LVar("boolA"); Why.LVar("v"); Why.LVar("i")])]), 
                                                         Why.LPred("eq", [Why.LApp("array_acc", [Why.LVar("objARestore"); Why.LVar("v"); Why.LVar("i")]); 
                                                                                Why.LApp("array_acc", [Why.LVar("objA"); Why.LVar("v"); Why.LVar("i")])])));;
let beginTransaction_arrays = Why.LForall("v", "value", Why.LForall("i", "int", beginTransaction_arrays_body));;
let beginTransaction_fields = Model.store_fields fields;;
let beginTransaction_post = Why.LAnd(Why.LPred("eq_int", [Why.LVar("transactionDepth"); Why.LConst(Why.Prim_int(1))]), 
                                         Why.LAnd(beginTransaction_arrays, beginTransaction_fields));;
let beginTransaction_annot =  Why.Annot_type(beginTransaction_pre, Why.unit_type, beginTransaction_reads, beginTransaction_writes, beginTransaction_post, None, None);;
let beginTransaction_type = Why.Prod_type("tt", Why.unit_type, beginTransaction_annot);;
let beginTransaction_decl = Why.Param("JCSystem_beginTransaction_parameter", beginTransaction_type);;
*)

(* CARD TEAR *)

(* NON-ATOMIC METHODS *)

