
open Java_context_types;;
open Java_context;;
open Ast_types;;
open Ast;;

exception Ill_typed of string;;

(*
let ill_typed msg = raise (Ill_typed msg);;

let translation_anomaly loc msg =
  Format.printf "Translation anomaly: %s%s@."
    (Location.to_string loc) msg;
  assert false
*)

let ill_typed msg = Typing.ill_typed Location.dummy msg;;

let translation_anomaly loc msg =
  if !Config.stop_on_error
  then
    failwith ("Translation anomaly: "^(Location.to_string loc)^msg)
  else
    Format.printf "Translation anomaly: %s%s@."
      (Location.to_string loc) msg


exception Unsupported of string;;

let unsupported loc msg =
  raise (Unsupported ((Location.to_string loc)^": "^msg))

let unsupported_no_loc msg = unsupported Location.dummy msg

let ill_typed msg = Typing.ill_typed Location.dummy msg;;

let get_logic_types env =
(*  Format.printf "get_logic_types@."; *)
  List.fold_left
    (fun acc (_, entry) -> match entry with
       | Logic_type_entry lte -> 
           (* Format.printf "lte: %s@." lte.logic_type_entry_name; *)
           lte::acc
       | _ -> acc)
    []
    env
;;

let rec get_fields acc env =
  List.fold_left
    (fun acc (_,entry) -> match entry with
       | Package_entry pe ->
	   get_fields acc pe.package_entry_contents
       | Class_entry te ->
	   List.fold_left
	   (fun acc (_,entry) -> match entry with
	      | Instance_variable_entry ive ->
		  (* if ive.instance_variable_entry_is_static
		  then acc
		  else*) ive::acc
	      | _ -> acc)
	   acc
	   te.class_entry_contents
       | _ -> acc)
    acc
    env
;;

let rec get_constructors acc env =
  List.fold_left
    (fun acc (_, entry) -> match entry with
       | Package_entry pe ->
	   get_constructors acc pe.package_entry_contents
       | Class_entry ce ->
	   List.append ce.class_entry_constructors acc
       | _ -> acc)
    acc
    env
;;

let rec get_methods acc env =
  List.fold_left
    (fun acc (_, entry) -> match entry with
       | Package_entry pe ->
	   get_methods acc pe.package_entry_contents
       | Class_entry ce ->
	   List.fold_left
	   (fun acc (_, entry) -> match entry with
	      | Method_entry mel ->
		  List.append mel acc
	      | _ -> acc)
	   acc
	   ce.class_entry_contents
       | Interface_entry ie ->
           List.fold_left
           (fun acc (_, entry) -> 
              match entry with
                | Method_entry mel -> List.append mel acc
                | _ -> acc)
           acc
           ie.interface_entry_contents
       | _ -> acc)
    acc
    env
;;

let rec check_logic_type_ambiguity l =
  match l with
    | [] -> ()
    | [_] -> ()
    | lte1::(lte2::_ as r) ->
	if lte1.logic_type_entry_name = lte2.logic_type_entry_name
	then
	  begin
	    lte1.logic_type_entry_name <- 
 	    lte1.logic_type_entry_name ^ "_1";
	    lte2.logic_type_entry_name <-
	    lte2.logic_type_entry_name ^ "_2";
	  end;
	check_logic_type_ambiguity r
;;

let rec check_ambiguity l =
  match l with
    | [] -> ()
    | [_] -> ()
    | ive1::(ive2::_ as r) ->
	if ive1.instance_variable_entry_name = ive2.instance_variable_entry_name
	then
	  begin
	    ive1.instance_variable_entry_trans_name <- 
 	    get_ive_class_or_interface_name ive1 ^ "_" ^ 
	    ive1.instance_variable_entry_name;
	    ive2.instance_variable_entry_trans_name <-
 	    get_ive_class_or_interface_name ive2 ^ "_" ^ 
	    ive2.instance_variable_entry_name;
	  end;
        (* reserved words in Why *)
        if Why.is_why_reserved ive1.instance_variable_entry_name
        then
          begin
          ive1.instance_variable_entry_trans_name <- 
 	    get_ive_class_or_interface_name ive1 ^ "_" ^ 
	    ive1.instance_variable_entry_name
          end;
	check_ambiguity r
;;

let rec java_type_to_string java_type =
  match java_type with 
    | Byte_type -> "byte"
    | Short_type -> "short"
    | Integer_type -> "int"
    | Float_type -> "float"
    | Boolean_type -> "bool"
    | Class_type(ce) -> ce.class_entry_name
    | Interface_type(ie) -> ie.interface_entry_name 
    | Array_type(jt) -> java_type_to_string jt ^ "T"
    | _ -> assert false
;;

let get_class_or_interface_name me =
  match me.method_entry_class_or_interface with
    | Class_entry ce -> ce.class_entry_name
    | Interface_entry ie -> ie.interface_entry_name
    | _ -> assert false (* should not occur *)
;;

let set_extended_constructor_name ce =
  if List.length ce.constructor_entry_routine.routine_entry_parameters > 0
  then
    begin
      ce.constructor_entry_trans_name <- 
      ce.constructor_entry_trans_name ^
      List.fold_left
        (fun s (_, entry) ->
           match entry with
	    | Local_variable_entry lve ->
                s ^ "_" ^ java_type_to_string lve.local_variable_entry_type
            | _ -> assert false (* should not occur *))
        ""
        ce.constructor_entry_routine.routine_entry_parameters;
    end;
;;

let set_extended_method_name me =
  if List.length me.method_entry_routine.routine_entry_parameters > 0
  then
    begin
      me.method_entry_trans_name <- 
      me.method_entry_name ^
      List.fold_left
        (fun s (_, entry) ->
           match entry with
	    | Local_variable_entry lve ->
                s ^ "_" ^ java_type_to_string lve.local_variable_entry_type
            | _ -> assert false (* should not occur *))
        ""
        me.method_entry_routine.routine_entry_parameters;
    end;
;;

let rec check_constructor_ambiguity l =
  match l with
    | [] -> ()
    | [_] -> ()
    | ce1::(ce2::_ as r) ->
	if ce1.constructor_entry_trans_name = ce2.constructor_entry_trans_name
	then
          begin
            set_extended_constructor_name ce1;
            set_extended_constructor_name ce2
          end;
	check_constructor_ambiguity r
;;

let rec check_method_ambiguity l =
  match l with
    | [] -> ()
    | [_] -> ()
    | me1::(me2::_ as r) ->
	if get_class_or_interface_name me1 = get_class_or_interface_name me2 &&
           me1.method_entry_name = me2.method_entry_name
	then
          begin
            set_extended_method_name me1;
            set_extended_method_name me2
          end;
	check_method_ambiguity r
;;

let rec check_lve_ambiguity fix l =
  begin
  match l with
    | [] -> ()
    | [_] -> ()
    | (_, Local_variable_entry lve1)::((_, Local_variable_entry lve2)::_ as r) ->
	if lve1.local_variable_entry_trans_name = lve2.local_variable_entry_trans_name
	then
          begin
	    lve1.local_variable_entry_trans_name <-
              lve1.local_variable_entry_trans_name ^ "_0";
	    lve2.local_variable_entry_trans_name <-
              lve2.local_variable_entry_trans_name ^ "_1";
	    check_lve_ambiguity false r
          end
        else
	  check_lve_ambiguity true r;
    | _ -> assert false (* should never happen *)
  end;
  if not fix then check_lve_ambiguity true l
;;

let rec check_constructor_parameters_problems l f =
  List.iter
  (fun ce ->
     let re = ce.constructor_entry_routine in
     let parameters = re.routine_entry_parameters in
     List.iter
     (fun (_, p) -> 
        match p with
          | Local_variable_entry lve ->
              if Why.is_why_reserved lve.local_variable_entry_name 
              then lve.local_variable_entry_trans_name <- "p_" ^ lve.local_variable_entry_name;
              List.iter
              (fun ive -> 
                 if ive.instance_variable_entry_trans_name = lve.local_variable_entry_trans_name
                 then 
                   ive.instance_variable_entry_trans_name <- 
                     get_ive_class_or_interface_name ive ^ "_" ^ ive.instance_variable_entry_trans_name;)
              f
          | _ -> ())
     parameters;
     check_lve_ambiguity true parameters)
  l  
;;

let rec check_method_parameters_problems l f =
  List.iter
  (fun me ->
     let re = me.method_entry_routine in
     let parameters = re.routine_entry_parameters in
     List.iter
     (fun (_, p) -> 
        match p with
          | Local_variable_entry lve ->
              if Why.is_why_reserved lve.local_variable_entry_name 
              then lve.local_variable_entry_trans_name <- "p_" ^ lve.local_variable_entry_name;
              List.iter
              (fun ive -> 
                 if ive.instance_variable_entry_trans_name = lve.local_variable_entry_trans_name
                 then 
                   ive.instance_variable_entry_trans_name <- 
                     get_ive_class_or_interface_name ive ^ "_" ^ ive.instance_variable_entry_trans_name;)
              f
          | _ -> ())
     parameters;
     check_lve_ambiguity true parameters)
  l
;;

let rec check_method_local_vars_problems l f =
  List.iter
  (fun me ->
     let re = me.method_entry_routine in
     let local_vars = re.routine_entry_local_vars in
     List.iter
       (fun (_, jee) ->
	 match jee with
	 | Local_variable_entry lve ->
	     if Why.is_why_reserved lve.local_variable_entry_name 
	     then lve.local_variable_entry_trans_name <- "local_" ^ lve.local_variable_entry_name;
	     List.iter
               (fun ive -> 
		 if ive.instance_variable_entry_trans_name = lve.local_variable_entry_trans_name
		 then 
		   ive.instance_variable_entry_trans_name <- 
		     get_ive_class_or_interface_name ive ^ "_" ^ ive.instance_variable_entry_trans_name)
	       f
	 | _ -> ())
       local_vars;
     check_lve_ambiguity true local_vars)
  l
;;

let logic_type_names_disambiguation env =
  (* Format.printf "logic_type_names_disambiguation@."; *)
  let logic_types = 
    List.sort
      (fun lte1 lte2 -> 
         String.compare 
           lte1.logic_type_entry_name 
	   lte2.logic_type_entry_name)
      (get_logic_types env) 
  in
  List.iter 
    (fun lte -> 
       if Why.is_why_reserved lte.logic_type_entry_name 
       then lte.logic_type_entry_name <- "lt" ^ lte.logic_type_entry_name) 
    logic_types;
  check_logic_type_ambiguity logic_types;
;;

let field_names_disambiguation env =
  let sfields = 
    List.sort 
      (fun ive1 ive2 -> 
	 String.compare 
	 ive1.instance_variable_entry_name 
	 ive2.instance_variable_entry_name)
      (get_fields [] env)
  in
  check_ambiguity sfields;
;;

let constructor_disambiguation env =
  let sconstructors =
    List.sort
      (fun ce1 ce2 ->
	 String.compare
	 ce1.constructor_entry_trans_name
	 ce2.constructor_entry_trans_name)
      (get_constructors [] env)
  in
  let sfields = 
    List.sort 
      (fun ive1 ive2 -> 
	 String.compare 
	 ive1.instance_variable_entry_name 
	 ive2.instance_variable_entry_name)
      (get_fields [] env)
  in
  check_constructor_ambiguity sconstructors;
  check_constructor_parameters_problems sconstructors sfields;
;;

let method_disambiguation env =
  let smethods =
    List.sort
      (fun me1 me2 ->
	 String.compare
	 me1.method_entry_name
	 me2.method_entry_name)
      (get_methods [] env)
  in
  let sfields = 
    List.sort 
      (fun ive1 ive2 -> 
	 String.compare 
	 ive1.instance_variable_entry_name 
	 ive2.instance_variable_entry_name)
      (get_fields [] env)
  in
  check_method_ambiguity smethods;
  check_method_parameters_problems smethods sfields;
  check_method_local_vars_problems smethods sfields;
;;


let rec compute_inv_args env =
  List.iter
    (fun (_,entry) -> match entry with
       | Package_entry pe ->
	   compute_inv_args pe.package_entry_contents
       | Class_entry ce ->
	   let eff = ce.class_entry_invariant_effects in
(*
	   let eff =
	     List.fold_left
	       (fun acc (_,entry) -> match entry with
		  | Instance_variable_entry ive ->
		      if ive.instance_variable_entry_is_static
		      then acc
		      else 
			begin
			  match ive.instance_variable_entry_type with
			    | Integer_type | Boolean_type | Float_type ->
				acc
			    | _ -> 
				{ eff with reads = Inttagset.add ive.instance_variable_entry_tag ive acc.reads }
			end
		  | _ -> acc)
	       eff
	       ce.class_entry_contents
	   in
*)
	   let inv_params = 
	     Effects.params_reads eff [] false
	   in
	   let static_inv_params = 
	     Effects.params_reads eff [] true 
	   in
           let restore_inv_params =
	     Effects.restore_params_reads eff [] 
	   in    
             ce.class_entry_invariant_args <- inv_params;
             ce.class_entry_static_invariant_args <- static_inv_params;
             ce.class_entry_restore_invariant_args <- restore_inv_params
       | _ -> ())
    env
;;



(* check if method can be translated to a predicate, ie returns
   a boolean and its postcondition has the form \result <==> ... *)
let find_predicate post = 
  match post.expr_node with
    | Bin(v,"<==>",e) ->
	begin
	  match v.expr_node with
	    | Typed_var(lve) 
		when lve.local_variable_entry_name = "result" -> e
	    | _ -> raise Not_found
	end
    | _ -> raise Not_found
;;

let find_predicate_method me jml =
  if me.method_entry_is_pure
  then 
    let jml_post = 
      match jml with
	| Typed_compact tcc -> tcc
	| _ -> ill_typed "find_predicates_method(jml desugaring)"
    in 
    try
      if (jml_post.typed_compact_signals = None)
(*i
	&& (jml_post.typed_compact_requires.expr_node = Lit(Bool(true)))
i*)
      then
	begin
	  let _ = find_predicate jml_post.typed_compact_ensures in
	  Format.fprintf Config.log 
	    "Boolean method %s is considered as predicate.@." 
	    me.method_entry_name;
	  me.method_entry_is_predicate <- true;
	end
    with Not_found -> ()
;;

let find_predicates_class_field (_,f) = 
  match f with 
    | Typed_method(_, Typed_compact_nested _, _) -> ()
    | Typed_method(me,jml,b) -> 
	find_predicate_method me jml
    | Typed_constructor(ce,jml,eci,b) -> ()
    | Typed_variable(ve,init) -> ()
    | Static_initializer _ -> ()
    | Method(_,_) -> 
	ill_typed "tr_field_declaration(method)"
    | Constructor(_) -> 
	ill_typed "tr_field_declaration(constructor)"
    | Variable(_) -> 
	ill_typed "tr_field_declaration(variable)"
    | Specification(_) -> ()

let find_predicates_interface_member (_,m) =
  match m with
    | Constant_declaration(_) -> 
	ill_typed "find_predicates_interface_member(constant)"
    | Typed_model_variable _
    | Interface_specification _
    | Typed_method_header(_, Typed_compact_nested _)
    | Typed_constant_declaration _ -> ()
    | Method_header(_) -> 
	ill_typed "find_predicates_interface_member(method)"
    | Typed_method_header(me,jml) -> 
	find_predicate_method me jml
    | Model_variable _ ->
	ill_typed "find_predicates_interface_member(variable)"

let find_predicates_type (_,td) =
  match td with
  | Interface _ -> 	
      ill_typed "find_predicates_type(interface)"
  | Class _ -> 	
      ill_typed "find_predicates_type(class)"
  | Axiom _ -> ()
  | Logic_typedecl _ -> ()
  | Logic_reads _ -> ()
  | Logic_def _ -> ()
  | Typed_logic_reads _ -> ()
  | Typed_logic_def _ -> ()
  | Typed_interface(_, i) -> 
      List.iter find_predicates_interface_member i.interface_members
  | Typed_class(_,c) -> 
      List.iter find_predicates_class_field c.class_fields
  

let find_predicates comp_unit =
  List.iter find_predicates_type comp_unit.cu_type_decls


let rec canonize_for_simplify env =
  List.iter
    (fun (_,entry) -> match entry with
       | Package_entry pe ->
	   canonize_for_simplify pe.package_entry_contents
       | Class_entry te ->
	   List.iter
	     (fun (_,entry) -> match entry with
		| Instance_variable_entry ive ->
		    let id = ive.instance_variable_entry_trans_name in
		    if String.get id 0 = '_'
		    then 
		      ive.instance_variable_entry_trans_name <- "U" ^ id
		| _ -> ())
	     te.class_entry_contents
       | _ -> ())
    env
;;

open Format

let pre_translation env comp_units =
  begin
    fprintf Config.log "Starting computation of effects.@.";
    Effects.compute env comp_units;
    fprintf Config.log "Starting names disambiguation.@.";
    logic_type_names_disambiguation env;
    field_names_disambiguation env;
    constructor_disambiguation env;
    method_disambiguation env;
    if !Config.output_language = `SIMPLIFY
    then canonize_for_simplify env;
    List.iter find_predicates comp_units;
    fprintf Config.log "Computing class invariants.@.";
    compute_inv_args env;
  end
;;

