
open Java_context_types;;
open Java_context;;
open Ast_types;;
open Ast;;
open Pre_translate;;


let sprintf = Format.sprintf

let tr_ive ive = ive.instance_variable_entry_trans_name

let type_default_value produce_a_value t =
  match t with
    | Byte_type
    | Short_type
    | Integer_type -> 
	let v = Why.Cte(Why.Prim_int("0")) in
	if produce_a_value then Why.make_app "Primitive_int" [v] else v
    | Float_type -> 
	let v = Why.Cte(Why.Prim_real("0.0")) in
	if produce_a_value then Why.make_app "Primitive_float" [v] else v
    | Boolean_type -> 
	let v = Why.Cte(Why.Prim_bool(false)) in
	if produce_a_value then Why.make_app "Primitive_bool" [v] else v
    | Class_type(_) -> Why.Var("Null")
    | Array_type _ ->  Why.Var("Null")
    | Interface_type _ -> Why.Var("Null")
    | Null_type | Type_type | Real_type -> assert false
    | Prop_type | Type_var _ | Logic_type _ -> assert false
;;
let has_head_term s = function 
    Why.LApp(s',_) -> s=s'
  | _ -> false 

let has_head_expr s = function 
    Why.App(Why.Var(s'),_) -> s=s'
  | _ -> false 

let arg_term = function 
    Why.LApp(_,[x]) -> x
  | _ -> assert false 

let arg_expr = function 
    Why.App(_,x) -> x
  | _ -> assert false 

let ref_modifiable = ref Why.LTrue


let coerce_to_value_term t accu = match t with 
      | Byte_type
      | Short_type
      | Integer_type -> 
	  if has_head_term "int_val" accu then arg_term accu 
	  else Why.LApp("Primitive_int",[accu])
      | Boolean_type -> 
	  if has_head_term "bool_val" accu then arg_term accu 
	  else Why.LApp("Primitive_bool",[accu])
      | Float_type -> 
	  if has_head_term "float_val" accu then arg_term accu 
	  else Why.LApp("Primitive_float",[accu])
      | Null_type -> ill_typed "coerce_to_value_term(null)"
      | Type_type -> ill_typed "coerce_to_value_term(type)"
      | Real_type -> ill_typed "coerce_to_value_term(real)"
      | Prop_type | Type_var _ | Logic_type _ -> assert false
      | Class_type(_) 
      | Array_type(_)
      | Interface_type(_) ->  accu
	
let coerce_to_value_expr t accu = match t with 
      | Byte_type
      | Short_type
      | Integer_type -> 
	  if has_head_expr "int_val" accu then arg_expr accu 
	  else Why.make_app "Primitive_int" [accu]
      | Boolean_type -> 
	  if has_head_expr "bool_val" accu then arg_expr accu 
	  else Why.make_app "Primitive_bool" [accu]
      | Float_type -> 
	  if has_head_expr "float_val" accu then arg_expr accu 
	  else Why.make_app "Primitive_float" [accu]
      | Null_type -> ill_typed "coerce_to_value_expr(null)"
      | Type_type -> ill_typed "coerce_to_value_term(type)"
      | Real_type -> ill_typed "coerce_to_value_term(real)"
      | Class_type(_) 
      | Array_type(_)
      | Interface_type(_) ->  accu
      | Prop_type | Type_var _ | Logic_type _ -> assert false

	
(* translation of types : to primitive type or to "Value" *)

let rec tr_type t =
  let t' =
    match t with
      | Byte_type | Short_type | Integer_type -> "int"
      | Float_type | Real_type -> "real"
      | Boolean_type -> "bool"
      | Type_var v -> "'" ^ v
      | Logic_type(s, tl) -> 
	  List.fold_right
	    (fun t acc ->
	      (snd (tr_type t)) ^ " " ^ acc)
	    tl
	    s
      | Prop_type -> "prop"
      | Class_type(ce) -> 
	  "value" (* sprintf "value (* %s *)" ce.class_entry_name *)
      | Array_type t -> 
	  "value"
	    (*Format.fprintf 
	      Format.str_formatter
	      "value (* %a array *)" Java_context.fprintf_java_type t;
	      Format.flush_str_formatter () *)
      | Interface_type ie -> 	
	  "value" (* sprintf "value (* %s *)" ie.interface_entry_name *)
      | Null_type -> "unit"
      | Type_type -> unsupported_no_loc "tr_type(type_type)"
  in ([],t')
;;


let rec tr_type_detailed t =
  match t with
  | Byte_type | Short_type | Integer_type -> "int"
  | Float_type | Real_type -> "real"
  | Boolean_type -> "bool"
  | Type_var v -> "'" ^ v
  | Logic_type(s, tl) -> 
      List.fold_right
	(fun t acc ->
	  tr_type_detailed t ^ " " ^ acc)
	tl
	s
  | Prop_type -> "prop"
  | Class_type(ce) -> ce.class_entry_name
  | Array_type _ -> unsupported_no_loc "tr_type_detailed(array_type)"
  | Interface_type _ -> unsupported_no_loc "tr_type_detailed(interface_type)"
  | Null_type -> "unit"
  | Type_type -> unsupported_no_loc "tr_type_detailed(type_type)"
;;
	
	
let rec tr_type_term t =
  match t with
  | Byte_type | Short_type | Integer_type -> Why.LVar("IntType")
  | Float_type -> Why.LVar("FloatType")
  | Boolean_type -> Why.LVar("BoolType")
  | Type_var v -> Why.LVar("'" ^ v)
  | Logic_type(s, tl) -> Why.LVar(s) (* TO DO - NR *)
  | Prop_type -> assert false
  | Class_type(ce) -> 
      Why.LApp ("ClassType", [ Why.LVar ce.class_entry_name ])
  | Array_type Byte_type | Array_type Short_type | Array_type Integer_type -> Why.LVar("ArrIntType")
  | Array_type Float_type -> Why.LVar("ArrFloatType")
  | Array_type Boolean_type -> Why.LVar("ArrBoolType")
  | Array_type t1 -> 
      Why.LApp
	("ArrayType",
	 [tr_type_term t1])
  | Interface_type ce -> 
      Why.LApp ("ClassType", [ Why.LVar ce.interface_entry_name ])
  | Null_type -> 
      unsupported_no_loc "tr_type_in_coq(null_type)"
  | Type_type ->
      unsupported_no_loc "tr_type_in_coq(type_type)"
  | Real_type ->
      unsupported_no_loc "tr_type_in_coq(real_type)"
	
let rec tr_type_expr t =
  match t with
  | Byte_type | Short_type | Integer_type -> Why.Var("IntType")
  | Float_type -> Why.Var("FloatType")
  | Boolean_type -> Why.Var("BoolType")
  | Type_var _ | Logic_type _ | Prop_type -> assert false
  | Class_type(ce) -> 
      Why.make_app "ClassType" [Why.Var ce.class_entry_name]
  | Array_type Byte_type | Array_type Short_type | Array_type Integer_type -> Why.Var("ArrIntType")
  | Array_type Float_type -> Why.Var("ArrFloatType")
  | Array_type Boolean_type -> Why.Var("ArrBoolType")
  | Array_type t1 -> 
      Why.make_app "ArrayType" [tr_type_expr t1]
  | Interface_type ie -> 
      Why.make_app "ClassType" [Why.Var ie.interface_entry_name] (* NR : temporary... to test *)
	(* unsupported_no_loc "tr_type_expr(interface_type)" *)
  | Null_type -> 
      unsupported_no_loc "tr_type_expr(null_type)"
  | Type_type -> 
      unsupported_no_loc "tr_type_expr(type_type)"
  | Real_type -> 
      unsupported_no_loc "tr_type_expr(real_type)"

let tr_literal loc l =
  match l with
    | Int(n) -> Why.Cte(Why.Prim_int(string_of_int n))
    | Long(n) -> Why.Cte(Why.Prim_int(Int64.to_string n))
    | Float(s) -> Why.Cte(Why.Prim_real(s))
    | Bool(b) -> Why.Cte(Why.Prim_bool(b))
    | String(s) -> unsupported loc "tr_literal(string)"
    | Char(c) -> Why.Cte(Why.Prim_int(string_of_int (int_of_char c)))
    | Null -> Why.Var("Null")
;;



let why_lzero = Why.LConst(Why.Prim_int("0"));;
let why_zero = Why.Cte(Why.Prim_int("0"));;
let why_one = Why.Cte(Why.Prim_int("1"));;

(*
let coercion_term t e =
  match t with
    | Integer_type -> Why.LApp("int_val",[e])
    | Boolean_type -> Why.LApp("bool_val",[e])
    | Float_type -> Why.LApp("float_val",[e])
    | Class_type(_) -> e
    | Array_type(_) -> e
    | Interface_type(_) -> e
    | Null_type -> ill_typed "coercion_term(null)"
;;
*)

let tr_logic_bin_op loc op t1 t2 =
  let op' = 
    match op with
      | "&&" -> "andb"
      | "||" -> "orb"
      | "le_int" | "ge_int" | "gt_int" | "lt_int" 
      | "le_float" -> "le_real"
      | "ge_float" -> "ge_real"
      | "gt_float" -> "gt_real"
      | "lt_float" -> "lt_real" 
      | "eq_int" -> "eq_int"
      | "eq_float" -> "eq_real"
      | "ne_int" -> op
      | "ne_float" -> "ne_real"	 
      | "add_int" | "sub_int" | "mul_int" | "div_int" | "mod_int" -> op
      | "add_float" -> "add_real"
      | "sub_float" -> "sub_real"
      | "mul_float" -> "mul_real"
      | "div_float" -> "div_real"
      | "Zland" | "Zlor" | "Zlsr" | "Zlsl" | "Zasr" -> op
      | "le_type" -> "subType" 
      | "eq_type" -> op
      | _ -> unsupported loc ("tr_logic_op(" ^ op ^ ")")
  in
  Why.LApp(op',[t1;t2])
;;

let tr_logic_un_op loc op t =
  let op' =
    match op with
    | "!" -> "negb"
    | "neg_int" -> op
    | "float_of_int" -> "real_of_int"
    | "neg_float" -> "neg_real"
    | _ -> unsupported loc ("tr_logic_un_op(" ^ op ^ ")")
  in
  Why.LApp(op',[t])
;;

let rec tr_jml_lit loc l =
  match l with
    | Int(n) -> Why.LConst(Why.Prim_int(string_of_int n))
    | Long(n) -> Why.LConst(Why.Prim_int(Int64.to_string n))
    | Bool(b) -> Why.LConst(Why.Prim_bool(b))
    | Char(c) -> Why.LConst(Why.Prim_int(string_of_int (int_of_char c)))
    | String(_) -> unsupported loc "tr_jml_lit(string)"
    | Float(s) -> Why.LConst(Why.Prim_real(s))
    | Null -> Why.LVar("Null")
;;



let array_access_funs ty =
  match ty with
    | Byte_type | Short_type | Integer_type -> "intA","array_acc_"
    | Float_type -> "floatA","array_acc_"
    | Boolean_type -> "boolA","array_acc_"
    | _ -> "objA","array_acc_"
;;

let array_update_funs ty =
  match ty with
    | Byte_type | Short_type | Integer_type -> "intA", (* "array_upd_", - NEVER USED *) false
    | Float_type -> "floatA", (* "array_upd_", - NEVER USED *) false
    | Boolean_type -> "boolA", (* "array_upd_" - NEVER USED, *)false
    | _ -> "objA", (* "array_upd", - NEVER USED *) true
;;



let method_args_with_effects oldify eff args =
  let args =
    if eff.reads_obj_array
    then
      (oldify "objA")::args
    else
      args
  in
  let args =
    if eff.reads_bool_array
    then
      (oldify "boolA")::args
    else
      args
  in
  let args =
    if eff.reads_int_array
    then
      (oldify "intA")::args
    else
      args
  in
  let args =
    if eff.reads_float_array
    then
      (oldify "floatA")::args
    else
      args
  in
  let args =
    if !Config.javacard
    then
      Inttagset.fold
        (fun ive args ->
             (oldify ive.instance_variable_entry_trans_name) :: (oldify (ive.instance_variable_entry_trans_name ^ "_restore")) ::args)
        eff.reads
        args
    else
      Inttagset.fold
        (fun ive args ->
	   (oldify ive.instance_variable_entry_trans_name) :: args)
        eff.reads
        args
  in
  let args =
    if eff.reads_alloc_table
    then
      (oldify "alloc")::args
    else
      args
  in
  let args = if !Config.javacard then (oldify "volatile")::args else args in
  args
;;


(* label is the label to refer when oldfiying *)

let this_is_result = ref false
(*
let result_is_returned = ref false
*)

let rec tr_term label oldify t =
  match t.expr_node with
    | Lit(l) -> tr_jml_lit t.expr_loc l
    | Var(s) -> ill_typed ("tr_term(var "^(snd s)^" has not been typed)")
    | Typed_var(lve) ->
	let id = lve.local_variable_entry_trans_name in
	if id = "this" && !this_is_result
	then Why.LVar("result")
	else
(*
	  if id = "result" && !result_is_returned
	then Why.LVar("returned")
	else
*)
	  if lve.local_variable_entry_is_ref 
	  then oldify id else Why.LVar(id)
    | Bin(t1, op, t2) ->
	tr_logic_bin_op t.expr_loc op 
	  (tr_term label oldify t1) (tr_term label oldify t2)
    | Un(op,t2) -> tr_logic_un_op t.expr_loc op (tr_term label oldify t2)
    | PrePostIncrDecr(op,t2) -> 
	ill_typed "tr_term(prepostincrdecr)"
    | Instanceof(_,_) -> ill_typed "tr_term(instanceof)"
    | Typed_instanceof (e, t) -> 
	Why.LApp("instanceof_bool",[tr_term label oldify e;tr_type_term t])
    | Cast (t, e) -> ill_typed "tr_term(cast)"
    | Typed_cast(ty,ty',e) -> 
	begin
	  match ty,ty' with
	    | (Float_type | Real_type), (Integer_type | Short_type) ->
		Why.LApp("real_of_int", [tr_term label oldify e])
	    | _ ->
		(* unsupported t.expr_loc "cast"; TODO ! *)
		tr_term label oldify e
	end
    | Array_access (e1, e2) -> ill_typed "tr_term(array_access)"
    | Typed_array_access (ty, e1, e2) -> 
	let alloc_name, access_fun = array_access_funs ty in
	let access_fun = "array_acc" in
	let alloc = oldify alloc_name in
	let acc = 	  
	  Why.LApp(access_fun,
		   [alloc;
		    tr_term label oldify e1;
		    tr_term label oldify e2])
	in (*coercion_term ty*) acc
    | Array_creation _ -> unsupported t.expr_loc "tr_term(new_array)"
    | Class_instance_creation (_, _) -> 
	ill_typed "tr_term(new_object)"
    | Typed_class_instance_creation (_, _) -> 
	unsupported t.expr_loc "tr_term(new_object)"
    | Method_call _ -> ill_typed "tr_term(method_call)"
    | Super_method_call _ -> ill_typed "tr_term(super_method_call)"
    | Typed_method_call(e,me,args) -> 
	if me.method_entry_is_pure then
	begin
	  let args = List.map (tr_term label oldify) args in
	  let args =
	    if me.method_entry_is_static
	    then 
              if args = []
              then [Why.LVar "void"]
              else args
	    else 
	      let tmp_expr = tr_term label oldify e in
	      tmp_expr::args
	  in
	  let args =
	    method_args_with_effects 
	      oldify me.method_entry_routine.routine_entry_effects args
	  in
	  let class_name =
	    match me.method_entry_class_or_interface with
	      | Class_entry ce -> ce.class_entry_name
	      | Interface_entry te -> te.interface_entry_name
	      | _ -> assert false
	  in
            Why.LApp (class_name ^ "_" ^ me.method_entry_trans_name,args)
	end
	else
	  ill_typed ("tr_term(non pure method '" ^ me.method_entry_trans_name ^ "')")
    | Logic_call(le, args) ->
	begin
	  let args = List.map (tr_term label oldify) args in
	  let args = method_args_with_effects oldify le.logic_entry_effects args in
	  Why.LApp(le.logic_entry_name,args)
	end
    | Field_access fa -> tr_field_access_term t.expr_loc label oldify fa
    | If_expr (e1, e2, e3) ->
	Why.LApp("if_then_else",
		 [tr_term label oldify e1;
		  tr_term label oldify e2;
		  tr_term label oldify e3])
    | Assign_array (_, _, _, _) -> ill_typed "tr_term(assign_array)"
    | Typed_assign_array (_, _, _, _, _) -> 
	ill_typed "tr_term(typed_assign_array)"
    | Assign_field (_, _, _) -> 
	ill_typed "tr_term(field_assign)"	
    | Assign_local_var (_, _, _) -> 
	ill_typed "tr_term(var_assign)"
    | Assign_name (_, _, _) -> 
	ill_typed "tr_term(assign_name)"
    | Static_class ce ->  
	unsupported t.expr_loc "tr_term(static_class)"
    | Static_interface(ie) -> 
	unsupported t.expr_loc "tr_term(static_interface)"
    | This -> 
	ill_typed "tr_term(this)"
    | Result -> 
	ill_typed "tr_term(\\result)"
    | Old t' -> 
	tr_term label (fun id -> Why.LVarAtLabel(id,label)) t'
    | Fresh t -> 
	Why.LApp("fresh",[Why.LVarAtLabel("alloc",label);tr_term label oldify t])
    | Not_conditionally_updated (te1, te2, te3) -> assert false
    | Type te -> ill_typed "tr_term(type)"
    | Typed_type te -> tr_type_term te
    | Typeof te -> ill_typed "tr_term(typeof)"
    | Typed_typeof (e, t) -> 
	Why.LApp("typeof_bool",[tr_term label oldify e;tr_type_term t])
    | Quantifier(_) -> 
	ill_typed "tr_term(quantifier)"
    | Typed_quantifier(_) -> 
	unsupported t.expr_loc "tr_term(quantifier)"
    | Fpi(_) -> 
	unsupported t.expr_loc "tr_term(\\fpi)"

and tr_field_access_term loc label oldify f =
  match f with
    | Super_access(id) -> 
	unsupported loc "tr_field_access_term(super)"
    | Primary_access(_,_) ->
	ill_typed "tr_field_access_term(primary)"
    | Typed_constant_access(ive) ->
	let c = ive.constant_entry_interface.interface_entry_name 
		^ "_" ^ ive.constant_entry_name
	in
	Why.LVar(c)
    | Typed_primary_access(e,ive) ->
	begin
	  let c = ive.instance_variable_entry_trans_name in
	  if ive.instance_variable_entry_is_static 
	  then
	    if not ive.instance_variable_entry_is_final 
	    then oldify c else Why.LVar(c)
	  else
            let field = oldify c in
	    let access_fun = "acc" in
	    let acc =
	      Why.LApp
	        (access_fun,
		 [field ; tr_term label oldify e])
	    in (*coercion_term ive.instance_variable_entry_type*) acc
	end
    | Array_length(e) ->
	Why.LApp("arraylength",
		 [oldify "alloc"; tr_term label oldify e])


let tr_term_variable_initializer loc i =
  match i with
    | Simple_initializer(e) -> tr_term "" (fun id -> Why.LVar(id)) e
    | Array_initializer _ -> 
	unsupported loc "tr_term_variable_initializer(array)"
;;


let make_bin label oldify t1 op t2 =
  let why_decl =
    if !Config.javacard
    then
      match t1.expr_node with
       | Typed_method_call(e, me, args) ->
           begin
           match me.method_entry_class_or_interface with
             | Class_entry(ce) ->
               if ce.class_entry_name = "JCSystem"
               then
                 begin 
                   match me.method_entry_name with 
                    | "getTransactionDepth" ->
                        Why.LPred(op, [oldify "transactionDepth"; tr_term label oldify t2])
                    | "isTransient" ->
                      begin 
                        match op with 
                         | "eq_int" ->
                            begin
			    match t2.expr_node with 
			     | Field_access(Typed_primary_access(e,ive)) ->
                               begin
                               match (get_ive_class_or_interface_name ive) with
                                | "JCSystem" ->
			          begin
                                  match ive.instance_variable_entry_name with 
                                   | "NOT_A_TRANSIENT_OBJECT" ->
                                     Why.LPred("eq_bool", [Why.LApp("acc", [Why.LVar("volatile"); tr_term label oldify (List.hd args)]);Why.LConst(Why.Prim_bool(false))])
                                   | "CLEAR_ON_RESET" | "CLEAR_ON_DESELECT" ->
                                     Why.LPred("eq_bool", [Why.LApp("acc", [Why.LVar("volatile"); tr_term label oldify (List.hd args)]);Why.LConst(Why.Prim_bool(true))])
                                   | _ -> 
                                     assert false
                                  end
                                | _ -> 
                                  assert false
                               end
                             | _ ->
		               assert false
                            end                  
                         | _ -> 
		           assert false
                      end
                    | _ ->
                       Why.LPred(op,[tr_term label oldify t1; tr_term label oldify t2])
                 end
               else
                 Why.LPred(op,[tr_term label oldify t1; tr_term label oldify t2])
             | _ ->  
               Why.LPred(op,[tr_term label oldify t1; tr_term label oldify t2])
           end
       | _ -> 
          Why.LPred(op,[tr_term label oldify t1; tr_term label oldify t2])
    else
      Why.LPred(op,[tr_term label oldify t1; tr_term label oldify t2])
    in why_decl


let is_assertion e =
  match e.expr_node with
  | Bin(_, ("&&" | "||" | "<==>" | "==>" | "<=="), _)
  | Un("!", _)
  | Lit(Bool _)
  | If_expr _
  | Typed_instanceof _
  | Fresh _
  | Not_conditionally_updated _
  | Typed_quantifier _ -> true
  | _ -> false
;;


let rec tr_predicate_bin_op loc label oldify op t1 t2 =
  match op with
    | "&&" -> Why.make_and 
	(tr_assertion_rec label oldify t1) 
	(tr_assertion_rec label oldify t2)
    | "||" -> Why.make_or
	  (tr_assertion_rec label oldify t1)
	  (tr_assertion_rec label oldify t2)
    | "==>" -> 
	Why.make_impl 
	(tr_assertion_rec label oldify t1) 
	(tr_assertion_rec label oldify t2)
    | "<==" -> 
	Why.make_impl 
	(tr_assertion_rec label oldify t2) 
	(tr_assertion_rec label oldify t1)
    | "<==>" -> 
	Why.make_equiv 
	(tr_assertion_rec label oldify t1) 
	(tr_assertion_rec label oldify t2)
    | "<=!=>" -> 
	Why.LNot(Why.make_equiv
		   (tr_assertion_rec label oldify t1) 
		   (tr_assertion_rec label oldify t2))
    | "le_int" | "ge_int" | "gt_int" | "lt_int"  ->
	make_bin label oldify t1 op t2	
    | "le_float" -> make_bin label oldify t1 "le_real" t2
    | "ge_float" -> make_bin label oldify t1 "ge_real" t2
    | "gt_float" -> make_bin label oldify t1 "gt_real" t2
    | "lt_float" -> make_bin label oldify t1 "lt_real" t2
    | "eq_int" -> make_bin label oldify t1 "eq_int" t2	
    | "eq_float" -> make_bin label oldify t1 "eq_real" t2	
    | "eq_bool" -> make_bin label oldify t1 op t2	
    | "ne_int" -> make_bin label oldify t1 "neq_int" t2
    | "ne_float" -> make_bin label oldify t1 "neq_real" t2
    | "ne_bool" -> make_bin label oldify t1 "neq_bool" t2
    | "neq" | "eq" -> make_bin label oldify t1 op t2
    | "le_type" -> make_bin label oldify t1 "subType" t2
    | "eq_type" -> make_bin label oldify t1 op t2
    | "add_int" | "sub_int" | "mul_int" | "div_int" | "mod_int" 
    | "add_float" | "sub_float" | "mul_float" | "div_float" 
    | "land" 
    | "lor" 
    | "lsr" 
    | "lsl" 
    | "asr" -> ill_typed "tr_predicate_op"
    | _ -> unsupported loc ("tr_predicate_bin_op(" ^ op ^ ")")


and tr_assertion_rec label oldify a =   
  match a.expr_node with
    | Bin(t1,op,t2) ->
	tr_predicate_bin_op a.expr_loc label oldify op t1 t2
    | Un("!",t2) ->
	Why.LNot(tr_assertion_rec label oldify t2)
    | If_expr(b, t1, t2) ->
	let a1 = tr_assertion_rec label oldify t1 in
	let a2 = tr_assertion_rec label oldify t2 in
	(* NR - if b corresponds to a Why assertion, it is not a term 
	   so (b ? t1 : t2) is interpreted as ((b ==> t1) && (!b ==> t2)) *)
	if is_assertion b then
	  let a = tr_assertion_rec label oldify b in
	  Why.make_and
	    (Why.make_impl a a1)
	    (Why.make_impl (Why.LNot a) a2)
	else
	  Why.LIf(tr_term label oldify b, a1, a2)
    | Lit(Bool(true)) -> Why.LTrue
    | Lit(Bool(false)) -> Why.LFalse
    | Typed_instanceof (e, t) -> 
	Why.LPred("instanceof",[Why.LVar("alloc");tr_term label oldify e;tr_type_term t])
    | Logic_call(le,args) ->
	begin
	  let args = List.map (tr_term label oldify) args in
	  let args = method_args_with_effects oldify le.logic_entry_effects args in
	  Why.LPred(le.logic_entry_name,args)
	end
    | Typed_method_call(e,me,args) when me.method_entry_is_predicate ->
	begin
	if me.method_entry_is_pure then
	begin
	  let args = List.map (tr_term label oldify) args in
	  let args =
	    if me.method_entry_is_static
	    then args
	    else 
	      let tmp_expr = tr_term label oldify e in
	      tmp_expr::args
	  in
	  let args =
	    method_args_with_effects oldify me.method_entry_routine.routine_entry_effects args
	  in
	  let class_name =
	    match me.method_entry_class_or_interface with
	      | Class_entry ce -> ce.class_entry_name
	      | Interface_entry te -> te.interface_entry_name
	      | _ -> assert false
	  in
	  Why.LPred
	    (class_name ^ "_" ^ me.method_entry_trans_name,args)
	end
	else
	  ill_typed "tr_assertion(non pure method)"
(*
	  match e.expr_node with
	    | Static_class(ce) ->
		let args =
		  if me.method_entry_effects.reads_alloc then
		    let alloc = "alloc" ^ oldify in
		    Why.LVar(alloc)::
		    List.map (tr_term label oldify) args
		  else
		    List.map (tr_term label oldify) args
		in
		Why.LPred(ce.class_entry_name ^ "_" ^ me.method_entry_name,
			  args)
	    | Typed_var(lve) ->
		begin
		  match lve.local_variable_entry_type with
		    | Class_type(ce) ->
			let args =
			  if me.method_entry_effects.reads_alloc then
			    let alloc = oldify "alloc" in
			    alloc::
			    List.map (tr_term label oldify) (e::args)
			  else
			    List.map (tr_term label oldify) (e::args)
			in
			Why.LPred
			  (ce.class_entry_name ^ "_" ^ me.method_entry_name,
			   args)
		    | _ -> ill_typed "tr_term(method_call(typed_var))"
		end
	    | Typed_cast(tt,{expr_node = Typed_var(lve)}) ->
		begin
		  match tt with
		    | Class_type(_) | Interface_type(_) ->
			let args =
			  if me.method_entry_effects.reads_alloc then
			    let alloc = oldify "alloc" in
			    alloc::
			    List.map (tr_term label oldify) (e::args)
			  else
			    List.map (tr_term label oldify) (e::args)
			in
			Why.LPred
			  ((type_entry_name tt)^ "_" ^ me.method_entry_name,
			   args)
		    | _ ->
 			unsupported e.expr_loc "tr_term(method_call(cast))"
		end
	    | _ ->
		Format.fprintf Config.log 
		"unsupported: tr_assertion(method_call(%a))@." 
		Ast.fprintf_expr e;
		unsupported e.expr_loc "tr_term(method_call(novar))"	
*)
	end
    | Typed_quantifier(Forall, idl, e1, e2) -> 
	List.fold_right
	  (fun lve acc ->
	    let ty = tr_type lve.local_variable_entry_type in
	    Why.LForall(lve.local_variable_entry_trans_name, ty, acc))
	  idl
	  (Why.LImpl(tr_assertion_rec label oldify e1,
		     tr_assertion_rec label oldify e2))
    | Typed_quantifier(Exists, idl, e1, e2) -> 
	List.fold_right
	  (fun lve acc ->
	    let ty = tr_type lve.local_variable_entry_type in
	    Why.LExists(lve.local_variable_entry_trans_name, ty, acc))
	  idl
	  (Why.make_and (tr_assertion_rec label oldify e1)
	     (tr_assertion_rec label oldify e2))
    | Old b -> tr_assertion_rec label (fun id -> Why.LVarAtLabel(id,label)) b
    | Fresh b -> Why.LPred("fresh",[Why.LVarAtLabel("alloc",label);
				    tr_term label oldify b])
    | Not_conditionally_updated (te1, te2, te3) -> 
	Why.LForall("i",
		    ([],"int"),
		    Why.LImpl(Why.LAnd(Why.LPred("le_int", [why_lzero; Why.LVar("i")]),
				       Why.LPred("lt_int", [Why.LVar("i"); tr_term label oldify te3])),
			      Why.LPred("eq_bool", [Why.LApp("array_acc", [Why.LVar("intA_assigned"); 
									   tr_term label oldify te1; 
									   Why.LApp("add_int", [tr_term label oldify te2; Why.LVar("i")])]);
						    Why.LConst(Why.Prim_bool(false))])))
    | Fpi (e,f1,f2) ->
	Why.LPred("fpi",[tr_term label oldify e; 
			 Why.LConst(Why.Prim_real(f1)); 
			 Why.LConst(Why.Prim_real(f2))])
    | Typed_typeof (e, t) -> 
	Why.LPred("typeof",[Why.LVar("alloc");tr_term label oldify e;tr_type_term t])
    | _ -> 
	Why.LPred("eq_bool",[tr_term label oldify a;
			     Why.LConst(Why.Prim_bool(true))])

let tr_assertion label = tr_assertion_rec label (fun id -> Why.LVar(id)) 


let tr_bin_op loc s =
  let s' = 
    match s with
      | "add_int" | "sub_int" | "mul_int" -> s
      | "div_int" | "mod_int" -> 
	  if !Config.nonnull_only then s else s ^ "_"
      | "add_float" -> "add_real"
      | "sub_float" -> "sub_real"
      | "mul_float" -> "mul_real"
      | "div_float" -> "div_real"
      | "le_int" | "ge_int" | "gt_int" | "lt_int" -> s ^ "_"
      | "le_float" | "ge_float" | "gt_float" | "lt_float" -> s
      | "||" | "&&" -> s
      | "Zland" | "Zlor" | "Zlxor" | "Zlsr" | "Zlsl" | "Zasr" -> s
      | "andb" | "orb" -> s
      | "eq_int" -> s ^ "_"
      | "eq_float" -> s ^ "_"
      | "eq_bool" -> s ^ "_"
      | "ne_int" -> "neq_int_"
      | "ne_float" -> "neq_float_"
      | "ne_bool" -> "neq_bool_"
      | "neq_value" | "eq_value" -> s 
      | "eq" -> s ^ "_value" 
      | "neq" -> "ne_value" 
      | _ -> unsupported loc ("tr_bin_op(" ^ s ^ ")")
  in
  Why.Var(s')

let tr_un_op loc s =
  let s' = 
   match s with
    | "neg_int" -> s
    | "neg_float" -> s
    | "float_of_int" -> s
    | "!" -> "negb"
    | "lcompl" -> "Zlcompl"
    | _ -> unsupported loc ("tr_un_op(" ^ s ^ ")")
  in
  Why.Var(s')


(*

 coercion

*)

(*
let coercion_expr t e =
  match t with
    | Byte_type | Short_type | Integer_type -> Why.make_app "int_val" [e]
    | Boolean_type -> Why.make_app "bool_val" [e]
    | Float_type -> Why.make_app "float_val" [e]
    | Class_type(_) -> e
    | Array_type(_) -> e
    | Interface_type(_) -> e
    | Null_type -> ill_typed "coercion_expr(null)"
;;
*)

(*

  auxiliary functions to detect purity of expressions, and decide when
to introduce intermediaite variables or not.

*)

let rec is_pure e =
  match e.expr_node with 
    | Assign_local_var(_, _, _) -> false
    | Method_call _ -> assert false
    | Typed_method_call _ -> false
    | PrePostIncrDecr _ -> false
    | Bin(e1,op,e2) -> is_pure e1 && is_pure e2
    | Typed_cast(_, _, e) -> is_pure e
    | Typed_array_access(_, e1, e2) -> is_pure e1 && is_pure e2
    | _ -> 
	let eff = Effects.expr e in 
	(not eff.writes_alloc) && (Inttagset.is_empty eff.writes) 
	&& (not eff.throws) && (not eff.breaks)
;;

let build_why_app e args =
  let rec build e args =
    match args with
      | [] -> e
      | p::l -> build (Why.App(e,p)) l
  in
  match args with
    | [] -> Why.App(e,Why.Void)
    | _ -> build e args
	  

let tempindex_count = ref(-1);;

let temp_index () =
  incr tempindex_count;
  if !tempindex_count = 0
  then "krak_index"
  else "krak_index" ^ (string_of_int !tempindex_count)
;;

(* Build a Let only if the expression is complex 
    [v : string,f : Why.expr -> Why.term -> Why.expr; e: Why.expr] 
*)

let why_let_var is_pure v f e = match e with 
  | (Why.Cte n) -> f e (Why.LConst n)
  | (Why.Var x) -> f e (Why.LVar x) 
  | (Why.Deref x) when is_pure -> f e (Why.LVar x) 
  | _ -> Why.Let(v,e,f (Why.Var v) (Why.LVar v))

let why_let_var_ref = why_let_var true
(*i
v f e = match e with 
    (Why.Var x) -> f e (Why.LVar x)
  | (Why.Cte n) -> f e (Why.LConst n)
  | (Why.Deref x) -> f e (Why.LVar x) 
  | _ -> Why.Let(v,e,f (Why.Var v) (Why.LVar v))
i*)


let tempvar_count = ref 0;;
let temp_switch_count = ref 0;;

let build_complex_app e args =
  let rec build n e args =
    match args with
      | [] -> e
      | [p] -> Why.App(e,p)
      | ((Why.Var _) | (Why.Cte _) as p)::l ->
	  build n (Why.App(e,p)) l
        | (Why.Let(s, e1, e2))::l ->
            Why.Let(s, e1, build n (Why.App(e, e2)) l)
      | p::l ->
	  incr tempvar_count;
	  let v = "krakv"^(string_of_int !tempvar_count) in
	  Why.Let(v,p,
	    build (succ n) (Why.App(e,Why.Var(v))) l)
  in
  match args with
    | [] -> Why.App(e,Why.Void)
    | _ -> build 1 e args

exception Impure

let rec build_app1 e args =
  let rec build e args =
    match args with
      | [] -> e
      | [p] -> Why.App(e, p)
      | (Why.App(e1, e2) as p)::l -> 
          incr tempvar_count;
          let v = "krakatoa_var" ^ (string_of_int !tempvar_count) in
          Why.Let(v, p, build (Why.App(e, Why.Var(v))) l)
      | (Why.Let(s, e1, e2))::l ->
          Why.Let(s, e1, build (Why.App(e, e2)) l)
(*        | (Why.Triple(pre, b, post, exc, ct) as p)::l ->
            begin
            match b with
              | Why.Let(v, e1, e2) ->
                  Why.Let(v, e1, build (Why.App(e, Why.Triple(pre, e2, post, exc, ct))) l)
              | _ -> build (Why.App(e, p)) l
            end *)
        | p::l -> build (Why.App(e, p)) l
    in
    build e args  
    (* List.fold_left Why.App(acc, tr_expr x)) e args *)

and build_app e args =
  match args with
    | [] -> Why.App(e,Why.Void)
    | _ ->
       if List.for_all is_pure args 
       then build_app1 e (List.map tr_expr args)
       else build_complex_app e (List.map tr_expr args)


(* translation of cast (pas au point !)*)

and tr_cast t t' e = 
  match t with
(*i
    | Base_type(s) -> tr_expr e (* TODO ! *)	
    | Type_name(qi) ->
	begin
	  match qi with
	    | [(loc,id)] -> 
		let result_type = Why.Var(id) (* correct ?? *) in
		Why.make_app "cast" [result_type;tr_expr e]
	    | _ ->unsupported e.expr_loc "tr_cast(longname)"
	end
    | Array_type_expr(t) -> unsupported e.expr_loc "tr_cast(array)"
i*)
  | Byte_type
  | Short_type
  | Integer_type
  | Float_type
  | Boolean_type -> tr_expr e (* TODO ! *)	
  | Null_type -> assert false
(*i
  | Class_type(te) ->
      let result_type = Why.Var(te.class_entry_name) in
      Why.make_app "cast" [result_type;tr_expr e]      
  | Interface_type(te) ->
      let result_type = Why.Var(te.interface_entry_name) in
      Why.make_app "cast" [result_type;tr_expr e]      
i*)
(*i
  | _ ->
      let result_type = tr_type_expr t in
      Why.make_app "cast" [result_type;tr_expr e]      
i*)
  | _ ->
      let result_type = tr_type_term t in
      why_let_var (is_pure e)
	"krak_cast"
        (fun whye loge ->
	    Why.make_pre
	      (if !Config.nonnull_only then Why.LTrue else Why.LPred("instanceof", [Why.LVar("alloc"); loge; result_type]))
	      whye)
	(tr_expr e)
	

and tr_primary_access e ive = 
  why_let_var_ref (* we allow deref here because (Field ive) is always pure *)
    "krak_access"
    (fun whye loge ->
       let field = tr_ive ive in
       let acc = 
           Why.make_app "acc_"
	     [Why.Deref(field) ; whye] 
       in
       match loge with
	 | Why.LVar "this" -> acc
	 | _ ->
	     Why.make_pre
	     (Why.LPred("neq",[loge;Why.LVar("Null")]))
	     acc)
    (tr_expr e)

and tr_field_access loc f =
  match f with
    | Super_access(id) -> unsupported loc "tr_field_access(super)"
    | Primary_access(e,id) -> ill_typed "tr_field_access(primary)"
    | Typed_primary_access(e,ive) ->
	begin
	  if ive.instance_variable_entry_is_static
	  then
	    if ive.instance_variable_entry_is_final
	    then Why.Var(ive.instance_variable_entry_trans_name) 
	    else Why.Deref(ive.instance_variable_entry_trans_name) 
	  else
	    tr_primary_access e ive
	end
    | Typed_constant_access(const) ->
	let c = const.constant_entry_interface.interface_entry_name in
	Why.Var(c ^ "_" ^ const.constant_entry_name) 
    | Array_length(e) ->
	why_let_var_ref 
	  "krak_access"
	  (fun whye loge ->
	     let acc =
	       Why.make_app "arraylength" [Why.Deref "alloc"; whye]
	     in
	     match loge with
	       | Why.LVar "this" -> acc
	       | _ -> 
		   let pre = Why.LPred("neq",[loge;Why.LVar("Null")]) in
		   Why.make_pre pre acc)
	  (tr_expr e)


and tr_pre_post_incr_decr op e =
  match e.expr_node with
    | Var(x) -> ill_typed "tr_pre_post_incr_decr(var)"
    | This -> ill_typed "tr_pre_post_incr_decr(this)"
    | Typed_var(lve) -> 
	let v = lve.local_variable_entry_trans_name in
	Why.Assign
	  (v, build_why_app (tr_bin_op e.expr_loc op) [Why.Deref v; why_one])
    | Field_access(fa) -> 
	begin
	  match fa with
	    | Array_length _ -> ill_typed "tr_pre_post_incr_decr(length)"
	    | Typed_constant_access _ -> ill_typed "tr_pre_post_incr_decr(constant)"
	    | Primary_access (_, _) -> 
		ill_typed "tr_pre_post_incr_decr(primary)"
	    | Super_access _ -> 
		unsupported e.expr_loc "tr_pre_post_incr_decr(super)"
	    | Typed_primary_access(f, ive) ->
		let tr = 
                  if ive.instance_variable_entry_is_static 
                  then Why.Void 
                  else tr_expr f 
                in
		tr_field_assignment_op e.expr_loc tr ive op why_one
	end
    | Typed_array_access(jt, e1, e2) ->
        begin
          match jt with
            | Byte_type | Short_type | Integer_type ->
               tr_array_assignment Integer_type e1 e2 (is_pure e) (Why.make_app op [tr_expr e; why_one])
            | _ -> assert false (* should never happen *)
        end
    | _ -> unsupported e.expr_loc "tr_pre_post_incr_decr()"

and tr_post_incr_decr op e =
  match e.expr_node with
    | Var(x) -> ill_typed "tr_post_incr_decr(var)"
    | This -> ill_typed "tr_post_incr_decr(this)"
    | Typed_var(lve) -> 
	let v = lve.local_variable_entry_trans_name in
	Why.Let("tmp", 
                Why.Deref v,
		Why.append
		  (Why.Assign(v, (build_why_app (tr_bin_op e.expr_loc op) 
			[Why.Deref(v); why_one])))
		  (Why.Var "tmp"))
    | Field_access(fa) -> 
	begin
	  match fa with
	    | Array_length _ -> ill_typed "tr_pre_post_incr_decr(length)"
	    | Typed_constant_access _ -> ill_typed "tr_pre_post_incr_decr(constant)"
	    | Primary_access (_, _) -> 
		ill_typed "tr_post_incr_decr(primary)"
	    | Super_access _ -> 
		unsupported e.expr_loc "tr_post_incr_decr(super)"
	    | Typed_primary_access(f,ive) ->
		unsupported e.expr_loc "tr_post_incr_decr(primary)"
	end
    | _ -> unsupported e.expr_loc "tr_post_incr_decr()"


and tr_pre_incr_decr op e =
  match e.expr_node with
    | Var(x) -> ill_typed "tr_post_incr_decr(var)"
    | This -> ill_typed "tr_post_incr_decr(this)"
    | Typed_var(lve) -> 
	let v = lve.local_variable_entry_trans_name in
	Why.Let("tmp",build_why_app (tr_bin_op e.expr_loc op) 
			[Why.Deref(v);why_one],
		Why.append
		  (Why.Assign (v,Why.Var "tmp"))
		  (Why.Var "tmp"))
    | Field_access(fa) -> 
	begin
	  match fa with
	    | Array_length _ -> ill_typed "tr_pre_post_incr_decr(length)"
	    | Typed_constant_access _ -> ill_typed "tr_pre_post_incr_decr(constant)"
	    | Primary_access (_, _) -> 
		ill_typed "tr_pre_incr_decr(primary)"
	    | Super_access _ -> 
		unsupported e.expr_loc "tr_pre_post_incr_decr(super)"
	    | Typed_primary_access(f, ive) ->
		let tr = 
                  if ive.instance_variable_entry_is_static 
                  then Why.Void 
                  else tr_expr f 
                in
		Why.append
		(tr_field_assignment_op e.expr_loc tr ive op why_one)
                (Why.Deref(ive.instance_variable_entry_trans_name))
	end
    | _ -> unsupported e.expr_loc "tr_pre_incr_decr()"


and tr_field_assignment loc f ive is_pure accu is_void =
  let why_expr = tr_field_assignment_expr loc f ive is_pure accu in
  if is_void then 
    why_expr 
  else 
    Why.Block([why_expr; Why.make_app "acc_" [Why.Deref(ive.instance_variable_entry_trans_name); tr_expr f]])

and tr_field_assignment_expr loc f ive is_pure accu = 
  if ive.instance_variable_entry_is_static
  then
    Why.Assign(ive.instance_variable_entry_trans_name, accu)
  else
    let _ = ive.instance_variable_entry_type in
    why_let_var is_pure "krak_update"
      (fun whyf logf -> 
	 why_let_var is_pure "krak_acc"
	   (fun whyaccu logaccu ->
	      let assign =
(*                if !Config.javacard && !Config.transactionException
                then 
		  Why.Assign
		    (ive.instance_variable_entry_trans_name, 
		     Why.make_app "upd_transaction"
		       [Why.Var(ive.instance_variable_entry_trans_name);
		        whyf;
		        whyaccu])
		else *)
		  Why.Assign
		    (ive.instance_variable_entry_trans_name, 
		     Why.make_app "upd_"
		       [Why.Deref(ive.instance_variable_entry_trans_name);
		        whyf;
		        whyaccu])
	      in 
              match logf with
	        | Why.LVar "this" -> assign
		| _ -> 
		    let pre =
		      Why.LPred("neq",[logf;Why.LVar("Null")])
		    in Why.make_pre pre assign)
	 accu)
      (tr_expr f)
    
and tr_field_assignment_op loc f ive op accu = 
  if ive.instance_variable_entry_is_static
  then
    let c = ive.instance_variable_entry_trans_name
    in
    Why.Assign(c, build_why_app (tr_bin_op loc op) [Why.Deref(c); accu])
  else
    (* on traduit f.ive op= accu par
       let krak_update = f in
       let kra_accu = (op (access !alloc krak_update (Field ive)) accu) in
       update f krak_update accu
       {}
       *)
    let _lfield = Why.LVar(tr_ive ive)
    and field = tr_ive ive
    in
    why_let_var false "krak_update"
      (fun whyf logf -> 
	 let acc =
           Why.make_app "acc_"
	     [Why.Deref(field) ; whyf]
	 in
	 let access =
           match logf with
             | Why.LVar "this" -> acc
             | _ -> 
                 let pre =
                   Why.LPred("neq",[logf;Why.LVar("Null")])
                 in
                 Why.make_pre pre acc
	 in
	 why_let_var false "krak_acc"
	   (fun whyaccu logaccu ->
	      let assign =
                if !Config.javacard && !Config.transactionException
                then 
		  Why.Assign
		    (field, 
		     Why.make_app "upd_transaction"
		       [Why.Var(field);
                        whyf;
                        whyaccu])
              else 
		  Why.Assign
		    (field, 
		     Why.make_app "upd_"
		       [Why.Deref(field);
                        whyf;
                        whyaccu])
              in 
		match logf with
                  | Why.LVar "this" -> assign
                  | _ -> 
                      let pre =
			Why.LPred("neq",[logf;Why.LVar("Null")])
                      in Why.make_pre pre assign)
	   (build_why_app (tr_bin_op loc op) [access;accu]))
    f
    
and tr_array_assignment t e1 e2 accu_is_pure accu = 
  let alloc_name,check_is_valid = array_update_funs t in
  why_let_var (accu_is_pure && (is_pure e2)) "krak_update" 
    (fun whye1 loge1 ->
       why_let_var accu_is_pure (temp_index ())
	 (fun whye2 loge2 ->
	    why_let_var true "krak_acc"
	      (fun whyaccu logaccu -> 
		 let pre1 =
(* NR is_valid_array_upd n'existe plus dans local_memory.why ---> erreur Why *)
(* NR		   if check_is_valid
		   then
		     Why.LPred
		       ("is_valid_array_upd",
			[Why.LVar("alloc");
			 loge1; 
			 loge2;
			 logaccu])
		   else *)
		   if !Config.nonnull_only then Why.LTrue else 
                   if !Config.unfold_pred >= 1
                   then
		     Why.LAnd(
                     Why.LPred("le_int", [Why.LConst(Why.Prim_int("0")); loge2]),
                     Why.LPred("lt_int", [loge2; Why.LApp("arraylength", [Why.LVar "alloc"; loge1])]))
                   else
		     Why.LPred
		       ("is_valid_index",
			[Why.LVar("alloc");
			 loge1; 
			 loge2])
		 in
		 let pre = 
		   Why.make_and 
		     (Why.LPred("neq", [loge1; Why.LVar("Null")])) 
		     pre1 
		 in
                 let e = (Why.Assign
			    (alloc_name, 
			     Why.make_app "array_upd_"
			       [Why.Deref(alloc_name);
				whye1;
				whye2;
				whyaccu])) 
                 in
		   Why.make_pre pre e)
	    accu)
       (tr_expr e2))
    (tr_expr e1)

and replace_by_restore_fa fa =
  match fa with
  | Super_access _ -> ill_typed "replace_by_restore_fa"
  | Primary_access(_, _) -> ill_typed "replace_by_restore_fa"
  | Typed_primary_access(e, ive) -> 
      Typed_primary_access(
        replace_by_restore e, 
        { ive with instance_variable_entry_trans_name = ive.instance_variable_entry_trans_name ^ "_restore" })
  | Typed_constant_access ce -> fa
  | Array_length e -> Array_length(replace_by_restore e)   

and replace_by_restore_ac ac =
  match ac with 
  | Explicit_array_creation(e, ac) -> Explicit_array_creation(replace_by_restore e, replace_by_restore_ac ac)
  | Implicit_array_creation _ -> failwith "typing error"
  | Typed_implicit_array_creation _ -> ac
   
and replace_by_restore e =
  match e.expr_node with
  | Lit _ -> e
  | Bin(e1, op, e2) -> mk_expr_no_loc (Bin(replace_by_restore e1, op, replace_by_restore e2))
  | Un(op, e1) -> mk_expr_no_loc (Un(op, replace_by_restore e1))
  | PrePostIncrDecr(i, e) -> mk_expr_no_loc (PrePostIncrDecr(i, replace_by_restore e))
  | Var _ -> failwith "typing error"
  | Typed_var _ -> e
  | Static_class _ -> e
  | Static_interface _ -> e
  | Assign_name(qid, s, e) -> mk_expr_no_loc (Assign_name(qid, s, replace_by_restore e))
  | Assign_local_var(lve, s, e) -> mk_expr_no_loc (Assign_local_var(lve, s, replace_by_restore e))
  | Assign_field(fa, s, e) -> mk_expr_no_loc (Assign_field(fa, s, replace_by_restore e))
  | Assign_array(_, _, _, _) -> failwith "typing error"
  | Typed_assign_array(t, e1, e2, s, e3) -> mk_expr_no_loc (Typed_assign_array(t, replace_by_restore e1, replace_by_restore e2, s, replace_by_restore e3))  
  | If_expr(e1, e2, e3) -> mk_expr_no_loc (If_expr(replace_by_restore e1, replace_by_restore e2, replace_by_restore e3))
  | This -> e
  | Field_access fa -> mk_expr_no_loc (Field_access(replace_by_restore_fa fa))
  | Method_call(_, _, _) -> failwith "typing error"
  | Super_method_call(_, _) -> failwith "typing error"
  | Typed_method_call (e1, me, el) -> 
      mk_expr_no_loc (Typed_method_call(replace_by_restore e1, me, List.map replace_by_restore el))
  | Logic_call(le, el) -> mk_expr_no_loc (Logic_call(le, List.map replace_by_restore el))
  | Class_instance_creation(_, _) -> failwith "typing error" 
  | Typed_class_instance_creation(ce, el) ->
      mk_expr_no_loc (Typed_class_instance_creation(ce, List.map replace_by_restore el))
  | Array_creation(ac) -> mk_expr_no_loc (Array_creation(replace_by_restore_ac ac))
  | Array_access(_, _) -> failwith "typing error"
  | Typed_array_access(t, e1, e2) -> 
      mk_expr e.expr_loc (Typed_array_access(t, replace_by_restore e1, replace_by_restore e2))
  | Cast(te, e) -> failwith "typing error"
  | Typed_cast(t, t', e) -> mk_expr_no_loc (Typed_cast(t, t', replace_by_restore e))
  | Instanceof(_, _) -> failwith "typing error"
  | Typed_instanceof(e1, java_type) -> mk_expr e.expr_loc (Typed_instanceof(replace_by_restore e1, java_type))
  | Result -> e  
  | Old(e) -> mk_expr_no_loc (Old(replace_by_restore e)) 
  | Fresh(e) -> mk_expr_no_loc (Fresh(replace_by_restore e)) 
  | Not_conditionally_updated(e1, e2, e3) -> mk_expr_no_loc (Not_conditionally_updated(replace_by_restore e1, replace_by_restore e2, replace_by_restore e3))
  | Type _ -> failwith "typing error"
  | Typed_type _ -> failwith "typing error"
  | Typeof _ -> failwith "typing error" 
  | Typed_typeof(_, _) -> failwith "typing error"
  | Quantifier(_, _, _, _, _) -> failwith "typing error"  
  | Typed_quantifier(q, lvel, e1, e2) -> 
      mk_expr
        e.expr_loc
        (Typed_quantifier(q, lvel, replace_by_restore e1, replace_by_restore e2))
  | Fpi(_, _, _) -> failwith "typing error"
    
and tr_method_call e me args =
  Format.fprintf Config.log 
    "%s: translating method call@." (Location.to_string e.expr_loc);
  let args =
    if me.method_entry_is_static
    then args
    else e::args
  in
  let class_name =
    match me.method_entry_class_or_interface with
      | Class_entry ce -> ce.class_entry_name
      | Interface_entry te -> te.interface_entry_name
      | _ -> assert false
  in
  if !Config.javacard
  then
    match class_name with
    | "JCSystem" -> 
      begin
      match me.method_entry_name with
       | "getTransactionDepth" ->
         build_app (Why.Var "getTransactionDepth") args
       | "makeTransientBooleanArray" ->
         build_app (Why.Var "makeTransientBooleanArray") args
       | "makeTransientByteArray" ->
         build_app (Why.Var "makeTransientByteArray") args
       | "makeTransientShortArray" ->
         build_app (Why.Var "makeTransientShortArray") args
       | "makeTransientObjectArray" ->
         build_app (Why.Var "makeTransientObjectArray") args
       | "commitTransaction" ->
	 build_app (Why.Var "commitTransaction") []
       | _ -> 
         build_app
           (Why.Var
              (class_name ^ "_" ^ 
	       me.method_entry_trans_name ^ "_parameter"))
            args
      end
    | "Util" ->
      begin
      match me.method_entry_name with
      | "arrayCopyNonAtomic" ->
         let args =
           (replace_by_restore (List.nth args 2))::args
         in
         let why_args = List.map tr_expr args in
         let why_args = 
           match why_args with
             | we::tl -> (Why.reset_assertions we)::tl
             | _ -> assert false (* should never happen *)
         in
         let array_copy_non_atomic =
	   if !Config.secure then "arrayCopyNonAtomic_secure"
	   else if !Config.sun then "arrayCopyNonAtomic_sun"
	   else if !Config.impl then "arrayCopyNonAtomic_impl"
	   else "arrayCopyNonAtomic_secure" (* secure is the default :-) *)
	 in
         if List.for_all is_pure args
         then build_app1 (Why.Var array_copy_non_atomic) why_args
         else build_complex_app (Why.Var array_copy_non_atomic) why_args
      | _ -> 
      build_app
      (Why.Var
       (class_name ^ "_" ^ 
		   me.method_entry_trans_name ^ "_parameter"))
      args
      end
    | _ ->
      if !Config.transactionException
      then
        let qe = Why.LPred("instanceof", [Why.LVar("alloc"); Why.LVar("result"); Why.LApp("ClassType", [Why.LVar("TransactionException")])]) in
        Why.append
          (build_app
             (Why.Var
                (class_name ^ "_" ^ 
	         me.method_entry_trans_name ^ "_parameter"))
              args)
          (Why.BlackBox(Why.Annot_type(Why.LTrue, Why.unit_type, ["alloc"], [], Why.LTrue, [("Exception", qe)])))
      else
        build_app
          (Why.Var
             (class_name ^ "_" ^ 
	      me.method_entry_trans_name ^ "_parameter"))
           args
  else 
  build_app
    (Why.Var
       (class_name ^ "_" ^ 
	me.method_entry_trans_name ^ "_parameter"))
    args

(*
  match e.expr_node with
    | Var(x) -> ill_typed "tr_method_call(var)"
    | This -> ill_typed "tr_method_call(this)"
    | Typed_var(lve) -> 
	begin
	  let v = lve.local_variable_entry_trans_name in
	  match me.method_entry_class_or_interface with
            | Class_entry(ce) ->
		let this = 
		  if lve.local_variable_entry_is_ref
		  then Why.Deref(v) else Why.Var(v)
		in
		build_app
		  (Why.App
		     (Why.Var
			(ce.class_entry_name ^ "_" ^ 
			 me.method_entry_name ^ "_parameter"),
			this))
		  args
	    | _ -> unsupported e.expr_loc "tr_method_call(var)"
	end
    | Static_class(ce) ->
	build_app
	  (Why.Var(ce.class_entry_name ^ "_" ^ me.method_entry_name ^ "_parameter")) 
	  args
    | _ -> unsupported e.expr_loc "tr_method_call()"
*)

and tr_expr e =
  Format.fprintf Config.log
    "%s: translating expr.@." (Location.to_string e.expr_loc);
  match e.expr_node with
    | Lit(l) -> tr_literal e.expr_loc l
    | Bin(e1,op,e2) ->
	begin
	  match op with
	    | "&&" -> Why.And(tr_expr e1,tr_expr e2)
	    | "||" -> Why.Or(tr_expr e1,tr_expr e2)
	    | _ ->
		build_app (tr_bin_op e.expr_loc op) [e1;e2]
	end
    | Un(op,e1) ->
	build_app (tr_un_op e.expr_loc op) [e1]
    | PrePostIncrDecr(op, e1) ->
	begin
	  match op with
	    | `PREINCR -> 
		tr_pre_incr_decr "add_int" e1
	    | `POSTINCR -> 
		tr_post_incr_decr "add_int" e1
	    | `PREDECR ->
		tr_pre_incr_decr "sub_int" e1
	    | `POSTDECR -> 
		tr_post_incr_decr "sub_int" e1 
	end
    | Var(_) -> ill_typed "tr_expr(var)"
    | Typed_var(lve) ->
	if lve.local_variable_entry_is_ref
	then Why.Deref(lve.local_variable_entry_trans_name)
	else Why.Var(lve.local_variable_entry_trans_name)
    | Static_class(_) -> unsupported e.expr_loc "tr_expr(static_class)"
    | Static_interface(_) -> 
	Format.fprintf Config.log
	"e = %a@." Ast.fprintf_expr e;
        unsupported e.expr_loc "tr_expr(static_interface)"
    | This -> ill_typed "tr_expr(this)" 
    | Assign_name(_,_,_) -> ill_typed "tr_expr(assign_name)"
    | Assign_local_var(lve, op, e1) ->
	begin
	  match op with
	    | "=" ->
                let n = lve.local_variable_entry_trans_name in
                Why.append (Why.Assign(n, tr_expr e1)) (Why.Deref n)
	    | _ -> unsupported e.expr_loc ("tr_expr("^op^")")
	end
    | Assign_field(fa, op, e1) ->
	begin
	  match fa with
	    | Super_access(id) -> 
		unsupported e.expr_loc "tr_expr(assign_field_super)"
	    | Primary_access(f,_) ->
		ill_typed "tr_expr(assign_field_primary)"
	    | Typed_constant_access(id) -> 
		ill_typed "tr_expr(assign_constant)"
	    | Typed_primary_access(f, ive) ->
	        tr_field_assignment e.expr_loc f ive false (tr_expr e1) false 
	    | Array_length(e2) -> 
		ill_typed "tr_expr(assign_field_array_length)"
	end
    | Assign_array(e1,e2,op,e3) -> ill_typed "tr_expr(assign_array)"
    | Typed_assign_array(t,e1,e2,op,e3) -> 
	unsupported e.expr_loc "tr_expr(array_assign)"
    | If_expr(e1,e2,e3) -> 
	Why.If(tr_abstract_expr e1,tr_expr e2,tr_expr e3)
    | Field_access(fa) -> tr_field_access e.expr_loc fa
    | Method_call _ -> ill_typed "tr_expr(method_call)"
    | Super_method_call _ -> ill_typed "tr_expr(super_method_call)"
    | Typed_method_call(e,me,args) ->
	tr_method_call e me args
(*
	  (List.combine 
	     (List.map 
		(fun (_,t) -> 
		   match t with
		     | Local_variable_entry le -> 
			 le.local_variable_entry_type
		     | _ -> assert false) 
		me.method_entry_parameters)
	     args)
*)
    | Logic_call _ -> assert false
    | Array_creation(a) -> tr_array_creation e.expr_loc a
    | Array_access(_) -> ill_typed "tr_expr(array_access)"
    | Typed_array_access(t, e1, e2) -> 	
	let alloc_name, access_fun = array_access_funs t in
	if is_pure e1 && is_pure e2 then
	  let we1 = tr_term "" (fun id -> Why.LVar(id)) e1 in
	  let pre =
	    Why.make_and
	      (Why.LPred("neq", [we1; Why.LVar "Null"]))
              (if !Config.nonnull_only then Why.LTrue else 
	      if !Config.unfold_pred >= 1
	      then
                  Why.LAnd(
                  Why.LPred("le_int", [Why.LConst(Why.Prim_int "0"); tr_term "" (fun id -> Why.LVar(id)) e2]),
                  Why.LPred("lt_int", [tr_term "" (fun id -> Why.LVar(id)) e2; Why.LApp("arraylength", [Why.LVar "alloc"; we1])]))
		else
      	          Why.LPred
		    ("is_valid_index",
		     [Why.LVar "alloc";
		      we1;
		      tr_term "" (fun id -> Why.LVar(id)) e2]))
	  and acc = 
	    Why.make_app access_fun
	      [Why.Deref(alloc_name);
	       tr_expr e1;
	       tr_expr e2]
	  in
	  Why.make_pre pre acc
	else
	  why_let_var (is_pure e1) "krak_access"
	  (fun whye1 loge1 ->
	     why_let_var true (temp_index ())
	       (fun whye2 loge2 ->
		  let pre = 
		    Why.make_and
	              (Why.LPred("neq", [loge1; Why.LVar "Null"]))
                      (if !Config.nonnull_only then Why.LTrue else 
		      if !Config.unfold_pred >= 1
			 then
                           Why.LAnd(
                           Why.LPred("le_int", [Why.LConst(Why.Prim_int "0"); loge2]),
                           Why.LPred("lt_int", [loge2; Why.LApp("arraylength", [Why.LVar "alloc"; loge1])]))
			 else
		           Why.LPred
	                     ("is_valid_index",
		            [Why.LVar "alloc"; loge1; loge2]))
		  and acc = 
		    Why.make_app access_fun
		      [Why.Deref(alloc_name);
		       whye1;
		       whye2] 
		  in 
		    Why.make_pre pre ((* coercion_expr t *) acc)
	       )
	       (tr_expr e2))
	  (tr_expr e1) 
    | Cast(t,e) -> ill_typed "tr_expr(cast)"
    | Typed_cast(t,t',e) -> tr_cast t t' e
    | Class_instance_creation(_,_) -> 
	ill_typed "tr_expr(new class)"
    | Typed_class_instance_creation(ce,args) -> 
	(* vieille version :
           let new_val = (new_obj alloc classid) in
           alloc := (store_new_ref alloc new_val);
           (constr_classId new_val args)
           nouvelle version:
             (new_classId args)
        *)
(*i
	let new_val = "krak_new" in
	Why.Let
	  (new_val,
	   Why.make_app "new_obj"
	     [Why.Deref("alloc");
	      Why.Var(ce.constructor_entry_class.class_entry_name)],
	   Why.Block
	     ([Why.Assign
		 ("alloc",
		  Why.make_app "store_new_ref"
		    [Why.Deref("alloc");
		     Why.Var(new_val)]);
	       build_why_app
		 (Why.Var(ce.constructor_entry_trans_name ^ "_parameter"))
		 (Why.Var(new_val)::(List.map tr_expr args));
	       Why.Var(new_val)]))
i*)
	build_why_app
	  (Why.Var(ce.constructor_entry_trans_name ^ "_parameter"))
		 (List.map tr_expr args);
    | Instanceof(e,t) -> ill_typed "tr_expr(instanceof)"
    | Typed_typeof(e,t) -> ill_typed "tr_expr(typef_typeof)"
    | Typeof(e) -> ill_typed "tr_expr(typeof)"
    | Typed_type(t) -> ill_typed "tr_expr(typed_type)"
    | Type(t) -> ill_typed "tr_expr(type)"
    | Typed_instanceof(e1,t) -> 
	Why.make_app "_instanceof" [Why.Deref "alloc" ; tr_expr e1; tr_type_expr t]
    | Result -> assert false (* should never happen *)
    | Old(_) -> assert false (* should never happen *)
    | Fresh(_) -> assert false (* should never happen *)
    | Not_conditionally_updated(_) -> assert false (* should neveer happen *)
    | Fpi(_) -> assert false (* should never happen *)
    | Quantifier(_,_,_,_,_) -> assert false (* should never happen *)
    | Typed_quantifier(_,_,_,_) -> assert false (* should never happen *)


and tr_array_creation loc a =
  match a with
    | Explicit_array_creation(e,Typed_implicit_array_creation(t)) ->
	begin
	  match t with
	    | Byte_type | Short_type | Integer_type ->
		Why.make_app
		  ("alloc_new_int_array" ^ if !Config.nonnull_only then "_nonnull_only" else "" ^ if !Config.javacard then "_javacard" else "")
		  [tr_expr e]
	    | Float_type ->
		Why.make_app
		  "alloc_new_float_array"
		  [tr_expr e]
	    | Boolean_type ->
		Why.make_app
		  ("alloc_new_bool_array" ^ if !Config.nonnull_only then "_nonnull_only" else "" ^ if !Config.javacard then "_javacard" else "")
		  [tr_expr e]
	    | _ ->
		Why.make_app
		  ("alloc_new_obj_array" ^ if !Config.nonnull_only then "_nonnull_only" else "" ^ if !Config.javacard then "_javacard" else "")
		  [tr_type_expr t;tr_expr e]
	end

    | Explicit_array_creation(e,a) ->
	unsupported loc "tr_array_creation(multidimensional)";
	(*  il faudrait
	    let length = e in
	    let new_val = new_array alloc length type_des_cellules in	      
	    alloc := (store_new_ref alloc new_val);
	    for i=0 to length-1 do alloc := update alloc new_val i a
	*)
	(*i
	Why.Let
	  ("length",
	   tr_expr e,
	   Why.Let
	     ("new_val", 
	      (Why.make_app "new_array"
	  [Why.Deref("alloc")),
		  Why.Var("length")],
	      Why.Let
		("new_vector", 
		 (Why.make_app
		    (Why.App
		       (Why.Var("new_array_map"),
			Why.Var("length")),
		       tr_array_creation loc a)),
		 (Why.Block
		    [Why.Assign
		       ("alloc",
			Why.make_app
			  (Why.App
			     (Why.App(Why.Var("store_add_value"),
				      Why.Deref("alloc")),
			      Why.Var("new_val")),
			     Why.Var("new_vector")));
		     Why.Var("new_val")]))))
	  i*)
    | Implicit_array_creation(_) -> ill_typed "tr_array_creation(implicit)"
    | Typed_implicit_array_creation(t) -> type_default_value true t

(* [tr_abstract_expr e] translates expression [e] and adds post-condition
   [result = e] when it is pure (or [if result then e else not e] when
   [e] is a boolean), if option -bb is set *)
and tr_abstract_expr e =
(*
  let no_effect e =
    let eff = Effects.expr e in 
    (not eff.writes_alloc) && (Inttagset.is_empty eff.writes) 
    && (not eff.throws) && (not eff.breaks)
  in
*)
  (* [pure_expr e] checks if [e] is pure and returns its type;
     raises [Impure] is [e] is not pure 
     (note: we know that [e] has already been typed) *)
  let rec pure_expr e = match e.expr_node with
    | Lit(Int _ | Char _) -> Integer_type
    | Lit(Float _) -> Float_type
    | Lit(Bool _) -> Boolean_type
    | Un(op,e) -> 
	let t = pure_expr e in
	(match op with
	   | "float_of_int" -> Float_type
	   | "neg_int" | "neg_float" -> t
	   | "!" -> Boolean_type
	   | "lcompl" -> Integer_type
	   | _ -> raise Impure)
    | Bin(e1,op,e2) ->
	ignore (pure_expr e1);
	ignore (pure_expr e2);
	(match op with
	   |"le_int"|"ge_int"|"gt_int"|"lt_int"
	   |"le_float"|"ge_float"|"gt_float"|"lt_float"
	   |"eq_int"|"ne_int"|"eq_bool"|"ne_bool"|"andb"|"orb"
	   |"eq_float"|"ne_float"
	   |"&&"|"||" -> Boolean_type
	   | "add_int"|"sub_int"|"mul_int"|"div_int_"|"mod_int_"
	   |"Zland"|"Zlor"|"Zasr"|"Zlsl"|"Zlsr" -> Integer_type
	   | "add_float"|"sub_float"|"mul_float"|"div_float" -> Float_type
	   | _ -> raise Impure)
    | Typed_var v -> v.local_variable_entry_type
    | Field_access (Typed_primary_access ({expr_node=This},f)) -> 
	f.instance_variable_entry_type
    | Field_access (Typed_primary_access (e,f)) -> 
	ignore (pure_expr e); 
	f.instance_variable_entry_type
    | Typed_method_call (e, me, el) when me.method_entry_is_pure ->
	(match e.expr_node with
	   | Static_class _ | Static_interface _ -> ()
	   | _ -> ignore (pure_expr e));
	List.iter (fun e -> ignore (pure_expr e)) el;
	me.method_entry_return_type
    | _ -> 
	raise Impure
  in
  tr_expr e

let tr_variable_id loc id =
  match id with
    | Simple_id s -> snd s
    | Array_id _ -> unsupported loc "tr_variable_id(array)"
;;

let tr_variable_initializer loc i =
  match i with
    | Simple_initializer(e) -> tr_abstract_expr e
    | Array_initializer _ -> unsupported loc "tr_variable_initializer(array)"
;;

let tr_local_var_declarator loc t v accu =
  let init =
    match v.variable_initializer with
      | None -> type_default_value false t
      | Some e -> tr_variable_initializer loc e
  in
  Why.Let_ref(tr_variable_id loc v.variable_id,init,accu)

;;

let tr_local_var_declaration loc lve init acc =
  let init =
    match init with
      | None -> type_default_value false lve.local_variable_entry_type
      | Some e -> tr_variable_initializer loc e
  in
  Why.Let_ref(lve.local_variable_entry_trans_name,init,acc)
;;

(*

  [abrupt_termination s] is [true] if [s] always terminate abruptly,
  by a return, a break, a continue or a throw statement. It returns
  [false] if it always terminate nromally, and raises [Not_found] if
  both are possible.

*)


let rec abrupt_termination s =
  abrupt_termination_bis s true

and abrupt_termination_bis s break =
  match s.statement_node with
    | Skip | Expr(_) | Typed_expr(_) -> false
    | Var_decl(v) -> false
    | Typed_var_decl(v,_,i) -> abrupt_termination_bis i break
    | Return(e) -> true
    | Typed_return(ty,e) -> true
    | Break(l) -> break (* NR - break = true <==> should be interpreted as an abrupt_termination *)
    | Continue(l) -> true
    | Label(l,s) -> abrupt_termination_bis s break
    | If_statement(e, s1, s2) ->
	let c1 = abrupt_termination_bis s1 false
	and c2 = abrupt_termination_bis s2 false
	in if c1 = c2 then c1 else raise Not_found
    | While(loop_inv,e,s) -> abrupt_termination_bis s false
    | For(_,_,e,el2,s) 
    | For_decl(_,_,e,el2,s) 
    | For_typed_decl(_,_,_,e,el2,s) -> 
	abrupt_termination_bis s false
    | Switch(e,cases) -> 
	List.for_all (fun (_,b) -> block_abrupt_termination b) cases
    | Block(s) -> block_abrupt_termination s
    | Synchronized (e, _) -> 
	unsupported e.expr_loc "abrupt_termination(synchronized)"
    | Try (_, _, _) -> assert false
    | Typed_try (b, catches, bopt) -> raise Not_found 
    | Do (inv,s,_) -> abrupt_termination_bis s false
    | Throw(_) ->  true
    | Jml_assert _ -> false
    | Jml_annot_statement _ -> false
    | Jml_ghost_var_decl _ -> false
    | Annotated _ -> assert false
    | Typed_annotated(_,st) -> abrupt_termination_bis st break
	
and block_abrupt_termination s =
  match s with
    | [] -> false
(*    | { statement_node = If_statement(_,s,{ statement_node = Skip})} :: l -> 
	abrupt_termination s || block_abrupt_termination l *)
    | i::l ->
	let x = 
	  try Some (abrupt_termination_bis i false)
	  with Not_found -> None
	in 
	match x with
	  | (Some x) -> if x then true else block_abrupt_termination l
	  | None ->
	      if block_abrupt_termination l
	      then true
	      else raise Not_found
	
;;


let rec chop_last_elem l =
  match l with
    | [] -> assert false
    | [x] -> [], x
    | h::l -> let r, x = chop_last_elem l in (h::r), x
;;

let rec make_case_expr l =
  match l with
    | [] -> assert false
    | [Case c] ->
	Why.make_app "eq_int_" [Why.Var ("temp_switch" ^ (string_of_int !temp_switch_count));
			       tr_expr c]
    | (Case c)::l -> 
	Why.Or
	  (Why.make_app "eq_int_" [Why.Var ("temp_switch" ^ (string_of_int !temp_switch_count));
				  tr_expr c],
	   make_case_expr l)
    | _ -> assert false
;;




let tr_signals_clause assign s =
  match s with
    | None -> []
    | Some(t,var,e) ->
	assert (match var with 
		  | None -> Format.eprintf "None!@.";false
		  | Some v -> 
		      (v.local_variable_entry_trans_name = "result") ||
		      (Format.eprintf "Some %s@." v.local_variable_entry_trans_name;
		       false));
	[(tr_type_detailed t, Why.make_and assign (tr_assertion "" e))]
;;


let tr_cardtear_clause c =
  match c with 
  | None -> []
  | Some(e) -> [("CardTear",tr_assertion "" e)]
;;


let make_inter_loc l1 l2 =
  if l1 = l2 then l1 
  else
    match l1,l2 with
      | (Why.LVar("nothing_loc"), _) -> l2
      | (_, Why.LVar("nothing_loc")) -> l1
      | _, _ -> Why.LApp("inter_loc", [l1; l2])
;;

let make_array_inter_loc l1 l2 =
  match l1,l2 with
    | (Why.LVar("array_nothing_loc"),_) -> l2
    | (_,Why.LVar("array_nothing_loc")) -> l1
    | _,_ -> Why.LApp("array_inter_loc",[l1;l2])
;;

type assignable_struct =
  {
    assignable_fields : 
      (Java_context_types.instance_variable_entry * 
       Ast_types.expr list) list;
    assignable_int_arrays : Ast_types.assignable_object list;
    assignable_float_arrays : Ast_types.assignable_object list;
    assignable_bool_arrays : Ast_types.assignable_object list;
    assignable_obj_arrays : Ast_types.assignable_object list;
  }

let rec add_field f sr fields =
  match fields with
    | [] -> [(f,[sr])]
    | ((g,srl) as h)::rem ->
	if f == g
	then (g,sr::srl)::rem
	else h::(add_field f sr rem)
;;
 
let tr_clause modif loc clause =
  match clause with
    | Assignable_everything ->
	modif (* unsupported loc "tr_assignable(everything)" *)
    | Assignable_nothing ->
	modif (* unsupported loc "tr_assignable(nothing)" *)
    | Assignable_set_ref(e,_) -> assert false
    | Typed_assignable_set_ref(t,e,[sr]) ->
	begin
(*	  match e.expr_node with
	    | Static_class(ce) -> 
		unsupported loc "tr_assignable(static field)"
	    | _ ->
*)
	  match sr with 
	    | Set_field(ive) -> 
                if ive.instance_variable_entry_is_static 
                then modif
                else
		  { modif with
		      assignable_fields =
                        add_field ive e modif.assignable_fields }
	    | Set_array_index _ | Set_array_interval _ | Set_array ->      
		begin
		  match t with
		    | Array_type Byte_type | Array_type Short_type | Array_type Integer_type ->
			{ modif with
			    assignable_int_arrays =
			      clause::modif.assignable_int_arrays }
		    | Array_type Float_type ->
			{ modif with
			    assignable_float_arrays =
			      clause::modif.assignable_float_arrays }
		    | Array_type Boolean_type -> 
			{ modif with
			    assignable_bool_arrays =
			      clause::modif.assignable_bool_arrays }
		    | Array_type _ ->
			{ modif with
			    assignable_obj_arrays =
			      clause::modif.assignable_obj_arrays }
		    | Null_type -> assert false
		    | Type_type -> assert false
		    | Byte_type -> assert false
		    | Short_type -> assert false
		    | Integer_type -> assert false
		    | Boolean_type -> assert false
		    | Float_type -> assert false
		    | Real_type -> assert false
		    | Class_type _ -> assert false
		    | Interface_type _ -> assert false
		    | Prop_type | Type_var _ | Logic_type _ -> assert false
                end
	    | Set_fieldraw(_) -> assert false
	end
    | Typed_assignable_set_ref(_,e,[]) -> assert false
    | Typed_assignable_set_ref(_,e,lv) -> 
	unsupported loc "tr_assignable(set_ref_not_singleton)"
    | Assignable_all_fields(e) ->                      
	unsupported e.expr_loc "tr_assignable(all fields)"
;;

	  
let rec make_inter_locs oldify el for_constructor =
  match el with
    | [] -> Why.LVar("nothing_loc")
    | e::rem ->
	begin
	  match e.expr_node with
	    | Typed_var(lve) when 
		lve.local_variable_entry_trans_name = "this" -> 
		  if for_constructor then
		    make_inter_locs oldify rem for_constructor
		  else 
		    make_inter_loc 
		      (Why.LApp("value_loc",[tr_term "?" oldify e])) 
	              (make_inter_locs oldify rem for_constructor)
	    | _ ->
		make_inter_loc 
		  (Why.LApp("value_loc",[tr_term "?" oldify e])) 
		  (make_inter_locs oldify rem for_constructor)
	end;;

let make_modif_field_nothing oldify f for_constructor acc =
  let mod_pred = "modifiable" in
  let name = f.instance_variable_entry_trans_name in
  let locs = Why.LVar "nothing_loc" in
  Why.LPred(mod_pred,
	    [oldify "alloc"; 
	     oldify name; 
	     Why.LVar name;
	     locs])::acc
;;

let make_modif_field_pred oldify writes_nothing_fields f el for_constructor =
  let mod_pred = "modifiable"
		   (*
    match f.instance_variable_entry_type with
      | Byte_type | Short_type | Integer_type -> "modifiable_int"
      | Boolean_type -> "modifiable_bool"
      | Class_type _ -> "modifiable_obj"
      | Array_type _ -> "modifiable_obj"
      | _ -> assert false
*)
  in
  let name = f.instance_variable_entry_trans_name in
  let locs =
    if Inttagset.mem f.instance_variable_entry_tag writes_nothing_fields
    then Why.LVar "nothing_loc"
    else make_inter_locs oldify el for_constructor
  in
  Why.LPred(mod_pred,
	    [oldify "alloc"; 
	     oldify name; 
	     Why.LVar name;
	     locs])
;;

let tr_array_loc oldify sr =
  match sr with 
    | Typed_assignable_set_ref(t,v,[sr]) ->
	begin
	  match sr with
	  | Set_field(f) -> assert false
	  | Set_array_index(e) ->           
	      Why.LApp
		("array_loc",
		 [tr_term "?" oldify v; tr_term "?" oldify e])
	  | Set_array_interval(e1,e2) ->
	      Why.LApp
		("array_sub_loc",
		 [tr_term "?" oldify v;
		  tr_term "?" oldify e1;
		  tr_term "?" oldify e2])
	  | Set_array ->                      
	      Why.LApp("array_all_loc",[tr_term "?" oldify v])
	  | Set_fieldraw(_) -> assert false
	end
| _ -> assert false

;;

let rec make_inter_array_locs oldify srl =
  match srl with
    | [] -> Why.LVar("array_nothing_loc")
    | sr::rem ->
	make_array_inter_loc (tr_array_loc oldify sr) (make_inter_array_locs oldify rem)
;;


let make_modif_array_pred oldify writes_nothing srl mod_pred mod_var =
  if writes_nothing
  then
    begin
      assert (srl=[]);
      Why.LPred(mod_pred,
		[oldify "alloc";
		 oldify mod_var; 
		 Why.LVar mod_var;
		 Why.LVar "array_nothing_loc"])
    end
  else    
    if srl = [] then Why.LTrue
    else
      Why.LPred(mod_pred,
		[oldify "alloc";
		 oldify mod_var; 
		 Why.LVar mod_var;
		 make_inter_array_locs oldify srl])
;;

let make_table_extends_pred oldify eff =
  if eff.writes_alloc_table 
  then
    Why.LPred("store_extends",
	      [oldify "alloc"; Why.LVar("alloc")])
  else Why.LTrue
  
       

let tr_assignable loc oldify (assign : Ast_types.assignable_clause) routine for_constructor =
  assert (assign <> []);
  let modif =
    List.fold_left
      (fun accu (clause,guard) ->
	 match guard with
	   | None -> tr_clause accu loc clause 
	   | Some e -> unsupported loc "tr_assignable(if)"
      )
      {
	assignable_fields = [];
	assignable_int_arrays = [];
	assignable_float_arrays = [];
	assignable_bool_arrays = [];
	assignable_obj_arrays = [];
      }
      assign
  in
  Why.make_and_list
    ([ make_table_extends_pred oldify routine.routine_entry_effects;
       make_modif_array_pred oldify 
	 routine.routine_entry_int_array_writes_nothing
	 modif.assignable_int_arrays "array_modifiable" (*_int*) "intA";
       make_modif_array_pred oldify 
	 routine.routine_entry_float_array_writes_nothing
	 modif.assignable_float_arrays "array_modifiable" (*_float*) "floatA";
       make_modif_array_pred oldify
	 routine.routine_entry_bool_array_writes_nothing
	 modif.assignable_bool_arrays "array_modifiable" (*_bool*) "boolA";
       make_modif_array_pred oldify
	 routine.routine_entry_obj_array_writes_nothing
	 modif.assignable_obj_arrays "array_modifiable" (*_obj*) "objA"]
     @
     (List.map
        (fun (f,srl) -> 
	   make_modif_field_pred oldify 
	     routine.routine_entry_fields_writes_nothing
	     f srl for_constructor)
	modif.assignable_fields)
     @
     (Inttagset.fold
        (fun f acc -> 
	   make_modif_field_nothing oldify 
	     f for_constructor acc)
	routine.routine_entry_fields_writes_nothing []))
;;

let tr_loop_inv label loop_spec loc =
  let inv = loop_spec.loop_invariant
  and modif = 
    match loop_spec.typed_loop_modifies with
      | None -> !ref_modifiable
      | Some clause ->
	  (Why.make_and
	     (tr_assignable loc 
		(fun id -> Why.LVarAtLabel(id,label)) 
		clause Java_context.default_routine false) 
	     !ref_modifiable)
  and dec = loop_spec.loop_decreases
  in
  let i = Why.make_and (tr_assertion "init" inv) modif    
  and d = tr_term "init" (fun id -> Why.LVar(id)) dec
  in (i,d) 

let tr_assign_op loc op =
match op with
  | "add_int" | "sub_int" | "mul_int" | "div_int_" | "mod_int_" -> op
  | "add_float" | "sub_float" | "mul_float" | "div_float" -> op
  | "land" -> "Zland"
  | "lor" | "Zlor" -> "Zlor"
  | "lsr" | "Zlsr" -> "Zlsr"
  | "lsl" -> "Zlsl"
  | "asr" -> "Zasr"
  | "andb" -> op
  | _ -> unsupported loc ("tr_assign_op("^op^")")
;;

let label_counter = ref 0;;

let ab_ret = ref None;;

let rec well_formed_switch cases model_ensures =
  match cases with
    | [] -> assert false
    | (case_list, b)::l ->
	if List.mem Default case_list 
	then
	  if l <> [] 
          then
            let only_cases = List.map fst l in
	    let c = 
              Why.Not 
                (List.fold_right 
                   (fun cl acc -> 
                      Why.make_or_expr 
                        (make_case_expr cl) acc) 
                   only_cases 
                   (Why.Cte (Why.Prim_bool false))) in
            let (p', l') = well_formed_switch l model_ensures in
            (None::p', (c, b)::l')
          else  
            ([None], [(Why.Cte (Why.Prim_bool true), b)])
        else
	  let c = make_case_expr case_list in
	  if l = []
          then
            ([None], [(c, b)])
	  else
	    let (hb, tb) = chop_last_elem b in
            try
	      if abrupt_termination tb
	      then 
		let (p', l') = well_formed_switch l model_ensures in 
		(None::p', (c, b)::l')
	      else  
		let p = Why.If(c, tr_block true b model_ensures, Why.Void) in
                let (p', l') = well_formed_switch ((List.append case_list (fst (List.hd l)), snd (List.hd l))::(List.tl l)) model_ensures in 
		((Some p)::p', (c, b)::l')
	    with
	      Not_found -> 
		let p = Why.If(c, tr_block true b  model_ensures, Why.Void) in
                let (p', l') = well_formed_switch ((List.append case_list (fst (List.hd l)), snd (List.hd l))::(List.tl l)) model_ensures 
		in 
		  ((Some p)::p', (c, b)::l')

and is_breakable cases =
  match cases with
    | [] -> false
    | (_, b)::tl ->
       let (_, tb) = chop_last_elem b in
       if tb.statement_node = Break("") then true else is_breakable tl


(*
let rec well_formed_switch cases model_ensures =
  match cases with
    | [] -> assert false
    | (case_list, b)::l ->
	if List.mem Default case_list 
	then
	  begin
	    if l <> [] then Format.printf "WARNING : unreachable statement, well_formed_switch@.";
              let (hb, tb) = chop_last_elem b in
              if tb.statement_node = Break("")
              then ([], [], hb)
              else ([], [], b) 
	  end
        else
	  let e = make_case_expr case_list in
	  let (hb, tb) = chop_last_elem b in
	  if l = []
          then
	    if tb.statement_node = Break("") 
	    then ([None], [(e, hb)], []) 
	    else ([None], [(e, b)], [])
	  else
	    if tb.statement_node = Break("") 
	    then 
	      let (p', l',b') = well_formed_switch l model_ensures in 
	      (None::p', (e,hb)::l', b')
	    else 
	      try
		if abrupt_termination tb
		then 
		  let (p', l', b') = well_formed_switch l model_ensures in 
		  (None::p', (e, b)::l', b')
		else  
		  let p = Why.If(e, tr_block true b model_ensures, Why.Void) in
                  let (p', l', b') = well_formed_switch ((List.append case_list (fst (List.hd l)), snd (List.hd l))::(List.tl l)) model_ensures
		  in 
		  ((Some p)::p', (e, b)::l', b')
	      with
		  Not_found -> 
		    let p = Why.If(e, tr_block true b  model_ensures, Why.Void) in
                    let (p', l', b') = well_formed_switch ((List.append case_list (fst (List.hd l)), snd (List.hd l))::(List.tl l)) model_ensures 
		    in 
		    ((Some p)::p', (e, b)::l', b')
*)

and tr_statement abrupt_return s model_ensures = 
(*  Format.printf "Entering tr_statement %b@." abrupt_return; *)
  tempvar_count := 0;
  tempindex_count := -1;
  Format.fprintf Config.log
    "%s: translating statement.@." (Location.to_string s.statement_loc);
  match s.statement_node with
    | Skip -> Why.Void
    | Expr(e) -> ill_typed "tr_statement(expr)"
    | Typed_expr(t,e) -> 
	begin
	  match e.expr_node with
	    | Lit(_) -> ill_typed "tr_statement(lit)"
	    | Bin(_,_,_) ->  ill_typed "tr_statement(binary)"
	    | Un(op,e) -> ill_typed "tr_statement(unary)"
	    | PrePostIncrDecr(op,e) ->
		begin
		  match op with
		    | `PREINCR | `POSTINCR -> 
			tr_pre_post_incr_decr "add_int" e
		    | `PREDECR | `POSTDECR -> 
			tr_pre_post_incr_decr "sub_int" e 
		end
	    | Var(_) -> ill_typed "tr_statement(var)"
	    | Typed_var(lve) -> ill_typed "tr_statement(typed_var)"
	    | Static_class(_) -> assert false (* ????????? *)
	    | Static_interface(_) -> assert false (* ????????? *)
	    | This -> ill_typed "tr_statement(this)"
	    | Assign_name(_,_,_) -> ill_typed "tr_statement(assign_name)"
	    | Assign_local_var(lve, op, e1) ->
		begin
		  match op with
		    | "=" ->
                        if !Config.cardtear
                        then
                          Why.append (Why.Assign(lve.local_variable_entry_trans_name,tr_abstract_expr e1)) (build_app (Why.Var("possibleCardTear")) [])
                        else
			  Why.Assign(lve.local_variable_entry_trans_name,tr_abstract_expr e1)
		    | _ ->
			Why.Assign(lve.local_variable_entry_trans_name,
				   Why.make_app (tr_assign_op e1.expr_loc op)
				     [Why.Deref(lve.local_variable_entry_trans_name);
				      tr_expr e1])
		end
	    | Assign_field(fa,op,e1) ->
		begin
		  match fa with
		    | Super_access(id) -> 
			unsupported e.expr_loc "assign_field(super)"
		    | Primary_access(f,(loc,id)) ->	
			ill_typed "assign_field(primary)"
		    | Typed_constant_access _ ->
			ill_typed "assign_field(constant)"			
		    | Typed_primary_access(f, ive) ->	
			begin
			  match op with
			    | "=" ->
                              if !Config.javacard
                              then
                                if !Config.transactionException
                                then
                                  if !Config.cardtear
                                  then
                                    Why.append (Why.App(Why.App(Why.App(Why.Var "upd_transaction", Why.Var ive.instance_variable_entry_trans_name), 
                                                                tr_expr f), 
                                                                tr_expr e1))
                                               (build_app (Why.Var("possibleCardTear")) [])
                                  else
				    if ive.instance_variable_entry_is_static
                                    then
                                      Why.make_app 
				        "upd_transaction_static" 
					[Why.Var ive.instance_variable_entry_trans_name; tr_expr e1]
				    else
                                      Why.make_app
					"upd_transaction"
					[Why.Var ive.instance_variable_entry_trans_name; tr_expr f; tr_expr e1]
                                else
                                  if !Config.cardtear
                                  then
                                    Why.append 
                                      (tr_field_assignment e.expr_loc f ive (is_pure e1) (tr_abstract_expr e1) true)
                                      (build_app (Why.Var("possibleCardTear")) [])
                                  else
       			            tr_field_assignment 
				        e.expr_loc f ive (is_pure e1) (tr_abstract_expr e1) true
                              else
       			              tr_field_assignment 
				        e.expr_loc f ive (is_pure e1) (tr_abstract_expr e1) true
			    | _ -> 
				tr_field_assignment_op
				  e.expr_loc (tr_expr f) ive op (tr_expr e1)
			end
		    | Array_length(_) -> ill_typed "tr_statement(array_length)"
		end
	    | Assign_array(e1,e2,op,e3) -> 
		ill_typed "tr_statement(assign_array)"
	    | Typed_assign_array(t,e1,e2,op,e3) -> 
		begin
		  match op with
		  | "=" ->
		      let we = tr_array_assignment t e1 e2 (is_pure e3) (tr_expr e3) in
                      let we_non_atomic =
			match t with 
			| Byte_type | Short_type | Integer_type -> 
			    build_app (Why.Var ("int_array_upd_javacard" ^ if !Config.unfold_pred >= 1 then "_unfold_pred" else "")) [e1; e2; e3]
			| _ -> we
                      in   
                      let we_transaction_exception =
			match t with 
			| Byte_type | Short_type | Integer_type -> build_app (Why.Var "int_array_upd_exception") [e1; e2; e3]
			| Boolean_type -> build_app (Why.Var "bool_array_upd_exception") [e1; e2; e3]
			| Class_type _ | Interface_type _ -> build_app (Why.Var "obj_array_upd_exception") [e1; e2; e3]
			| _ -> assert false (* should not happen *)
                      in
		      let we = if !Config.javacard then we_non_atomic else we in
		      let we = if !Config.transactionException then we_transaction_exception else we in
                      let we = if !Config.cardtear then Why.append we (build_app (Why.Var("possibleCardTear")) []) else we in
		      we
                  | "add_int" | "Zland" -> 
                      let e4 = mk_expr e.expr_loc (Typed_array_access(t, e1, e2)) in
                      tr_array_assignment t e1 e2 (is_pure e4 && is_pure e3) (Why.make_app op [tr_expr e4; tr_expr e3])
		  | _ -> 
		      unsupported e.expr_loc 
			("tr_statement(assign_array("^op^"))")
		end			
	    | If_expr(e1,e2,e3) -> ill_typed "tr_statement(if_expr)"
	    | Field_access(fa) -> ill_typed "tr_statement(field_access)"
(*
	    | Method_call(Static_class(ce),(_,id),args) ->
		build_why_app 
		  (Why.Var("static_method_" ^ (snd ce.class_entry_name) ^ "_" ^ id)) args
	    | Method_call(e,(_,id),args) ->
		Why.Let
		  ("temp_for_method_call",tr_expr e,
		   build_why_app 
		     (Why.make_app
			(Why.Var("method_" ^ id ^ "_parameter"),
			 Why.Var("temp_for_method_call")))
		     args)
*)
	    | Method_call(e,(_,id),args) -> 
		ill_typed "tr_statement(method_call)"
	    | Super_method_call _ -> 
		ill_typed "tr_statement(super_method_call)"
	    | Typed_method_call(e,me,args) -> 
		    let tre = tr_method_call e me args
		    in
		    if t = Null_type
		    then tre
		    else Why.Let("krak_ignore",tre,Why.Void)
	    | Logic_call _ -> assert false
	    | Array_creation(a) -> ill_typed "tr_statement(array_creation)"
	    | Array_access(e1,e2) -> ill_typed "tr_statement(array_access)"
	    | Typed_array_access(t, e1,e2) -> ill_typed "tr_statement(typed_array_access)"
	    | Cast(t,e) -> ill_typed "tr_statement(cast)"
	    | Typed_cast(t,_,e) -> ill_typed "tr_statement(typed_cast)"
	    | Class_instance_creation(_,_) -> ill_typed "tr_statement(new_object)"
	    | Typed_class_instance_creation(ce, args) -> 
	        build_why_app
	          (Why.Var(ce.constructor_entry_trans_name ^ "_parameter"))
		  (List.map tr_expr args)
	    | Instanceof(e,t) -> ill_typed "tr_statement(instanceof)"
	    | Typed_instanceof(e,t) -> ill_typed "tr_statement(instanceof result)"
	    | Typed_typeof(e,t) -> ill_typed "tr_statement(typef_typeof)"
	    | Typeof(e) -> ill_typed "tr_statement(typeof)"
	    | Typed_type(t) -> ill_typed "tr_statement(typed_type)"
	    | Type(t) -> ill_typed "tr_statement(type)"
	    | Result -> assert false (* should never happen *)
	    | Old(_) -> assert false (* should never happen *)
	    | Fresh(_) -> assert false (* should never happen *)
	    | Not_conditionally_updated(_) -> assert false (* should never happen *)
	    | Fpi(_) -> assert false (* should never happen *)
	    | Quantifier(_) -> assert false (* should never happen *)
	    | Typed_quantifier(_) -> assert false (* should never happen *)
	end
    | Var_decl(v) -> ill_typed "tr_statement(var_decl)"
    | Typed_var_decl(lve, init, i) ->
	let init =
	  match init with
	  | None -> type_default_value false lve.local_variable_entry_type
	  | Some e -> tr_variable_initializer Location.dummy e
	in
	Why.Let_ref(lve.local_variable_entry_trans_name, init, tr_statement abrupt_return i model_ensures)
    | Return _ -> ill_typed "tr_statement(return)"
    | Typed_return(ty,e) -> 
	begin
	  if true (* abrupt_return *) then
	    match e with
	      | None -> 
		  ab_ret := Some "Return_void";
		  Why.Raise("Return_void",Some Why.Void)
	      | Some e -> 
		  let ex =
		    match ty with
		      | Byte_type | Short_type | Integer_type -> "Return_int"
		      | Boolean_type -> "Return_bool"
		      | Float_type -> "Return_float"
		      | _ -> "Return_value"
		  in
		  ab_ret := Some ex;
		  Why.Raise(ex,Some (tr_abstract_expr e))
	  else
            Why.Block([model_ensures;
	    match e with
	      | None -> Why.Void (* return; *) 
	      | Some e -> tr_abstract_expr e ])
(*
            if rep_inv <> Why.Void 
            then
              Why.Block([rep_inv;
	        match e with
	        | None -> Why.Void (* return; *) 
	        | Some e -> tr_abstract_expr e ])
            else
              begin
              match e with
	        | None -> Why.Void (* return; *) 
	        | Some e -> tr_abstract_expr e 
              end
*)
	end
    | Break(l) -> 
	if l="" then
	  Why.Raise("Break", None)
	else
	  unsupported Location.dummy "tr_statement(break label)"
	    (*i Why.Raise("Break",Why.Var("krak_label_" ^ l)) i*)
    | Continue(l) ->
        if l="" then
          Why.Raise("Continue", Some why_zero)
        else 
          unsupported Location.dummy "tr_statement(continue label)"
    | Label(l,s) -> unsupported Location.dummy "tr_statement(label)"
    | If_statement(e,s1,s2) ->
	Why.If(tr_abstract_expr e,
	       tr_statement abrupt_return s1 model_ensures,
	       tr_statement abrupt_return s2 model_ensures)
    | While(l, e, s) ->
	incr label_counter;
	let new_label = "krak_loop" ^ (string_of_int !label_counter) in
	let (inv,var) = tr_loop_inv new_label l e.expr_loc in
	let eff = Effects.statement s in
        let body = tr_statement true s model_ensures in
        let body =
          if eff.continue then
            Why.Try(body, "Continue", Some "krak_label", Why.Void)
          else
            body  
        in
	let why_var =
	  if !Config.nonnull_only then None else (Some (var, None))
	in
	let tr = 	  
          Why.make_while (tr_expr e) inv why_var body 
        in
	let tr =
	  if eff.breaks then
	    Why.Try(tr,"Break", None (* Some "krak_label" *), Why.Void)
	  else
	    tr
	in
	Why.make_label new_label tr
    | For(l, el1, e, el2, s) -> 
	incr label_counter;
	let new_label = "krak_loop" ^ (string_of_int !label_counter) in
	let (inv,var) = tr_loop_inv new_label l e.expr_loc in
	let eff = Effects.statement s in
        let body = tr_block true (s::el2) model_ensures in
        let body =
          if eff.continue then
            Why.Try(body, "Continue", Some "krak_label", Why.Void)
          else
            body  
        in
	let why_var =
	  if !Config.nonnull_only then None else (Some (var, None))
	in
	let tr =
	  Why.append
	    (tr_block abrupt_return el1 model_ensures)
	    (Why.make_while (tr_expr e) inv why_var body)
	in
	let tr = 
	  if eff.breaks then
	    Why.Try(tr,"Break", None (* Some "krak_label" *), Why.Void)
	  else
	    tr
	in
	Why.make_label new_label tr
    | For_decl(_) -> ill_typed "tr_statement(for_decl)"
    | For_typed_decl(lve, lvi, l, e, el2, s) -> 
	incr label_counter;
	let new_label = "krak_loop" ^ (string_of_int !label_counter) in
	let (inv,var) = tr_loop_inv new_label l e.expr_loc in
	let eff = Effects.statement s in
        let body = tr_block true (s::el2) model_ensures in
        let body =
          if eff.continue then
            Why.Try(body, "Continue", Some "krak_label", Why.Void)
          else
            body  
        in
	let why_var =
	  if !Config.nonnull_only then None else (Some (var, None))
	in
	let tr = 
	  tr_local_var_declaration Location.dummy lve lvi
	    (Why.make_while (tr_expr e) inv why_var body)
	in
	let tr =
	  if eff.breaks then
	    Why.Try(tr,"Break", None (* Some "krak_label" *), Why.Void)
	  else
	    tr
	in
	Why.make_label new_label tr
(*    | Switch(e, cases) -> 
	begin
	  try
   let (l, b) = well_formed_switch cases in
	    Why.Let("temp_switch",
		    tr_expr e,
		    List.fold_right
		      (fun (c,b) acc -> Why.If(c,tr_block true b model_ensures,acc))
		      l (tr_block true b model_ensures))
	  with
	      Exit -> 
		unsupported e.expr_loc 
		  ("tr_statement(switch not well formed)") 
	end *)
    | Switch(e, cases) -> 
	begin
	  try
	    incr temp_switch_count;
	    let temp_switch = ("temp_switch" ^ (string_of_int !temp_switch_count)) in
	    (* let (p, l, b) = well_formed_switch cases model_ensures in *)
            let (p, l) = well_formed_switch cases model_ensures in
            let body_expr =
	      Why.Let(
	        temp_switch,
	        tr_expr e,
                List.fold_left2
                 (fun acc p (c, b) -> 
                    let e = 
                      match p with
                        | None ->
                           if c = (Why.Cte(Why.Prim_bool true))
                           then
                             tr_block true b model_ensures
                           else
                             Why.If(c, tr_block true b model_ensures, Why.Void)
                        | Some s -> s
                    in Why.append acc e)
                 Why.Void
                 p l)
            in
            if is_breakable cases
            then
	      Why.Try(body_expr, "Break", None, Why.Void)
            else body_expr
	      (* List.fold_right2
		(fun p (c, b) acc -> 
		   begin
		   match p with
		    | None -> Why.If(c, tr_block true b model_ensures, acc)
		    | Some s -> Why.Block([s; acc])
		   end)
		p l 
		(tr_block true b model_ensures)) *)
	  with
	      Exit -> 
		unsupported e.expr_loc 
		  ("tr_statement(switch not well formed)") 
	end
    | Block(s) -> tr_block abrupt_return s model_ensures
    | Synchronized (e, _) -> 
	unsupported e.expr_loc "tr_statement(synchronized)"
    | Try (_, _, _) -> 	ill_typed "tr_statement(try-catch)" 
    | Typed_try (b, catches, None) -> 
	let tmp_catch = "tmp_catch" in
	Why.Try
	  (tr_block true b model_ensures, 
	   "Exception", 
	   Some tmp_catch,
	   List.fold_right
	     (fun (id, ty, s) acc ->
	       let whye = 
		 Why.If
		   (Why.And
		      (Why.make_app "_instanceof" [ Why.Deref "alloc" ; Why.Var tmp_catch ; tr_type_expr ty ],
		       Why.make_app "ne_value" [Why.Var tmp_catch; Why.Var "Null"]),
		    Why.Let 
		      (id,
		       Why.Var tmp_catch,
		       tr_block true s model_ensures),
		    acc)
	       in
	       if !Config.transactionException
               then
                 match ty with 
                 | Class_type(ce) -> 
		     if ce.class_entry_name = "TransactionException"
		     then 
		       Why.If
			 (Why.make_app "_instanceof" [ Why.Deref "alloc" ; Why.Var tmp_catch ; tr_type_expr ty ],
			  Why.Let 
			    (id,
			     Why.Var tmp_catch, 
			     Why.Block([build_app (Why.Var("JCSystem_abortTransaction_parameter")) []; tr_block true s model_ensures])),
			  acc)
		     else
		       whye
                 | _ -> 
		     whye
               else
		 whye)
	     catches 
	     (Why.Raise("Exception", Some (Why.Var tmp_catch))))
    | Typed_try (b, catches, bopt) -> 
	unsupported Location.dummy "tr_statement(finally)"
    | Do (inv, s, e) ->
	incr label_counter;
	let new_label = "krak_loop" ^ (string_of_int !label_counter) in
	let (inv,var) = tr_loop_inv new_label inv e.expr_loc in
	let eff = Effects.statement s in
        let body = tr_statement true s model_ensures in
        let body =
          if eff.continue then
            Why.Try(body, "Continue", Some "krak_label", Why.Void)
          else
            body  
        in
	let why_var =
	  if !Config.nonnull_only then None else (Some (var, None))
	in
	let tr = Why.make_while (tr_expr e) inv why_var body in
        let tr = Why.make_label new_label tr in
        let tr = Why.Block [body; tr] in
	if eff.breaks then
	  Why.Try(tr,"Break", None (* Some "krak_label" *), Why.Void)
	else
	  tr
    | Throw(e) ->  
	Why.Raise("Exception",Some (tr_expr e))
    | Jml_assert(e) -> (* assert false *)
	(* NR - TO SEE *)
	  Why.Assert(tr_assertion "init" e, Why.Void)
	(**)
    | Jml_annot_statement(a) ->
        begin
          match a with
          | Set_statement e -> tr_statement abrupt_return ({ statement_loc = s.statement_loc ; statement_node = Typed_expr(Boolean_type,e) }) model_ensures
        end
    | Jml_ghost_var_decl(vd) ->
        ill_typed "tr_statement(local_spec_var_decl)" 
    | Annotated _ ->
	assert false
    | Typed_annotated(spec,st) ->
	let effect = Effects.statement st 
	in let store_extends = 
	    if effect.writes_alloc_table then 
	      Why.LPred("store_extends",[Why.LVar("alloc@");Why.LVar("alloc")]) 
	    else Why.LTrue	   
	in
	let assign = Why.make_and store_extends
	    (tr_assignable Location.dummy 
	       (fun id -> Why.LVarAtLabel(id,"")) 
	       spec.typed_compact_assignable 
	       Java_context.default_routine false)
	in 
	Why.Triple(true,
		   Why.make_and
		     (* old in requires is interpreted in init state *)
		     (tr_assertion "init" spec.typed_compact_requires)
		     !ref_modifiable
		     ,
		   tr_statement abrupt_return st model_ensures,
		   Why.make_and
		     (tr_assertion "" spec.typed_compact_ensures)
		     assign,
		   (tr_signals_clause assign spec.typed_compact_signals)@
		   (tr_cardtear_clause spec.typed_compact_cardtear))
	  

and tr_begin_end abrupt_return i l model_ensures =
  Why.Block([(tr_statement abrupt_return i model_ensures); (tr_block abrupt_return l model_ensures)])


and tr_block abrupt_return (l : statement list) model_ensures : Why.expr = 
(*  Format.printf "Entering tr_block %b@." abrupt_return; *)
  match l with
    | [] -> Why.Void
    | [i] ->
	tr_statement abrupt_return i model_ensures
    | i::l -> 
	match i.statement_node with
	  | Skip -> tr_block abrupt_return l model_ensures
	  | Jml_assert(e) ->
	      Why.Assert(tr_assertion "init" e,tr_block abrupt_return l model_ensures)
	  | If_statement(e,s1,s2) ->
	      begin
		try
		  match (abrupt_termination s1, abrupt_termination s2) with
		    | (false, false) ->
			tr_begin_end abrupt_return i l model_ensures
		    | (true,true) ->
			Format.printf "WARNING : unreachable statement, tr_block@.";
			Why.If
			  (tr_abstract_expr e,
			   tr_statement abrupt_return s1 model_ensures,
			   tr_statement abrupt_return s2 model_ensures)
		    | (false,true) ->
			Why.If
			  (tr_abstract_expr e,
			   tr_block abrupt_return (s1::l) model_ensures, tr_statement abrupt_return s2 model_ensures)
		    | (true,false) ->
			Why.If
			  (tr_abstract_expr e,
			   tr_statement abrupt_return s1 model_ensures,
			   Why.Block([tr_block abrupt_return (s2::l) model_ensures]))
(*
			   tr_statement abrupt_return s1 rep_inv,
			   if rep_inv <> Why.Void 
                           then Why.Block([tr_block abrupt_return (s2::l) rep_inv; rep_inv])
                           else tr_block abrupt_return (s2::l) rep_inv)
*)
		with
		    Not_found ->
		      Why.append
			(Why.If
			  (tr_abstract_expr e,
			   tr_statement true s1 model_ensures, tr_statement true s2 model_ensures))
			(tr_block abrupt_return l model_ensures)
	      end
	  | Typed_var_decl(lve,init,i') -> 
	      (**)
	      tr_statement abrupt_return 
		(mk_statement_no_loc 
		   (Typed_var_decl(lve,init,
				   mk_statement_no_loc (Block(i'::l))))) model_ensures
	      (*
	      Format.printf "***Assert false in tr_block: %a@."
		(Format_misc.fprintf_list Ast.fprintf_statement) (i::l);
	      assert false
	      *)
	  (* | _ -> tr_begin_end i l *)
	  | _ -> Why.Block [tr_statement true i model_ensures; tr_block abrupt_return l model_ensures]

and tr_body pl l model_ensures =
  match pl with
  | [] -> ([], tr_block false l model_ensures)
  | (s, Local_variable_entry lve)::tl ->
      let old_name = lve.local_variable_entry_trans_name in
      lve.local_variable_entry_trans_name <- ("mutable_" ^ old_name);
      lve.local_variable_entry_is_ref <- true;
      let (palv, wb) = tr_body tl l model_ensures in
      ((lve.local_variable_entry_trans_name, Why.Var(old_name))::palv, wb);
  | _ -> assert false
;;
  
let tr_parameter p = 
  match p with
    | Local_variable_entry le ->
	let t = tr_type le.local_variable_entry_type in
	(le.local_variable_entry_trans_name,t)
    | _ -> ill_typed "tr_parameter(non_local)"
;;

let rec params_type params return_type = 
  match params with
    | [] -> return_type
    | (id,t)::l ->
	Why.Prod_type(id,Why.Base_type(t),params_type l return_type)
;;


let rec call_inherited_invariant x extends acc =
  match extends with
  | None -> acc
  | Some ce -> 
      Why.make_and
	(Why.LPred(ce.class_entry_name ^ "_invariant", 
		   List.fold_right
		     (fun (s, _) acc -> Why.LVar(s)::acc)
	             ce.class_entry_invariant_args
	             [Why.LVar(x)]))
	(call_inherited_invariant x ce.class_entry_extends acc)
;;


let arg_pre x class_name class_invariant_args class_restore_invariant_args extends static_invs pre_def = 
  let inherited_inv =
    call_inherited_invariant x extends Why.LTrue
  in
  let inv_prefix =
    if !Config.nonnull_only && not pre_def then "_no" else ""
  in
  let inv = 
    Why.LPred(class_name ^ inv_prefix ^ "_invariant", 
	      List.fold_right
		(fun (s,_) acc -> Why.LVar(s)::acc)
		class_invariant_args
		[Why.LVar(x)])
  in
  let inv = Why.make_and inherited_inv inv in
(* let inv = Why.make_and inv static_invs in *)
  let restore_inv = 
    Why.LPred(class_name ^ "_restore_invariant", 
	      List.fold_right
                (fun (s,_) acc -> Why.LVar(s)::acc)
		class_restore_invariant_args
		[Why.LVar(x)])
  in
   (Why.make_and
     (if !Config.nonnull_only then Why.LTrue else
     (Why.LPred("instanceof",
	        [Why.LVar("alloc");Why.LVar(x);
		 Why.LApp("ClassType",[Why.LVar(class_name)])])))
     (if !Config.javacard 
      then
        (if x="this" then Why.make_and inv restore_inv else
             (Why.LImpl
	       (Why.LPred("neq",[Why.LVar(x);Why.LVar("Null")]), Why.make_and inv restore_inv)))
      else
          (if x="this" then inv else
             (Why.LImpl
	       (Why.LPred("neq",[Why.LVar(x);Why.LVar("Null")]),inv)))),
    Why.make_and
     (if !Config.nonnull_only then Why.LTrue else
     (Why.LPred("instanceof",
	        [Why.LVar("alloc");Why.LVar(x);
		 Why.LApp("ClassType",[Why.LVar(class_name)])])))
          (if x="this" then inv else
             (Why.LImpl
	       (Why.LPred("neq",[Why.LVar(x);Why.LVar("Null")]),inv))))
;;

let arg_pre_class x ce static = 
  fst (arg_pre x ce.class_entry_name ce.class_entry_invariant_args ce.class_entry_restore_invariant_args (if static then None else ce.class_entry_extends) Why.LTrue false)

let arg_pre_interface x ie = 
  if !Config.nonnull_only then Why.LTrue else
  (Why.LPred("instanceof",
	     [Why.LVar("alloc"); 
              Why.LVar(x); 
              Why.LApp("ClassType", [Why.LVar(ie.interface_entry_name)])]))
    
    
let rec args_pre p l static =
  match l with
    | [] -> p
    | (_, xe)::r ->
	let (x, t) = 
	  match xe with
	  | Local_variable_entry lve -> 
	      (lve.local_variable_entry_trans_name, lve.local_variable_entry_type)
	  | _ -> ill_typed "args_pre(nonlocal)"
	in
	match t with
	| Class_type ce ->
	    args_pre 
	      (Why.make_and p (arg_pre_class x ce static))
	      r
	      static
        | Interface_type ie ->
            args_pre
              (Why.make_and p (arg_pre_interface x ie))
              r
	      static
	| Array_type _ ->
	    args_pre
	      (Why.make_and 
		 p 
		 (if !Config.nonnull_only then Why.LTrue else
		 (Why.LPred("instanceof",
			    [Why.LVar("alloc");
			     Why.LVar(x);
			     tr_type_term t]))))
	      r
	      static
	| _ -> args_pre p r static
	      
	      
	      
(* Store the modifiable clause of method for automatic insertion in 
   loop invariant *)

let store_tr_assignable loc eff l for_constructor =
  ref_modifiable := 
    tr_assignable loc (fun id -> Why.LVarAtLabel(id,"init"))
      l eff for_constructor;
  tr_assignable loc (fun id -> Why.LVarAtLabel(id,"")) l eff for_constructor

let reset_tr_assignable () = ref_modifiable := Why.LTrue

let tr_effects e =
  let reads =
    Inttagset.fold
      (fun ive acc ->
         if !Config.javacard && not ive.instance_variable_entry_is_ghost (* && not ive.instance_variable_entry_is_model *)
         then
	   ive.instance_variable_entry_trans_name :: (ive.instance_variable_entry_trans_name ^ "_restore") :: acc
         else
	   ive.instance_variable_entry_trans_name :: acc)
      e.reads
      (if e.reads_alloc_table || e.reads_alloc then ["alloc"] else []) 
  and writes = 
    Inttagset.fold
      (fun ive acc ->
	 ive.instance_variable_entry_trans_name :: acc)
      e.writes
      (if e.writes_alloc_table then ["alloc"] else [])
  in
  let reads = if e.reads_int_array then "intA"::reads else reads in
  let reads = if e.reads_int_assigned_array then "intA_assigned"::reads else reads in
  let writes = if e.writes_int_array then "intA"::writes else writes in
  let reads = if e.reads_float_array then "floatA"::reads else reads in
  let writes = if e.writes_float_array then "floatA"::writes else writes in
  let reads = if e.reads_bool_array then "boolA"::reads else reads in
  let writes = if e.writes_bool_array then "boolA"::writes else writes in
  let reads = if e.reads_obj_array then "objA"::reads else reads in
  let writes = if e.writes_obj_array then "objA"::writes else writes in
  let reads = if !Config.javacard then "volatile"::reads else reads in
  let reads = if !Config.javacard then "transactionDepth"::reads else reads in
  (reads,writes)


let rec use_transactions_expr e =
  match e.expr_node with 
  | If_expr(_, e1, e2) -> use_transactions_expr e1 || use_transactions_expr e2
  | Typed_method_call(_, me, _) -> 
      if me.method_entry_name = "beginTransaction" ||
         me.method_entry_name = "commitTransaction" ||
         me.method_entry_name = "abortTransaction"
      then true 
      else false  
  | _ -> false

let rec use_transactions_statement s =
  match s.statement_node with
  | Typed_expr(t, e) -> use_transactions_expr e
  | Typed_var_decl(_, _, s) -> use_transactions_statement s
  | Label(_, st) -> use_transactions_statement st
  | If_statement(_, s1, s2) -> use_transactions_statement s1 || use_transactions_statement s2
  | While(_, e, s) -> use_transactions_expr e || use_transactions_statement s
  | Do(_, s, e) -> use_transactions_statement s || use_transactions_expr e
  | For(_, sl1, e, sl2, s) -> use_transactions sl1 || use_transactions_expr e || use_transactions sl2 || use_transactions_statement s
  | For_typed_decl(_, _, _, e, sl, s) -> use_transactions_expr e || use_transactions sl || use_transactions_statement s
  | Typed_try(b, [], None) -> use_transactions b
  | Typed_try(b1, [(_, _, b2)], None) -> use_transactions b1 || use_transactions b2 
  | Typed_try(b1, hd::tl, None) -> use_transactions b1 || use_transactions_statement (mk_statement s.statement_loc (Typed_try(b1, tl, None)))
  | Typed_try(b, l, Some f) -> use_transactions_statement (mk_statement s.statement_loc (Typed_try(b, l, None))) || use_transactions f
  | Switch(_, l) -> use_transactions_list (snd(List.split l))
  | Block(b) -> use_transactions b
  | Synchronized(_, b) -> use_transactions b
  | Typed_annotated(_, s) -> use_transactions_statement s
  | _ -> false

and use_transactions body =
  if !Config.cardtear 
  then true
  else
  match body with 
  | [] -> false
  | s::l -> use_transactions_statement s || use_transactions l

and use_transactions_list l =
  match l with
  | [] -> false
  | b::l -> use_transactions b || use_transactions_list l


let collect_model_fields c ef =
  List.fold_left
    (fun acc (_,e) -> 
       match e with
	 | Instance_variable_entry ive -> 
	     if ive.instance_variable_entry_is_model &&
	       Inttagset.mem ive.instance_variable_entry_tag ef.writes
	     then
	       ive.instance_variable_entry_trans_name::acc
	     else acc
	 | _ -> acc
    )
    []
    c
;;

let tr_model_ensures ret_type ef e =
  let pre = Why.LTrue in
  let rt = Why.unit_type (* Why.Base_type(tr_type ret_type) *) in
  let model_fields = List.filter (fun ive -> ive.instance_variable_entry_is_model) (Inttagset.elements ef.writes) in
  let rw = List.map (fun ive -> ive.instance_variable_entry_trans_name) model_fields (* collect_model_fields ce ef *) in
  let post =
    List.fold_left
      (fun post s ->
	Why.make_and
	  post
	  (Why.LPred("modifiable", [Why.LVar("alloc@"); Why.LVar(s ^ "@"); Why.LVar(s); Why.LApp("value_loc", [Why.LVar("this")])])))
      e
      rw
  in
  let exc = [] in
  let wt = Why.Annot_type(pre, rt, rw, rw, post, exc) in
(*
  Format.eprintf "blackbox for %s:" mn;
  List.iter (fun s -> Format.eprintf " %s" s) rw;
  Format.eprintf "@\n  post: %a@." Why.fprintf_assertion post;
 *)
  Why.BlackBox(wt)
;;


let rec find_interface_decl tdl ie =
  match tdl with
    | [] -> assert false (* should never occur *)
    | td::tdl ->
       begin
       match td with
         | Typed_interface(ientry, idecl) ->
             if ie.interface_entry_name = ientry.interface_entry_name 
             then idecl 
             else find_interface_decl tdl ie
         | _ -> find_interface_decl tdl ie
       end
;;

let rec find_impl_method tdl iel me =
  match iel with
    | [] -> None
    | ie::iel ->
        let idecl = find_interface_decl tdl ie in
        try 
          Some(
            snd
            (List.find 
               (fun (_, imd) -> 
                  match imd with
                   | Typed_method_header(mentry, ms) -> me.method_entry_name = mentry.method_entry_name
                   | _ -> false) 
               idecl.interface_members))
        with
          Not_found -> find_impl_method tdl iel me
;;


let rec get_all_class_entries cus =
  List.fold_left
    (fun acc cu ->
      List.append 
	acc
	(List.fold_left
	   (fun acc (_, td) ->
	     match td with
	     | Typed_class (ce, _) -> ce::acc
	     | _ -> acc)
	   []
	   cu.cu_type_decls))
    []
    cus
;;


let tr_method_declaration spec_only ?(class_contents=[]) class_name class_inv_args class_restore_inv_args extends jml me body cus inv restore_inv =  
  temp_switch_count := 0;
  if !Config.javacard &&
     (me.method_entry_name = "beginTransaction" ||
      me.method_entry_name = "abortTransaction" ||
      me.method_entry_name = "arrayCopyNonAtomic")
  then ([], [])
  else 
  begin
  reset_tr_assignable ();
  begin
    match body with
    | None -> ()
    | Some(b) -> me.method_entry_routine.routine_entry_use_transactions <- use_transactions b
  end;
  let routine = me.method_entry_routine in
  let l = routine.routine_entry_parameters in
  let params =
    if me.method_entry_is_static
    then
      if l = []
      then ["tt",([],"unit")]
      else List.map (fun (_,pe) -> tr_parameter pe) l
    else
      ("this",([],"value"))::
      (List.map (fun (_,pe) -> tr_parameter pe) l)
  in
  let tdl = List.flatten (List.map (fun cu -> List.map snd cu.cu_type_decls) cus) in
  let (impl_assignable, impl_ensures, impl_model_ensures, impl_effects) =
    if spec_only
    then 
      ([], None, None, default_effects)
    else
      begin
        match me.method_entry_class_or_interface with
          | Class_entry ce -> 
             begin
             match ce.class_entry_implements with
               | [] -> ([], None, None, default_effects)
               | iel -> 
                  match find_impl_method tdl iel me with
                   | None ->  ([], None, None, default_effects)
                   | Some (Typed_method_header(me, ms)) ->
                      begin
                      match ms with 
                        | Typed_compact tcs -> (tcs.typed_compact_assignable, Some tcs.typed_compact_ensures, Some tcs.typed_compact_model_ensures, me.method_entry_routine.routine_entry_effects)
                        | _ -> assert false (* not supported msg to put here *)
                      end
                   | _ -> assert false (* should never occur *)
             end 
          | _ -> ([], None, None, default_effects)
      end
  in
  let impl_ensures = match impl_ensures with | None -> Why.LTrue | Some e -> tr_assertion "" e in
  let impl_model_ensures = match impl_model_ensures with | None -> Why.LTrue | Some e -> tr_assertion "" e in
  routine.routine_entry_effects <- 
    Effects.effects_union routine.routine_entry_effects impl_effects;
  let (jml_pre,jml_assignable, model_ensures_body, model_ensures, jml_post, jml_post_expr, jml_signals, jml_cardtear) = 
    match jml with
    | Typed_compact_nested tcns ->
        let writes = 
          List.map 
            (fun ive -> ive.instance_variable_entry_trans_name) 
            (Inttagset.elements routine.routine_entry_effects.writes) in
	let requires = 
	 Why.make_and 
	    (tr_assertion "" tcns.typed_compact_nested_nre.typed_compact_requires) 
	    (Why.make_or_list (List.map (fun tcn -> tr_assertion "" tcn.typed_compact_requires) tcns.typed_compact_nested_fs))
	and assignable = tcns.typed_compact_nested_nre.typed_compact_assignable
	and ensures = 
	  Why.make_and 
	    (tr_assertion "" tcns.typed_compact_nested_nre.typed_compact_ensures)
	    (Why.make_and_list 
	       (List.map 
		  (fun tcs -> 
		    Why.LImpl(
                      tr_assertion_rec 
                        "" 
                        (fun var -> 
                           let var = if List.mem var writes then var ^ "@" else var in
                           Why.LVar(var)) 
                        tcs.typed_compact_requires, 
                      Why.make_and 
                        (tr_assertion "" tcs.typed_compact_ensures) 
                        (store_tr_assignable Location.dummy routine tcs.typed_compact_assignable false)))
		  tcns.typed_compact_nested_fs))
	and signals = 
          let new_signals_clause =
            List.fold_left
              (fun acc tcs ->
                 match tcs.typed_compact_signals with
                   | None -> acc
                   | Some(_, _, e) ->
                       begin
                         match acc with
                           | None -> tcs.typed_compact_signals
                           | Some(jt, lveo, eacc) -> 
                               Some(jt, 
                                    lveo, 
                                    mk_bin Location.dummy eacc "&&" (mk_bin Location.dummy tcs.typed_compact_requires "==>" e)) 
                       end 
              )
              None
              tcns.typed_compact_nested_fs       
          in
          tr_signals_clause Why.LTrue new_signals_clause
	and cardtear = [] (* TO DO *)
	in
	(requires, assignable, Why.LTrue, ensures, ensures, tcns.typed_compact_nested_nre.typed_compact_ensures, signals, cardtear)
    | Typed_compact tcc ->
	let requires = tr_assertion "" tcc.typed_compact_requires
	and assignable = tcc.typed_compact_assignable
	and ensures = tr_assertion "" tcc.typed_compact_ensures
	and signals = tr_signals_clause Why.LTrue tcc.typed_compact_signals
        and cardtear = tr_cardtear_clause tcc.typed_compact_cardtear
	in      
(*
	result_is_returned := true;
*)
	let model_ensures_body = tr_assertion "init" tcc.typed_compact_model_ensures
	and model_ensures = tr_assertion "" tcc.typed_compact_model_ensures
	in
(*
	result_is_returned := false;	
*)
	(requires, assignable, model_ensures_body, model_ensures, ensures, tcc.typed_compact_ensures, signals, cardtear)
    | _ -> ill_typed "tr_method_declaration(jml desugaring)"
  in 
  let jml_assignable = List.append jml_assignable impl_assignable in
  let jml_post = Why.make_and jml_post impl_ensures in
  let model_ensures_body = Why.make_and model_ensures_body impl_model_ensures in
  let model_ensures = Why.make_and model_ensures impl_model_ensures in
  let jml_post = Why.make_and jml_post model_ensures in
  (* assertion about \result in the post; so maybe 'args_pre' is not the right name anymore for that function - NR *)
  let jml_post = 
    let lve_result = 
      {
        local_variable_entry_name = "result";
        local_variable_entry_trans_name = "result";
        local_variable_entry_is_ref = false;
        local_variable_entry_type = me.method_entry_return_type
      }
    in
    args_pre 
      jml_post 
      ["result", Local_variable_entry lve_result] 
      me.method_entry_is_static
  in
  let (pre, pre_def, post, post_def, signals) =    
    if me.method_entry_is_static
    then (args_pre jml_pre l true, args_pre jml_pre l true, jml_post, jml_post, jml_signals)
    else 
      let make_inv a =
        if !Config.unfold_pred >= 2 
        then inv
        else 
	  (Why.LPred(class_name ^ "_invariant",
			       (List.fold_right
			          (fun (s,_) acc -> Why.LVar(s)::acc) 
			          a
			          [Why.LVar("this")])))
      in
      (* NR - TO DO : only static invariants of 'visible' classes *)
      let static_invs =
	List.fold_left
	  (fun acc ce ->
	    if List.length ce.class_entry_static_invariant_args = 0
	    then acc
	    else
	      Why.make_and 
		acc
		(Why.LPred(ce.class_entry_name ^ "_static_invariant",
			   List.map
			     (fun (s, _) -> Why.LVar(s))
			     ce.class_entry_static_invariant_args)))
	  Why.LTrue
	  (get_all_class_entries cus)
      in
      let make_restore_inv a =
        if !Config.unfold_pred >= 2 
        then restore_inv
        else 
	  (Why.LPred(class_name ^ "_restore_invariant",
			       (List.fold_right
			          (fun (s,_) acc -> Why.LVar(s)::acc) 
			          a
			          [Why.LVar("this")])))
      in
      let extends = if me.method_entry_is_static then None else extends in
      ((args_pre 
	  (Why.make_and jml_pre 
	     (Why.make_and 
		(Why.LPred("neq",[Why.LVar "this";Why.LVar "Null"]))
		(snd (arg_pre "this" class_name class_inv_args class_restore_inv_args extends static_invs false)))) l) false,
       (args_pre 
	  (Why.make_and jml_pre 
	     (Why.make_and 
		(Why.LPred("neq",[Why.LVar "this";Why.LVar "Null"]))
		(fst (arg_pre "this" class_name class_inv_args class_restore_inv_args extends static_invs true)))) l) false,
       (if me.method_entry_is_pure 
       then jml_post 
       else
	 Why.make_and 
	   jml_post 
	   (* (Why.make_and *)
	   (if !Config.javacard 
	   then Why.make_and (make_inv class_inv_args) (make_restore_inv class_restore_inv_args) 
	   else (make_inv class_inv_args))
	   (*static_invs*)),
       (if me.method_entry_is_pure then jml_post else
       Why.make_and 
	 jml_post 
	 (*Why.make_and*) (make_inv class_inv_args) (*static_invs*)),
       (if me.method_entry_is_pure then jml_signals else
       List.map 
	 (fun (e,p) -> 
	   (e, Why.make_and p  
	      (* if !Config.javacard 
		 then Why.make_and (make_inv class_inv_args) (make_restore_inv class_restore_inv_args) 
		 else *) (* Why.make_and static_invs*) (make_inv class_inv_args)))
	 jml_signals))
  in
  let assignable =
    if routine.routine_entry_effects.writes_alloc then 
      store_tr_assignable Location.dummy routine jml_assignable false
    else Why.LTrue
  in
  let post = Why.make_and post assignable in
  let post_def = Why.make_and post_def assignable in
  let return_base_type = tr_type me.method_entry_return_type in
  let return_type = Why.Base_type(return_base_type) in
  let (reads,writes) = tr_effects routine.routine_entry_effects in
  let reads =
    match me.method_entry_class_or_interface with
    | Class_entry ce -> List.append reads (List.map fst ce.class_entry_invariant_args)
    | _ -> reads
  in
  let rec inherited_reads extends reads = 
    match extends with
    | None -> reads
    | Some ce -> 
	List.append 
	  (inherited_reads ce.class_entry_extends reads) 
	  (fst (tr_effects ce.class_entry_invariant_effects))
  in
  let reads = inherited_reads extends reads in
  let reads = 
    List.append 
      (fst (tr_effects (Effects.param_effects routine.routine_entry_effects routine))) 
      reads 
  in
(*i
  Format.fprintf Config.log "Effects: method %s writes" me.method_entry_name;
  List.iter (fun s -> Format.fprintf Config.log " %s" s) writes;
  Format.fprintf Config.log ".@.";
i*)
  let signals =
    match (signals,routine.routine_entry_effects.throws) with
      | ([], false) -> []
      | ([],true) ->
	  Format.printf "Warning: method %s may throw an exception, whereas its specification does not allows it.@." me.method_entry_trans_name;
	  [("Exception",Why.LFalse)]
      | (l, true) -> 
	  List.map (fun (s,a) -> (s,Why.make_and a assignable)) signals
      | (l, false) ->
	  Format.printf "Warning: method %s body cannot throw an exception,@ whereas its specification allows it.@." me.method_entry_name;
(*	  if !Config.javacard
          then List.map (fun (s,a) -> (s, Why.make_and a assignable)) signals
          else *) [] (* Some ("Exception", Why.LFalse) *)
  in  
  let signals = signals @ jml_cardtear in
  let all_params = 
    if !Config.javacard
    then
      Effects.all_params_reads routine.routine_entry_effects params 
    else
      Effects.params_reads routine.routine_entry_effects params false 
  in
  let _why_type = params_type params return_type 
  and why_annot_type = 
    params_type params 
      (Why.Annot_type(pre,return_type,
		      reads,writes,
		      post,signals))
  in
  let why_decls =
    let extern_decl =
      Why.Param(true, (* external *)
		class_name ^ "_" ^ me.method_entry_trans_name ^ "_parameter", 
		why_annot_type)
    in
    if me.method_entry_is_pure 
    then 
      if me.method_entry_is_predicate then
	let logic_decl =
	  let ret = ([],"prop")
(*	  and params = List.map (fun (id,t)
*)	  in
	  Why.Logic(false,class_name ^ "_" ^ me.method_entry_trans_name, params, ret)
	in 
	[logic_decl; extern_decl]
	     (*i
	       let logic_bool_decl =
	       Why.Logic
	       (class_name ^ "_" ^ me.method_entry_name ^ "_bool" , 
	       Why.Prod_type("",Why.Base_type(alloc_table_type_name),why_type))	  
	       and logic_prop_decl =
	       Why.Logic
	       (class_name ^ "_" ^ me.method_entry_name, 
	       Why.Prod_type("",Why.Base_type(alloc_table_type_name),
	       params_type params (Why.Base_type "prop")))	  
	       in [logic_prop_decl;logic_bool_decl;extern_decl]
	       i*)
      else
	let logic_decl =
	  Why.Logic(false,class_name ^ "_" ^ me.method_entry_trans_name,all_params,return_base_type)
	in 
	  [logic_decl; extern_decl]
    else
      [extern_decl]
  and coq_decls =
    if me.method_entry_is_pure 
    then 
      if me.method_entry_is_predicate then
	let predicate =
	  match jml_post_expr.expr_node with
	    | Bin(_,"<==>",e) -> e 
	    | _ -> assert false
	in 	
	
	let prop_decl =
	  Why.Predicate
	    (false,
	     class_name ^ "_" ^ me.method_entry_trans_name, 
	     all_params,
	     (Why.make_and jml_pre (tr_assertion "" predicate)))
	in
	[prop_decl]
      else
	let parameter_decl =
	  Why.Logic(false,class_name ^ "_" ^ me.method_entry_trans_name, all_params,return_base_type)
	and axiom_decl =
	  let arguments =
	    List.map (fun (x,_) -> Why.LVar x) all_params
	  in
	  let pred =
	    match signals with
	      | [] -> post
	      | l -> Why.make_or post (Why.make_or_list (List.map snd l))
	  in
	  let fapp = Why.LApp 
	    (class_name ^ "_" ^ me.method_entry_trans_name,arguments)
	  in 
          let pred_result = if Why.occ_result_bound pred then 
	     Why.LForall("result",return_base_type,
		(Why.LImpl(Why.LPred("eq",[Why.LVar("result"); fapp]),pred)))
          else Why.subst_resulta  fapp pred
	  in
          let axiom = 
	    List.fold_right
	      (fun (x,t) acc ->
		 Why.LForall(x,t,acc))
	      all_params
	      (Why.LImpl(jml_pre,pred_result))
	  in
(*
		 Why.LForall("result",return_base_type,
		(Why.LImpl(Why.LPred("eq",[Why.LVar("result");                         Why.LApp (class_name ^ "_" ^ me.method_entry_trans_name,arguments)]), pred)))))
	  in
*)

	  Why.Axiom
	    (class_name ^ "_" ^ me.method_entry_trans_name ^ "_axiom", 
	     axiom)
	in      
	[parameter_decl;axiom_decl]
    else []
  in
  if spec_only 
  then 
    why_decls, coq_decls
  else
    match body with
      | Some l ->
	  begin
	  Format.fprintf Config.log "Translating body of method %s.@." me.method_entry_trans_name;
	    try
	      ab_ret := None;
	      let model_ensures = 
		if me.method_entry_is_static
		then Why.Void
		else tr_model_ensures me.method_entry_return_type routine.routine_entry_effects model_ensures_body 
	      in 
	      let l =
		match me.method_entry_return_type with
		  | Null_type -> 
		      l@[mk_statement_no_loc (Typed_return(Null_type,None))]
		  | _ -> l
	      in
	      let (palv, tbody) = tr_body routine.routine_entry_parameters_as_local_var l model_ensures in
	      let tbody =
		match !ab_ret with
		  | None -> assert false
		  | Some(ex) -> 
                      Why.Try(tbody, 
                              ex, 
                              Some "returned", 
                              if me.method_entry_is_static 
                              then Why.Var "returned" 
                              else 
				(* Why.Let("model_returned",model_ensures,
				   Why.Assert(Why.LPred("eq",[Why.LVar "returned"; Why.LVar "model_returned"]), Why.Var "returned")) *)
				Why.append model_ensures (Why.Var "returned"))
	      in
	      let params =
		List.map (fun (x,t) -> (x, Why.Base_type(t))) params 
	      in
              let tbody = Why.Label("init", tbody) in
              let tbody =
                List.fold_left
                  (fun acc (n, o) ->
                     Why.Let_ref(n, o, acc))
                  tbody
                  palv
              in
	      ((Why.Def
		  (class_name ^ "_" ^ me.method_entry_trans_name, 
		   Why.Fun
		     (params, pre_def, tbody, post_def, signals))
		::why_decls),coq_decls)
	    with
	      | Unsupported msg ->
		  Format.printf 
		    "**Warning: translation of body of %s not performed due to unsupported feature in translation: %s@."  
		    me.method_entry_trans_name msg;
		  (why_decls,coq_decls)
	  end
      | None -> why_decls,coq_decls
  end
;;


let fd_to_af ce fd = 
  match (snd fd) with
    | Typed_variable(ive, vio) ->
        let this =
	  { 
	    local_variable_entry_name = "this";
	    local_variable_entry_trans_name = "this";
	    local_variable_entry_is_ref = false;
	    local_variable_entry_type = Class_type ce;
	  } 
        in
        let fa = Typed_primary_access(mk_expr_no_loc (Typed_var(this)), ive) in
        let e = match vio with
          | None -> 
              mk_expr_no_loc
              (match ive.instance_variable_entry_type with
                | Byte_type | Short_type | Integer_type -> Lit(Int 0)
                | Float_type | Real_type -> assert false (* TO DO *)
		| Prop_type | Type_var _ | Logic_type _ -> assert false
                | Boolean_type -> Lit(Bool false)
                | Null_type | Type_type -> assert false (* should never happen *)
                | Class_type _ | Interface_type _ | Array_type _ -> Lit Null)
          | Some vi -> 
              match vi with
                | Simple_initializer e -> e
                | Array_initializer vil -> assert false (* TO DO *)
        in
        (ive.instance_variable_entry_type, mk_expr_no_loc (Assign_field(fa, "=", e)))
    | _ -> assert false (* should never happen *)
;;


let tr_constr_declaration spec_only class_entry jml ce body c inv restore_inv super super_args =
  temp_switch_count := 0;
  let class_name = class_entry.class_entry_name in
  let routine = ce.constructor_entry_routine in
  let l = routine.routine_entry_parameters in
  let params = 
    if l=[]
    then ["tt",([],"unit")]
    else
      List.map (fun (_,pe) -> tr_parameter pe) l
  in
  let (jml_pre, jml_assignable, model_ensures, model_ensures_this, jml_post, jml_signals, jml_cardtear) = 
    match jml with
      | Typed_compact tcc ->
	  let requires = 
            tr_assertion "" tcc.typed_compact_requires
	  and assignable = tcc.typed_compact_assignable
	  and model_ensures,model_ensures_this,ensures = 
	    begin
	      let mrt = tr_assertion "" tcc.typed_compact_model_ensures in
	      this_is_result := true;
	      let mr = tr_assertion "" tcc.typed_compact_model_ensures in
	      let r = tr_assertion "" tcc.typed_compact_ensures in
	      this_is_result := false;
	      mr,mrt,r
	    end
	  and signals = tr_signals_clause Why.LTrue tcc.typed_compact_signals
          and cardtear = tr_cardtear_clause tcc.typed_compact_cardtear
	  in      
	    (requires, assignable, model_ensures, model_ensures_this, ensures, signals, cardtear)
      | _ -> ill_typed "JML Desugaring failed"
  in
  let jml_post = Why.make_and jml_post model_ensures in
  let (pre,post, post_unfold_pred,signals) =    
    if !Config.javacard && !Config.cardtear
    then
      (args_pre 
       jml_pre
       l false,
       Why.make_and jml_post 
	 (Why.make_and 
	    (Why.LPred("neq",[Why.LVar "result";Why.LVar "Null"]))
	    (Why.make_and 
	       (Why.LPred
		  ("instanceof",
		   [Why.LVar("alloc");Why.LVar("result");
		    Why.LApp
		      ("ClassType",
		       [Why.LVar
			  (class_name)])]))
	       (Why.make_and
		  (Why.LPred("fresh",[Why.LVar("alloc@");Why.LVar("result")]))
                     (Why.LPred(class_name ^ "_invariant", 
			     (List.fold_right
				(fun (s,_) acc -> Why.LVar(s)::acc) 
				class_entry.class_entry_invariant_args
				[Why.LVar("result")])))))) ,
       Why.make_and jml_post 
	 (Why.make_and 
	    (Why.LPred("neq",[Why.LVar "result";Why.LVar "Null"]))
	    (Why.make_and 
	       (Why.LPred
		  ("instanceof",
		   [Why.LVar("alloc");Why.LVar("result");
		    Why.LApp
		      ("ClassType",
		       [Why.LVar
			  (class_name)])]))
	       (Why.make_and
		  (Why.LPred("fresh",[Why.LVar("alloc@");Why.LVar("result")]))
                     (Why.rtbra inv)))),
	 jml_signals)
    else
      (args_pre 
       jml_pre
       l false,
       Why.make_and jml_post 
	 (Why.make_and 
	    (Why.LPred("neq",[Why.LVar "result";Why.LVar "Null"]))
	    (Why.make_and 
	       (if !Config.nonnull_only then Why.LTrue else
	       (Why.LPred
		  ("instanceof",
		   [Why.LVar("alloc");Why.LVar("result");
		    Why.LApp
		      ("ClassType",
		       [Why.LVar
			  (class_name)])])))
	       (Why.make_and
		  (Why.LPred("fresh",[Why.LVar("alloc@");Why.LVar("result")]))
                     (Why.LPred(class_name ^ "_invariant", 
			     (List.fold_right
				(fun (s,_) acc -> Why.LVar(s)::acc) 
				class_entry.class_entry_invariant_args
				[Why.LVar("result")])))))) ,
       Why.make_and jml_post 
	 (Why.make_and 
	    (Why.LPred("neq",[Why.LVar "result";Why.LVar "Null"]))
	    (Why.make_and 
	       (if !Config.nonnull_only then Why.LTrue else
	       (Why.LPred
		  ("instanceof",
		   [Why.LVar("alloc");Why.LVar("result");
		    Why.LApp
		      ("ClassType",
		       [Why.LVar
			  (class_name)])])))
	       (Why.make_and
		  (Why.LPred("fresh",[Why.LVar("alloc@");Why.LVar("result")]))
                     (Why.rtbra inv)))),
	 jml_signals)
  in
  let assignable = 
    store_tr_assignable Location.dummy routine 
      jml_assignable true
  in
  let post = Why.make_and post assignable 
  and signals =
    match (signals,routine.routine_entry_effects.throws) with
      | ([],false) -> []
      | ([],true) ->
	  (*
	    Format.printf "Warning: method %s may throw an exception, whereas its specification does not allows it.@." me.method_entry_name;
	  *)
	  [("Exception",Why.LFalse)]
      | (_,true) -> 
          
	  List.map (fun (s,a) -> (s,Why.make_and a assignable)) signals
      | (_,false) ->
	  Format.printf "Warning: constructor %s body cannot throw an exception,@ whereas its specification allows it.@." ce.constructor_entry_class.class_entry_name;
	  []
  in  
  let return_type = Why.base_type("value") in
  let (reads, writes) = tr_effects routine.routine_entry_effects in 
  let _why_type = params_type params return_type 
  and why_annot_type = 
    params_type params
      (Why.Annot_type(pre, return_type, reads, writes, post, signals@ jml_cardtear))
  in
  let why_annot_type_super =
    params_type (("this", ([], "value"))::params)
      (Why.Annot_type(Why.rrbta pre, 
		      Why.unit_type, 
		      reads, 
		      writes, 
		      Why.rrbta post, 
		      List.map (fun (s, a) -> (s, Why.rrbta a)) signals@jml_cardtear))
  in
  let extern_decl =
    Why.Param(true, (* external *)
	      ce.constructor_entry_trans_name ^ "_parameter", 
	      why_annot_type) 
  in
  let extern_decl_super =
    Why.Param(true, (* external *)
	      ce.constructor_entry_trans_name ^ "_super", 
	      why_annot_type_super) 
  in  
  if spec_only then [extern_decl; extern_decl_super], []
  else
    let class_fields = c.class_fields in
    let instance_variables = 
      List.filter 
      (fun (_, fd) -> 
	 match fd with 
	   | Typed_variable (ive, _) -> 
	       not (ive.instance_variable_entry_is_final || 
		      ive.instance_variable_entry_is_model) 
	   | _ -> false) 
      class_fields 
    in
    let assign_fields = List.map (fd_to_af class_entry) instance_variables in
    let init_body = (List.map (fun (t, e) -> mk_statement_no_loc (Typed_expr(t, e))) assign_fields) in
    let body = List.append init_body body in
    let model_ensures = tr_model_ensures (Class_type class_entry) 
      (* class_entry.class_entry_contents *)
      ce.constructor_entry_routine.routine_entry_effects
      model_ensures_this
    in
    ab_ret := None;
    let (palv, tbody) = tr_body ce.constructor_entry_routine.routine_entry_parameters_as_local_var body model_ensures in
    let tbody =
      if super
      then
        match class_entry.class_entry_extends with
	| Some cee ->
	    Why.append 
	      (Why.make_app 
		 (cee.class_entry_name ^ "_" ^ cee.class_entry_name ^ "_super") 
		 ((Why.Var "this")::super_args)) 
	      tbody
	| None -> failwith "typing error"
      else 
	tbody
    in
    let tbody =
      match !ab_ret with
	| None -> 
	    Why.append tbody 
	      (* (Why.Let("model_returned", *) model_ensures (*, Why.Void) *)
	| Some(ex) -> 
	    assert false (* Why.Try(tbody,ex,Some "returned",Why.Block([model_ensures; Why.Var "returned"])) *)
    in
    let tbody = 
      ab_ret := None;
      let new_val = "this" in
      Why.Let
	(new_val,
	 Why.make_app
	   ("alloc_new_obj" ^ if !Config.javacard then "_javacard" else "")
	   [Why.Var(ce.constructor_entry_class.class_entry_name)],
	   Why.Block
	     ([tbody;
	       Why.Var(new_val)]))
    in
    let params =
      List.map (fun (x,t) -> (x, Why.Base_type(t))) params 
    in
    let tbody = Why.Label("init", tbody) in
    let tbody =
      List.fold_left
        (fun acc (n, o) ->
           Why.Let_ref(n, o, acc))
        tbody
        palv
    in
    ([Why.Def(ce.constructor_entry_trans_name, 
	      Why.Fun(params,pre,
		      tbody,
		      (if !Config.unfold_pred >= 2 then post_unfold_pred else post), 
                      signals@ jml_cardtear));
      extern_decl; extern_decl_super],[])
;;

let tr_instance_variable_declaration class_name ve = 
  if !Config.verbose >=2 then
    Format.printf "producing declaration of field %s.@."
      ve.instance_variable_entry_trans_name;
   ([ Why.Param
       (false,
	tr_ive ve, 	     
	Why.Ref_type
	  (Why.Base_type(Effects.compute_translated_type ve)))
       
   ] ,
   [])
;;


let tr_static_variable_declaration loc v init =
  let _i = 
    match init with
    | None -> type_default_value false v.instance_variable_entry_type
    | Some e -> 
	type_default_value false v.instance_variable_entry_type
	(*i 
	  tr_variable_initializer loc e
	  i*)
  in
  let name = v.instance_variable_entry_trans_name
  and t = tr_type v.instance_variable_entry_type
  in
  if v.instance_variable_entry_is_final
  then
    let coq_decl =
      match init with
	| None -> [Why.Param(false,name,Why.Base_type t)]
	| Some vi ->
	   begin
	   match vi with
	     | Simple_initializer _ ->
		[Why.Logic(false,name,[],t) ;
		 Why.Axiom
		   (name ^ "_value",
		    Why.LPred("eq",
			      [Why.LVar(name);
			       tr_term_variable_initializer 
				 Location.dummy vi]))
		]	      
	     | Array_initializer vil ->
                let array_type = match v.instance_variable_entry_type with
                  | Array_type t -> 
                      begin
                      match t with
                        | Byte_type | Short_type | Integer_type -> Why.LVar "ArrIntType"
                        | Float_type -> Why.LVar "ArrFloatType"
                        | Boolean_type -> Why.LVar "ArrBoolType"
                        | Class_type(ce) -> Why.LApp("ClassType", [Why.LVar ce.class_entry_name])
                        | Interface_type(ie) -> Why.LApp("ClassType", [Why.LVar ie.interface_entry_name]) (* CORRECT ??? *)
			| _ -> assert false (* should never happen *)
                      end
                  | _ -> assert false (* should never happen *)
                in
                [Why.Logic(false, name, [], t);
		 Why.Axiom
		   (name ^ "_value",
                    Why.LAnd(
		      Why.LPred("neq", 
				[Why.LVar name; Why.LVar "Null"]),
		      Why.LForall("alloc", 
			          ([], "alloc_table"),
                                  Why.LAnd(
			            Why.LPred("instanceof", [Why.LVar "alloc"; Why.LVar name; array_type]),
				    Why.LPred("eq_int", 
					      [Why.LApp("arraylength", [Why.LVar "alloc"; Why.LVar name]); 
					       Why.LConst(Why.Prim_int(string_of_int (List.length vil)))])))))] (* TO COMPLETE - NR *)
	   end
    in
    ([Why.Logic(false,name,[],t)], coq_decl )
  else
    ([Why.Param(false,name,Why.Ref_type(Why.Base_type t))],
     [])


let rec find_method_name m =
  match m with
    | Simple_method_declarator((loc,id),_) -> id
    | Array_method_declarator(m) -> find_method_name m

let name_of_field_declaration f =
  match f with 
    | Typed_method(me,jml,b) -> "method " ^ me.method_entry_trans_name
    | Typed_constructor(ce,jml,eci,b) -> 
	"constructor " ^ ce.constructor_entry_class.class_entry_name
    | Typed_variable(ve,init) -> "field " ^ ve.instance_variable_entry_name
    | Static_initializer(_) -> "static initializer"
    | Method(md,_) -> "method " ^ find_method_name md.method_declarator
    | Constructor(cd) -> "constructor " ^ (snd cd.constr_name)
    | Variable(v) -> "field"
    | Specification(_) -> "jml specification"

(*
let get_writes model_ensures =
  match model_ensures with
    | Why.BlackBox(Why.Annot_type(_, _, _, w, _, _)) -> w
    | Why.Void -> []
    | _ -> assert false (* should never happen *)
;;
*)

let tr_field_declaration class_entry f c tdl inv restore_inv = 
  try
    match f with 
      | Typed_method(me, jml, b) -> 
	  Format.fprintf Config.log "Translating method %s.@." me.method_entry_name;
(*
          let me_writes = snd (tr_effects me.method_entry_routine.routine_entry_effects) in
          let model_ensures = 
            if List.exists (fun s -> List.mem s me_writes) (get_writes model_ensures) 
            then Why.Void
            else model_ensures
          in
*)
	  begin
            match jml with
              | Typed_compact_nested tcns ->
		  if me.method_entry_routine.routine_entry_produce_body
		  then
                    let (wd_nre, cd_nre) = 
                      tr_method_declaration
			~class_contents:class_entry.class_entry_contents
			false
			class_entry.class_entry_name  
			class_entry.class_entry_invariant_args
			class_entry.class_entry_restore_invariant_args
                        class_entry.class_entry_extends
			(Typed_compact tcns.typed_compact_nested_nre) 
			({ me with method_entry_trans_name = me.method_entry_trans_name ^ "_nre" }) 
			b tdl inv restore_inv
                    in 
                    let rec mk_wd_cd tcsl i =
                      match tcsl with
                      | [] -> ([], [])
                      | hd::tl -> 
                          let (wd, cd) = 
                            tr_method_declaration 
			      ~class_contents:class_entry.class_entry_contents
	                      (not me.method_entry_routine.routine_entry_produce_body)
	                      class_entry.class_entry_name  
	                      class_entry.class_entry_invariant_args
	                      class_entry.class_entry_restore_invariant_args
			      class_entry.class_entry_extends
                              (Typed_compact 
                                 { hd with typed_compact_requires = 
                                     mk_bin 
                                       Location.dummy 
                                       hd.typed_compact_requires 
                                       "&&" 
                                       tcns.typed_compact_nested_nre.typed_compact_requires }) 
                              (let label = List.nth me.method_entry_sep_specs i in
			      if label = "" 
			      then { me with method_entry_trans_name = me.method_entry_trans_name ^ "_f" ^ string_of_int i }
			      else { me with method_entry_trans_name = me.method_entry_trans_name ^ "_" ^ label }) 
                              b tdl inv restore_inv;
                          in
                          (List.append wd (fst (mk_wd_cd tl (i + 1))), List.append cd (snd (mk_wd_cd tl (i + 1))))
                    in 
                    let (wd_fs, cd_fs) = mk_wd_cd tcns.typed_compact_nested_fs 0 in
                    (List.append wd_nre wd_fs, List.append cd_nre cd_fs)
		  else
                    tr_method_declaration
		      ~class_contents:class_entry.class_entry_contents
		      true
		      class_entry.class_entry_name  
		      class_entry.class_entry_invariant_args 
		      class_entry.class_entry_restore_invariant_args 
		      class_entry.class_entry_extends
		      jml 
		      me 
		      b tdl inv restore_inv
              | _ -> 
                  tr_method_declaration 
		    ~class_contents:class_entry.class_entry_contents
	            (not me.method_entry_routine.routine_entry_produce_body)
	            class_entry.class_entry_name  
	            class_entry.class_entry_invariant_args 
                    class_entry.class_entry_restore_invariant_args
		    class_entry.class_entry_extends
                    jml me b tdl inv restore_inv
		    end
      | Typed_constructor(ce, jml, eci, b) ->
	  begin
	    match eci with
	      | Invoke_none -> 
		  tr_constr_declaration 
		  (not ce.constructor_entry_routine.routine_entry_produce_body)
		  class_entry jml ce b c inv restore_inv false []
	      | Invoke_this(el) -> 
		  if ce.constructor_entry_routine.routine_entry_produce_body then
		    unsupported Location.dummy
		      "tr_field_declaration(constructor_invoke(this))"
		  else
		    tr_constr_declaration true
		      class_entry jml ce b c inv restore_inv false []
	      | Invoke_super(el) -> 
		  if ce.constructor_entry_routine.routine_entry_produce_body then
		    let tel = List.map tr_expr el in
                    tr_constr_declaration false
                      class_entry jml ce b c inv restore_inv true tel
		      (* unsupported Location.dummy
			 "tr_field_declaration(constructor_invoke(super))" *)
		  else
		    tr_constr_declaration true
		      class_entry jml ce b c inv restore_inv false []
	  end      
      | Typed_variable(ve,init) ->
	  if ve.instance_variable_entry_is_static
	  then tr_static_variable_declaration Location.dummy ve init
	  else tr_instance_variable_declaration class_entry.class_entry_name ve
      | Static_initializer _ -> 
	  unsupported Location.dummy "tr_field_declaration(static_init)"
      | Method(_,_) -> 
	  ill_typed "tr_field_declaration(method)"
      | Constructor(_) -> 
	  ill_typed "tr_field_declaration(constructor)"
      | Variable(_) -> 
	  ill_typed "tr_field_declaration(variable)"
      |	Specification(_) -> ([],[]) (* already computed before *)
  with
    | Unsupported msg when not !Config.stop_on_error  ->
	Format.printf 
	  "**Warning: translation of %s not performed due to unsupported feature in translation: %s@." 
	  (name_of_field_declaration f) msg;
	([],[])
    | Typing.Ill_typed msg when not !Config.stop_on_error  -> 
	Format.printf 
	  "**Warning: translation of %s not performed due to typing errors (%s)@."
	  (name_of_field_declaration f) msg;
	([],[])



let collect_class_restore_invariant class_name class_fields =
  try
    List.fold_left
      (fun acc (doc, f) ->
	 match f with
	   | Specification s ->
	       List.fold_left
		 (fun acc s ->
		   match s with
		   | Representation_invariant e 
		   | Invariant e ->
        	       Why.make_and acc (tr_assertion "" (replace_by_restore e))
		   | Static_invariant _
		   | Represents _ 
		   | Typed_represents _
		   | Jml_variable_declaration _
		     -> acc
		   )
		 acc s
	   | Typed_variable(ive, init) ->
	       if ive.instance_variable_entry_is_static (* ||
                  ive.instance_variable_entry_is_model *)
	       then acc
	       else
		 let ty = ive.instance_variable_entry_type in
		 begin
		   match ty with
		     | Byte_type | Short_type | Integer_type | Boolean_type | Float_type | Logic_type _ -> acc
		     | _ ->
                         Why.make_and
			   (Why.LPred
			      ("instanceof",
			       [Why.LVar "alloc";
				Why.LApp
				  ("acc",
				   [Why.LVar 
				      (ive.instance_variable_entry_trans_name ^ "_restore");
				    Why.LVar "this" ]);
				tr_type_term ty;
			       ]))
			   acc
		 end
	   | _ -> acc)
      Why.LTrue
      class_fields
  with
    | Unsupported msg when not !Config.stop_on_error  ->
	Format.printf 
	  "**Warning: translation of class invariant of %s not performed due to unsupported feature in translation: %s@." 
	  class_name msg;
	Why.LTrue
    | Typing.Ill_typed msg when not !Config.stop_on_error  -> 
	Format.printf 
	  "**Warning: translation of class invariant of %s not performed due to typing errors (%s)@."
	  class_name msg;
	Why.LTrue	
;;


let collect_class_invariant class_name class_fields no =
  try
    List.fold_left
      (fun acc (doc,f) ->
	 match f with
	   | Specification s ->
	       List.fold_left
		 (fun acc s ->
		   match s with
		   | Representation_invariant e
		   | Invariant e -> Why.make_and acc (tr_assertion "" e)
		   | Static_invariant _
		   | Typed_represents _
		   | Represents _
		   | Jml_variable_declaration _
		     -> acc 
		 )
		 acc s
	   | Typed_variable(ve,init) ->
	       if ve.instance_variable_entry_is_static
	       then acc
	       else 
		 let ty = ve.instance_variable_entry_type in
		 begin
		   match ty with
		     | Byte_type | Short_type | Integer_type | Boolean_type | Float_type | Logic_type _ -> acc
(*                     | Array_type(Class_type(ce)) -> 
                         let v =
                            Why.LApp
				    ("acc",
				     [Why.LVar 
				        ve.instance_variable_entry_trans_name;
				      Why.LVar "this" ])
                         in
                         (Why.LAnd
                           (Why.LPred
			      ("instanceof",
			       [Why.LVar "alloc";
			        v;
				tr_type_term ty;]),
                            Why.LForall
                              ("i",
                               ([], "int"),
                               Why.LImpl
                                 (Why.LAnd
                                    (Why.LPred("le_int", [Why.LConst (Why.Prim_int "0"); Why.LVar "i"]),
                                     Why.LPred("lt_int", [Why.LVar "i"; Why.LApp("arraylength", [Why.LVar "alloc"; v])])),
                                  Why.LPred
                                    ("instanceof", 
                                     [Why.LVar "alloc"; 
                                      Why.LApp("array_acc", [Why.LVar "objA"; v; Why.LVar "i"]); 
                                      Why.LApp("ClassType", [Why.LVar ce.class_entry_name])]))))) *)
		     | _ ->
			 (Why.make_and
(*
			   (Why.LPred
			      ("wf_memory",
			       [Why.LVar "alloc";
				tr_type_term ty;
				Why.LVar 
				  ve.instance_variable_entry_trans_name
			       ]))
*)
			   (if no then Why.LTrue else
			   (Why.LPred
			      ("instanceof",
			       [Why.LVar "alloc";
				Why.LApp
				  ("acc",
				   [Why.LVar 
				      ve.instance_variable_entry_trans_name;
				    Why.LVar "this" ]);
				tr_type_term ty;
			       ])))
			   acc)
		 end
	   | _ -> acc)
      Why.LTrue
      class_fields
  with
    | Unsupported msg when not !Config.stop_on_error  ->
	Format.printf 
	  "**Warning: translation of class invariant of %s not performed due to unsupported feature in translation: %s@." 
	  class_name msg;
	Why.LTrue
    | Typing.Ill_typed msg when not !Config.stop_on_error  -> 
	Format.printf 
	  "**Warning: translation of class invariant of %s not performed due to typing errors (%s)@."
	  class_name msg;
	Why.LTrue	
;;

let collect_static_invariant class_name class_fields =
  try
    List.fold_left
      (fun acc (doc,f) ->
	match f with
	| Specification s ->
	    List.fold_left
	      (fun acc s ->
		match s with
		| Representation_invariant e
		| Static_invariant e -> Why.make_and acc (tr_assertion "" e)
		| Invariant _
		| Typed_represents _
		| Represents _
		| Jml_variable_declaration _
		  -> acc 
	      )
	      acc s
	| Typed_variable(ive, init) ->
	    if ive.instance_variable_entry_is_static
	    then  
	      let ty = ive.instance_variable_entry_type in
	      begin
		match ty with
		| Byte_type | Short_type | Integer_type | Boolean_type | Float_type | Logic_type _ -> acc
		| _ ->
		    (Why.make_and
		       (Why.LPred
			  ("instanceof",
			   [Why.LVar "alloc";
			    Why.LVar ive.instance_variable_entry_trans_name;
			    tr_type_term ty]))
		       acc)
	      end
            else acc
	| _ -> acc)
      Why.LTrue
      class_fields
  with
  | Unsupported msg when not !Config.stop_on_error  ->
      Format.printf 
	"**Warning: translation of static invariant of %s not performed due to unsupported feature in translation: %s@." 
	class_name msg;
      Why.LTrue
  | Typing.Ill_typed msg when not !Config.stop_on_error  -> 
      Format.printf 
	"**Warning: translation of static invariant of %s not performed due to typing errors (%s)@."
	class_name msg;
      Why.LTrue	
;;


(*
let collect_class_representation_invariant class_name class_fields =
  try
    List.fold_left
      (fun (acc, lvl) (doc,f) ->
	 match f with
	   | Specification s ->
	       List.fold_left
		 (fun (acc, lvl) s ->
		   match s with
		   | Representation_invariant e -> (Why.make_and acc (tr_assertion "" e), lvl)
		   | Invariant _
		   | Typed_represents _
		   | Represents _
		   | Jml_variable_declaration _
		     -> (acc, lvl) 
		 )
		 (acc, lvl) s
	   | Typed_variable(ve,init) ->
	       if ve.instance_variable_entry_is_static
	       then acc, lvl
	       else 
		 let ty = ve.instance_variable_entry_type in
		 begin
		   match ty with
                     | Logic_type t -> (acc, ve::lvl)
		     | _ -> (acc, lvl)
		 end
	   | _ -> (acc, lvl))
      (Why.LTrue, [])
      class_fields
  with
    | Unsupported msg when not !Config.stop_on_error  ->
	Format.printf 
	  "**Warning: translation of class invariant of %s not performed due to unsupported feature in translation: %s@." 
	  class_name msg;
	(Why.LTrue, [])
    | Typing.Ill_typed msg when not !Config.stop_on_error  -> 
	Format.printf 
	  "**Warning: translation of class invariant of %s not performed due to typing errors (%s)@."
	  class_name msg;
	(Why.LTrue, [])	
;;
*)

(*
let tr_representation_invariant ce c =
  let pre = Why.LTrue in
  let rt = Why.unit_type in
  let (r, w) = tr_effects ce.class_entry_representation_invariant_effects in
  let (post, lvl) = collect_class_representation_invariant ce.class_entry_name c.class_fields in
  let why_lvl = List.map (fun ive -> ive.instance_variable_entry_trans_name) lvl in
  let w_logic = List.filter (fun ive -> List.mem ive r) why_lvl in
  let w = List.append w_logic w in
  (* add 'modifiable' assertions *)
  let post =
    List.fold_left
    (fun post s ->
       Why.LAnd(post, Why.LPred("modifiable", [Why.LVar("alloc"); Why.LVar(s ^ "@"); Why.LVar(s); Why.LApp("value_loc", [Why.LVar("this")])])))
    post
    w
  in
  let exc = [] in
  let wt = Why.Annot_type(pre, rt, r, w, post, exc) in
  if post = Why.LTrue then Why.Void else Why.BlackBox(wt)
;;
*)

let tr_class_declaration ce c cus =
  let class_name = ce.class_entry_name in
  try
    let inv = collect_class_invariant class_name c.class_fields false in
    let no_inv = collect_class_invariant class_name c.class_fields true in
    let static_inv = collect_static_invariant class_name c.class_fields in
    let restore_inv = collect_class_restore_invariant class_name c.class_fields in
    let inv_name = class_name ^ "_invariant" in
    let no_inv_name = class_name ^ "_no_invariant" in
    let static_inv_name = class_name ^ "_static_invariant" in
    let restore_inv_name = class_name ^ "_restore_invariant" in
    let inv_init_name = class_name ^ "_invariant_init" in
(*
    let model_ensures = tr_model_ensures ce c in 
*)
    let create_decl =
(*i ??
      if List.mem `FINAL c.class_modifiers then ([],[])
      else
i*)
      if !Config.javacard && !Config.cardtear
      then
	(begin  
           if ce.class_entry_name <> "TransactionException"
	   then 
             [Why.Logic(false, class_name,[],([],"classId"));
	      Why.Logic
	        (false,
	         inv_init_name,
	         List.fold_right
	           (fun (s,t) acc -> (s,t)::acc)
	           ce.class_entry_invariant_args
	           [("this",([],"value (*" ^ class_name ^ "*)"))],
	         ([],"prop"))] 
           else 
             [Why.Logic
	        (false,
	         inv_init_name,
	         List.fold_right
	           (fun (s,t) acc -> (s,t)::acc)
	           ce.class_entry_invariant_args
	           [("this",([],"value (*" ^ class_name ^ "*)"))],
	         ([],"prop"))] 
         end,
	   [Why.Predicate
	      (false,
	       inv_name,
	       (List.fold_right
	          (fun (s,t) acc -> (s,t)::acc)
		  ce.class_entry_invariant_args
		  ["this", ([], "value")]),
	       inv);
            Why.Predicate
	      (false,
	       restore_inv_name,
	       (List.fold_right
	          (fun (s,t) acc -> (s,t)::acc)
		  ce.class_entry_restore_invariant_args
		  ["this", ([], "value")]),
	       restore_inv);
	    Why.Predicate
	      (false,
	       inv_init_name,
	       (List.fold_right
	          (fun (s,t) acc -> (s,t)::acc)
		  ce.class_entry_invariant_args
		  ["this", ([], "value")]),
	       inv)])          
        else
	  (begin	  
             if ce.class_entry_name = "TransactionException"
	     then
               []
	     else
	       [Why.Logic(false, class_name, [], ([], "classId"));] 
           end,
             if !Config.javacard
             then
	       [Why.Predicate
	          (false,
	           inv_name,
	           (List.fold_right
	              (fun (s,t) acc -> (s,t)::acc)
		      ce.class_entry_invariant_args
		      ["this",([],"value")]),
	           inv);
	        Why.Predicate
	          (false,
	           restore_inv_name,
	           (List.fold_right
	              (fun (s,t) acc -> (s,t)::acc)
		      ce.class_entry_restore_invariant_args
		      ["this", ([], "value")]),
	           restore_inv)]
             else
	       let static_inv_pred =
		 if List.length ce.class_entry_static_invariant_args <> 0 then
		   [Why.Predicate
	              (false,
	               static_inv_name,
		       ce.class_entry_static_invariant_args,
	               static_inv)]
		 else []
	       in
	       let no_inv_pred =
		if !Config.nonnull_only then
		  [Why.Predicate
	             (false,
	              no_inv_name,
	              (List.fold_right
			 (fun (s,t) acc -> (s,t)::acc)
			 ce.class_entry_invariant_args
			 ["this",([],"value")]),
	              no_inv)]
		else [] 
	       in
	       (Why.Predicate
	          (false,
	           inv_name,
	           (List.fold_right
	              (fun (s,t) acc -> (s,t)::acc)
		      ce.class_entry_invariant_args
		      ["this",([],"value")]),
	           inv)) 
	       ::(no_inv_pred @ static_inv_pred))
    in
    List.fold_left
      (fun (wdacc,cdacc) (_,f) -> 
	 let (wd,cd) = tr_field_declaration ce f c cus inv restore_inv
	 in 
	 if !Config.javacard
         then
           match f with
            | Typed_variable(ve,init) -> 
               let ve = 
                 if ve.instance_variable_entry_is_ghost
                 then ve
                 else { ve with instance_variable_entry_trans_name = ve.instance_variable_entry_trans_name ^ "_restore" } 
               in
               let (wdr, cdr) = tr_field_declaration ce (Typed_variable(ve, init)) c cus inv restore_inv in
               (List.rev_append wdr (List.rev_append wd wdacc), List.rev_append cd cdacc) 
            | _ -> (List.rev_append wd wdacc,List.rev_append cd cdacc)
         else 
           (List.rev_append wd wdacc,List.rev_append cd cdacc))
      create_decl
      c.class_fields
      
  with
    | Unsupported msg when not !Config.stop_on_error  ->
	Format.printf 
	  "**Warning: translation of class %s not performed due to unsupported feature in translation: %s@." 
	  class_name msg;
	([],[])
    | Typing.Ill_typed msg when not !Config.stop_on_error  -> 
	Format.printf 
	  "**Warning: translation of class %s not performed due to typing errors (%s)@."
	  class_name msg;
	([],[])	
;;



let tr_interface_member_declaration interface_name interface_invariant interface_inv_args f =
  try
    match f with
    | Constant_declaration(_) -> 
	ill_typed "tr_interface_member_declaration(constant)"
    | Typed_constant_declaration(ive,i) ->
	let name = interface_name ^ "_" ^ ive.constant_entry_name 
	and t = tr_type ive.constant_entry_type
	in
	let coq_decl =
	  match i with
	    | None -> [Why.Logic(false,name,[],t)]
	    | Some vi ->
		begin
		  match vi with
		    | Simple_initializer _ ->
			[Why.Logic(false,name,[],t) ;
			 Why.Axiom
			   (name ^ "_value", 
			    Why.LPred("eq",
				      [Why.LVar(name);
				       tr_term_variable_initializer 
					 Location.dummy vi]))
			]
		    | Array_initializer _ -> 
			[Why.Logic(false,name,[],t)]
		end
	in
	([Why.Logic(false,name,[],t)],coq_decl)
    | Method_header(_) -> 
	ill_typed "tr_interface_member_declaration(method)"
    | Typed_method_header(me,jml) -> 
	(* ([],[]) *)
	tr_method_declaration true
	  interface_name interface_inv_args [] (* Why.Void *) None jml me None [] Why.LTrue Why.LTrue
    | Model_variable _ ->
        ill_typed "tr_interface_member_declaration(variable)"
    | Typed_model_variable ive ->
        if ive.instance_variable_entry_is_static
	then tr_static_variable_declaration Location.dummy ive None
	else tr_instance_variable_declaration interface_name ive
    | Interface_specification _ -> ([],[]) (* already computed before *)
  with
  | Unsupported msg when not !Config.stop_on_error ->
      Format.printf 
	"**Warning: translation of interface member %s not performed due to unsupported feature in translation: %s@." 
	"(?)" msg;
      ([],[])
  | Typing.Ill_typed msg when not !Config.stop_on_error  -> 
      Format.printf 
	"**Warning: translation of interface member %s not performed due to typing errors (%s)@."
	"(?)" msg;
      ([],[])	
;;

let tr_interface_declaration ie i =
  let n = ie.interface_entry_name in
  let inv_name = n ^ "_invariant" in
  let no_inv_name = n ^ "_no_invariant" in
  let restore_inv_name = n ^ "_restore_invariant" in
  let no_inv_logic =
    if !Config.nonnull_only then
      [Why.Logic(false,
		 no_inv_name,
		 [("this", ([],"value"))],
		 ([], "prop"))]
    else []
  in
  let no_inv_pred =
    if !Config.nonnull_only then
      [Why.Predicate
	 (false,
	  no_inv_name,
	  ["this", ([], "value")],
	  Why.LTrue)]
    else []
  in
  let why_decls =
    [(* Why.External(n,Why.base_type("classId")); *)
      Why.Logic(false,n,[],([],"classId"));
      Why.Logic(false,
		inv_name,
		[("this",([],"value (*" ^ n ^ "*)"))],
		([],"prop"));
      Why.Logic(false,
		restore_inv_name,
		[("this",([],"value (*" ^ n ^ "*)"))],
		([],"prop"));] 
    @ no_inv_logic
  in
  let coq_decls =
    [Why.Predicate
       (false,
	inv_name,
	["this",([],"value")],
	Why.LTrue);
     Why.Predicate
       (false,
	restore_inv_name,
	["this",([],"value")],
	Why.LTrue)]
    @ no_inv_pred
  in
  List.fold_left
    (fun (wdacc,cdacc) (_,f) -> 
       let (wd,cd) = tr_interface_member_declaration 
		       (snd i.interface_name) (Lit(Bool(true))) [] f
       in 
       if !Config.javacard
       then
         match f with
          | Typed_model_variable(ive) -> 
              let ive = 
                if ive.instance_variable_entry_is_ghost
                then ive
                else { ive with instance_variable_entry_trans_name = ive.instance_variable_entry_trans_name ^ "_restore" } 
              in
              let (wdr, cdr) = tr_interface_member_declaration (snd i.interface_name) (Lit(Bool(true))) [] (Typed_model_variable(ive)) in
               (List.rev_append wdr (List.rev_append wd wdacc), List.rev_append cd cdacc) 
          | _ -> (List.rev_append wd wdacc,List.rev_append cd cdacc)
       else 
         (List.rev_append wd wdacc, List.rev_append cd cdacc))
    (why_decls,coq_decls)
    i.interface_members
;;

let tr_logic_type id idl =
  let e = Why.Type(id, idl) in
  ([e],[])

let tr_axiom ae e =
  let e =
    List.fold_right
      (fun (x,t) acc ->
	 Why.LForall(x,t,acc))
      (Effects.params_reads ae.axiom_entry_effects [] false)
      (tr_assertion "" e)
  in
  ([Why.Axiom(ae.axiom_entry_name,e)],[])

let tr_logic_reads le r =
  let ret_ty = tr_type le.logic_entry_return_type in
  let ty =
    List.fold_right
      (fun (_,pe) acc ->
	 let (a,b) = tr_parameter pe in
	 (a,b)::acc)
      le.logic_entry_parameters
      []
  in
  let all_p = Effects.params_reads le.logic_entry_effects ty false in
  let e = Why.Logic(false,le.logic_entry_name,all_p,ret_ty) in
  ([e],[])

let tr_logic_def le e =
  let ty =
    List.fold_right
      (fun (_,pe) acc ->
	 let (a,b) = tr_parameter pe in
	 (a,b)::acc)
      le.logic_entry_parameters
      []
  in
  let all_p = Effects.params_reads le.logic_entry_effects ty false in
  let e = 
    match le.logic_entry_return_type with
      | Prop_type ->
	  Why.Predicate(false,le.logic_entry_name,all_p,tr_assertion "" e) 
      | t ->
	  Why.Function(false,le.logic_entry_name,all_p,tr_type t,tr_term ""  (fun id -> Why.LVar(id)) e) 
  in
	([e],[])

let tr_type_declaration d cus = 
  match d with
  | Interface _ -> assert false
  | Class _ -> assert false
  | Axiom(ae,e) -> tr_axiom ae e
  | Logic_reads(id,t,p,r) -> assert false
  | Typed_logic_reads(le,r) -> tr_logic_reads le r
  | Logic_def(id,t,p,e) -> assert false
  | Typed_logic_def(le,e) -> tr_logic_def le e
  | Logic_typedecl(id, idl) -> tr_logic_type (snd id) (List.map snd idl)
  | Typed_interface(ie, i) -> tr_interface_declaration ie i
  | Typed_class(ce,c) -> tr_class_declaration ce c cus
;;

let tr_compilation_unit a cus = 
  List.fold_left
    (fun (wdacc,cdacc) (_,d) -> 
       let (wd,cd) = tr_type_declaration d cus
       in (List.rev_append wd wdacc,List.rev_append cd cdacc)) 
    ([],[])
    a.cu_type_decls

(*
Local Variables: 
mode: tuareg
compile-command: "make"
End: 
*)

