Require Export local_memory_why.
Set Implicit Arguments.

Notation " this # f " := (acc f this) (at level 30, f at level 0).
Notation " mem : this [ i ] " := (array_acc mem this i)
  (at level 30, this at level 0, i at level 50).

Hint Unfold is_valid_index.
Hint Resolve instanceof_typeof.

Ltac Simpl1Hyps :=
  match goal with
  | id1:?X1,id2:?X1 |- _ => clear id2
  | id:True |- _ => clear id
  | id:(?X1 = ?X1) |- _ => clear id
  | id:(?X1 /\ ?X2) |- _ =>
      decompose record id; clear id
  | id:(store_extends ?X1 ?X1) |- _ =>
      clear id
  | id:(modifiable ?X2 ?X1 ?X1 ?X3) |- _ =>
      clear id
  | id:(array_modifiable ?X2 ?X1 ?X1 ?X3) |- _ => clear id
  | _ => fail
  end.

Ltac SimplHyps := repeat Simpl1Hyps.

Ltac Acc_upd_eq := match goal with 
  |- context [acc (upd ?m ?v ?w) ?t] => rewrite acc_upd || 
               (casetype (w=t); [progress auto with *|rewrite acc_upd])
| |- context [array_acc (array_upd ?m ?v ?i ?w) ?t ?j] => rewrite array_acc_upd ||
               (casetype (i=j); [omega | casetype (v=t); [progress auto with *|rewrite array_acc_upd]])
end.

Ltac Acc_upd1 :=
  Acc_upd_eq ||
          (rewrite acc_upd_other; [ idtac | progress auto with * ]) ||
            (rewrite array_acc_upd_other;
              [ idtac | progress auto with * ]); auto with *.

Ltac krakatoa := repeat Acc_upd1.

Ltac CleanMod :=
  match goal with
  | id:(store_extends ?X1 ?X1) |- _ =>
      clear id; CleanMod
  | id:(modifiable ?X2 ?X1 ?X1 ?X3) |- _ =>
      clear id; CleanMod
  | id:(array_modifiable ?X2 ?X1 ?X1 ?X3) |- _ => clear id; CleanMod
  | id:(?X1 = ?X1) |- _ => clear id; CleanMod
  | _ => idtac
  end.
(*
Ltac Subtype :=
  match goal with
  |  |- (subType ?X1 ?X2) => exact (dec_A_proof (subType_dec X1 X2) I)
  end.

Ltac NotSubtype :=
  match goal with
  |  |- (~ subType ?X1 ?X2) =>
      exact (dec_not_A_proof (subType_dec X1 X2) I)
  end.
*)

Ltac StoreRec :=
  match goal with
(* Allways succeed *)
  | id:(store_extends ?X1 ?X2) |- (store_extends ?X1 ?X2) => 
  exact id
  |  |- (store_extends ?X1 ?X1) =>
      exact (store_extends_refl X1)
(* X2<>X3, generates (store_extends X2 X3 *)
  | id:(store_extends ?X1 ?X2) |- (store_extends ?X1 ?X3) =>
      apply (store_extends_trans (s1:=X1) (s2:=X2) (s3:=X3) id); StoreRec
 (* X2<>X1, generates (store_extends X1 X2 *)
 | id:(store_extends ?X2 ?X3) |- (store_extends ?X1 ?X3) =>
      apply (store_extends_trans (s1:=X1) (s2:=X2) (s3:=X3));
      [StoreRec | exact id]
  end.

Ltac Store := CleanMod; StoreRec.

Ltac P n m := constr:(pair n m).

Ltac Instance :=
  match goal with
(* allways succeed *)
  | id:(instanceof ?X1 ?X2 ?X3) |- (instanceof ?X1 ?X2 ?X3) =>
      exact id
  | id:(typeof ?X1 ?X2 ?X3) |- (instanceof ?X1 ?X2 ?X3) =>
      exact (instanceof_typeof id)
(* generates (subType X3 X4) equivalent goal *)
   | id:(typeof ?X1 ?X2 (ArrayType ?X3)) |- (instanceof ?X1 ?X2 (ArrayType ?X4)) =>
     apply (instanceof_array (t1:=X3) (t2:=X4)) with (2:=id)
 | id:(instanceof ?X1 ?X2 ?X3) |- (instanceof ?X4 ?X2 ?X3) =>
(* generates (store_extends X1 X4) may be not provable *)
      apply (instanceof_extends_stable (s1:=X1) (s2:=X4) (t:=X3) (v:=X2)) with (2:=id); 
      StoreRec
(*
  | id:(instanceof ?X1 ?X2 ?X3) |- (instanceof ?X4 ?X2 ?X5) =>
    match (P X3 X5) with (pair ?X ?X) =>
(* generates (store_extends X1 X4) may be not provable *)
      apply (instanceof_extends_stable (s1:=X1) (s2:=X4) (t:=X3) (v:=X2)) with (2:=id); 
      StoreRec
      | _ => (* X3 <> X5 *)
(* generates (instanceof X4 X2 X3) and (subtype X3 X5) *)
      apply (instanceof_subType_trans (h:=X4) (this:=X2) (t1:=X3));
       [ Instance | Subtype ]
    end
*)
  end.

Hint Resolve instanceof_null.

Lemma null_alive : forall s, alive s Null.
intro s; assert (instanceof s Null ArrIntType); auto.
exact (instanceof_alive H).
Save.
Hint Resolve null_alive.

Lemma alive_extends_stable : forall s1 s2 v, store_extends s1 s2 -> alive s1 v -> alive s2 v.
repeat red; intros.
apply H0; apply store_extends_fresh with s2; auto.
Save.

Ltac AliveRec :=
  match goal with
 (* Allways succeed *)
  | id:(alive ?X1 ?X2) |- (alive ?X1 ?X2) => exact id
  |  |- (alive ?X1 Null) =>      apply null_alive
  | id:(instanceof ?X1 ?X2 ?X3) |- (alive ?X1 ?X2) =>
      exact (instanceof_alive (h:=X1) (t:=X3) (v:=X2) id)
  | id:(typeof ?X1 ?X2 ?X3) |- (alive ?X1 ?X2) => let Ha := fresh in
      assert (Ha:=instanceof_typeof id); exact (instanceof_alive (h:=X1) (t:=X3) (v:=X2) Ha)
(* generates (alive X2 X4) *)
  | id:(?X1 = ?X2) |- (alive ?X1 ?X4) =>
      rewrite id; AliveRec
(* generates (alive X1 X3) *)
  | id:(store_extends ?X1 ?X2) |- (alive ?X2 ?X3) =>
      apply (alive_extends_stable (s1:=X1) (s2:=X2) (v:=X3) id); AliveRec
  end.

Ltac Alive := CleanMod; AliveRec.

Ltac Fresh :=
  match goal with
  (* Allways succeed *)
 | h1:(fresh ?X1 ?X2) |- (fresh ?X1 ?X2) => exact h1
 (* generates (store_extends X1 X3) *)
 | h1:(fresh ?X3 ?X2) |- (fresh ?X1 ?X2) =>
      apply (store_extends_fresh (s1:=X1) (s2:=X3)) with (2:=h1); progress auto
(* generates (fresh X1 X3) *)
  | h1:(?X2 = ?X3) |- (fresh ?X1 ?X2) => rewrite h1; Fresh
 (* generates (fresh X3 X2) *)
  | h1:(store_extends ?X1 ?X3) |- (fresh ?X3 ?X2) =>
      apply (store_extends_fresh (s1:=X1) (s2:=X3) h1); Fresh
  end.

Lemma fresh_alive_diff : forall t v1 v2, fresh t v1 -> alive t v2 -> v1 <> v2.
intros; apply sym_not_eq.
apply alive_fresh_diff with t; auto.
Save.

Lemma fresh_not_null : forall t v, fresh t v -> v <> Null.
intros; apply fresh_alive_diff with t; auto.
Save.

Lemma arraylength_pos_val_not_null : forall t v, 0 < arraylength t v -> v <> Null.
Admitted.

Ltac Neqv :=
  match goal with
 (* Allways succeed *)
 | id:(fresh ?X1 ?X2) |- (?X2 <> Null) =>
      apply (fresh_not_null (t:=X1) (v:=X2) id)
(* generates (alive X1 X3) *)
  | id:(fresh ?X1 ?X2) |- (?X2 <> ?X3) =>
      apply (fresh_alive_diff (t:=X1) (v1:=X2) (v2:=X3) id); Alive
(* generates (alive X1 X3) *)
  | id:(fresh ?X1 ?X2) |- (?X3 <> ?X2) =>
      apply (alive_fresh_diff (t:=X1) (v1:=X3) (v2:=X2));
       [ Alive | exact id ]
 (* generates arraylength X1 X2 > 0 *)
 | id:?X3 = (arraylength ?X1 ?X2) |- (?X2 <> Null) =>
      apply (arraylength_pos_val_not_null (t:=X1) (v:=X2));
       progress auto with zarith
  | id:(arraylength ?X1 ?X2 = ?X3) |- (?X2 <> Null) =>
      apply (arraylength_pos_val_not_null (t:=X1) (v:=X2));
       progress auto with zarith
  | id:(?X3 <= arraylength ?X1 ?X2) |- (?X2 <> Null) =>
      apply (arraylength_pos_val_not_null (t:=X1) (v:=X2));
       progress auto with zarith
  | id:(?X3 < arraylength ?X1 ?X2) |- (?X2 <> Null) =>
      apply (arraylength_pos_val_not_null (t:=X1) (v:=X2));
       progress auto with zarith
  | id:(arraylength ?X1 ?X2 >= ?X3) |- (?X2 <> Null) =>
      apply (arraylength_pos_val_not_null (t:=X1) (v:=X2));
       progress auto with zarith
  | id:(arraylength ?X1 ?X2 > ?X3) |- (?X2 <> Null) =>
      apply (arraylength_pos_val_not_null (t:=X1) (v:=X2));
       progress auto with zarith
(* generates X2 <> Null *)
  |  |- (Null <> ?X2) => apply sym_not_eq; Neqv
  end.

(*
Ltac ModRec :=
  match goal with
(* Allways succeed *)
  | id:(modifiable ?X1 ?X2 ?X3 ?X4) |- (modifiable ?X1 ?X2 ?X3 ?X4) =>
      exact id
  |  |- (modifiable ?X1 ?X2 ?X2 ?X4) =>
      exact (modifiable_refl (h:=X1) X2 (unchanged:=X4))
(* generates (modifiable X1 X7 X2 X3) *)
  | id:(fresh ?X1 ?X3) |- (modifiable ?X1 ?X7 (upd ?X2 ?X3 ?X4) ?X6) =>
      apply
       (modifiable_fresh_upd (h:=X1) (m1:=X7) (m2:=X2) (unch:=X6) (v:=X3)
          X4 id); ModRec
  |  |- (modifiable ?X1 ?X7 (upd ?X2 ?X3 ?X4) ?X6) =>
           match (P X7 X2) with (pair ?X ?X) => 
(* generates ~(X6 X3) *)
           apply (modifiable_upd (h:=X1) X (unch:=X6) (v:=X3) X4); subst;
           krakatoa; progress auto 
                                                 | _ => (*X2 <>X7 *)
(* generates (modifiable X1 X7 X2 X6) et (modifiable X1 X2 (upd X2 X3 X4) X6)*)
           apply (modifiable_trans (h:=X1) (m1:=X7) (m2:=X2) 
                      (m3:=(upd X2 X3 X4)) (unch:=X6)); ModRec
            end
(* generates (modifiable X1 X4 X3 X5) *)
  | id:(?X2 = ?X3) |- (modifiable ?X1 ?X4 ?X2 ?X5) =>
      rewrite id; ModRec
(* generates X7 subset X4 and (store_extends X1 X5) 
    and (modifiable X5 X6 X2 X7) *)
  | id:(modifiable ?X1 ?X2 ?X3 ?X4) |- (modifiable ?X5 ?X6 ?X3 ?X7) =>
      apply modifiable_trans_sub with (4:=id);
       [ subst; krakatoa; unfold inter_loc; intuition; fail 
       | Store | ModRec ]
  end.

Ltac Modifiable := CleanMod; ModRec.

Ltac ArrayModRec :=
  match goal with
 (* Allways succeed *)
  | id:(array_modifiable ?X1 ?X2 ?X3 ?X4) |-
  (array_modifiable ?X1 ?X2 ?X3 ?X4) => exact id
  |  |- (array_modifiable ?X1 ?X2 ?X2 ?X4) =>
      exact (array_modifiable_refl (h:=X1) X2 (unchanged:=X4))
(* generates (array_modifiable X1 X7 X2 X6) *)
  | id:(fresh ?X1 ?X3) |-
  (array_modifiable ?X1 ?X7 (array_upd ?X2 ?X3 ?X4 ?X8) ?X6) =>
      apply
       (array_modifiable_fresh_upd (h:=X1) (m1:=X7) (m2:=X2) (unch:=X6)
          (v:=X3) X4 X8 id); ArrayModRec
  |  |- (array_modifiable ?X1 ?X7 (array_upd ?X2 ?X3 ?X4 ?X8) ?X6) =>
           match (P X7 X2) with (pair ?X ?X) => 
(* generates ~(X6 X3 X4) *)
     apply
       (array_modifiable_upd (h:=X1) X2 (unch:=X6) (v:=X3) (i:=X4) X8);
       subst; krakatoa; progress auto
                                                 | _ => (*X2 <>X7 *)
(* generates (array_modifiable X1 X7 X2 X6) and
                      (array_modifiable X1 X2 (upd X2 X3 X4 X8) X6)*)
           apply (array_modifiable_trans (h:=X1) (m1:=X7) (m2:=X2) 
                      (m3:=(array_upd X2 X3 X4 X8)) (unch:=X6)); ArrayModRec
            end
(* generates (modifiable X1 X4 X3 X5) *)
  | id:(?X2 = ?X3) |- (array_modifiable ?X1 ?X4 ?X2 ?X5) =>
      rewrite id; ArrayModRec
(* generates X7 subset X4 and (store_extends X1 X5) *)
  | id:(array_modifiable ?X1 ?X2 ?X3 ?X4) |-
  (array_modifiable ?X5 ?X6 ?X3 ?X7) =>
      apply
       (array_modifiable_trans_sub (h1:=X5) (h2:=X1) (m1:=X6) (m2:=X2)
          (m3:=X3) (unch1:=X7) (unch2:=X4));
       [ subst; krakatoa; unfold array_inter_loc; intuition; fail 
       | Store | ArrayModRec | exact id ]
  end.


Ltac ArrayModifiable := CleanMod; ArrayModRec.
*)

Hint Extern 0 (alive _ _) => Alive.
Hint Extern 0 (fresh _ _) => Fresh.
Hint Extern 0 (store_extends _ _) => Store.
Hint Extern 0 (instanceof _ _ _) => Instance.
Hint Extern 0 (_ <> _) => Neqv.


